// ============================================================================
// $Id: Append.java,v 1.1 2008/02/08 04:24:05 davidahall Exp $
// Copyright (c) 2006  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.algorithms;

import java.util.Collections;
import java.util.Iterator;

import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.util.ArrayUtils;
import net.sf.jga.util.EmptyIterator;

/**
 * Algorithms that iterate over variable-length sets of inputs, ie, arrays, collections, or 
 * iterations, returning the combined contents of all inputs.  For the common case, iterating 
 * over the combined contents of two inputs, the Merge class provides a type-safe wrapper
 * around these algorithms.  These algorithms are more flexible than the append forms of Merge,
 * at the cost of being less type-safe.
 * <p>
 * NOTE: Use of these algorithms <b>will</b> result in type-safety warnings from the compiler: 
 * specifically "A generic array of ... is created for a varargs parameter".  This is pretty much 
 * unavoidable given the mismatch in java between generics and arrays and given that variable length 
 * argument lists are converted to arrays by the compiler. 
 * <p>
 * NOTE: Unlike most other classes in this package, Append does not support the form that puts the 
 * result of the algorithm into a collection passed as the last argument.  The CollectionUtils 
 * class contains an append method that handles one input, and the Merge class contains an append
 * method that handles two inputs.  A varargs form of append to a given collection would be 
 * ambiguous in this class if it had the same name, and it would be ambiguous with the append in
 * either CollectionUtils or Merge (or both) if all were statically imported.  If you need to append
 * three or more inputs to a given collection, the workaround is to call the appropriate method in
 * this class to reduce the inputs to one, then use Collection.addAll (or CollectionUtils.append). 
 * <p>
 * Copyright &copy; 2008  David A. Hall
 */

public class Append {

    //  A function that returns an iterable for an array
    @SuppressWarnings("serial")
    static private class Iterablerator<T> extends UnaryFunctor<T[], Iterable<? extends T>> {
            public Iterable<T> fn(T[] input) {
                return ArrayUtils.iterable(input);
            }
        };
        
    /**
     * Returns an empty iterable.  Special handling for the degenerate case.
     */
    static public <T> Iterable<? extends T> append() {
        return Collections.emptyList();
    }
        
    // ============
    // Arrays
    // ============
    
    /**
     * Returns all elements of each array
     */
    static public <T> Iterable<T> append(T[] ts) {
        return ArrayUtils.iterable(ts);
    }
 
    /**
     * Returns all elements of each array
     */
    static public <T> Iterable<T> append(T[]... ts) {
        return new AppendIterable<T>(ts);
    }
 
    // ============
    // Iterables
    // ============
    
    /**
     * Returns an Iterable that 'contains' all of the elements of the input(s)
     */
    static public <T> Iterable<? extends T> append(Iterable<? extends T>... inputs) {
        return new AppendIterable<T>(inputs);
    }
    
    // ============
    // Iterators
    // ============
    
    /**
     * Returns an Iterator that will return elements from all of the iterators
     */
    static public <T> Iterator<? extends T> append(Iterator<? extends T>... input) {
        return new AppendIterator<T>(input);
    }

    // ===============
    // AppendIterable
    // ===============
   
    /**
     * Produces iterators that return the contents of a two-dimensional data structure
     */
    @SuppressWarnings("serial")
    static public class AppendIterable<T> implements Iterable<T> {

        // the base resource
        private Iterable<? extends Iterable<? extends T>> base;
    
        // A function that returns an iterator for an iterable resource
        private UnaryFunctor<Iterable<? extends T>, Iterator<? extends T>> iteratorator =
            new UnaryFunctor<Iterable<? extends T>,Iterator<? extends T>>(){
                public Iterator<? extends T> fn(Iterable<? extends T> input) {
                    return input.iterator();
                }
        };

        /**
         * Builds a AppendIterable for the given set of inputs
         * @param input
         */
        public AppendIterable(Iterable<? extends T>... input) {
            this(ArrayUtils.iterable(input));
        }

        /**
         * Builds a AppendIterable for the given set of inputs
         */
        public AppendIterable(Iterable<? extends Iterable<? extends T>> input) {
            base = input;
        }
        
        /**
         * Builds a AppendIterable for the given set of inputs
         */
        public AppendIterable(T[]... input) {
            this.base = Transform.transform(input, new Iterablerator<T>());
        }
        
        // - - - - - - - - - - -
        // Iterable<T> interface
        // - - - - - - - - - - -

        @SuppressWarnings("serial")
        public Iterator<T> iterator() {
            return new AppendIterator<T>(Transform.transform(base.iterator(), iteratorator));
        }
    }

    
    /**
     * Iterator that returns the contents of a two-dimensional datastructure.
     */
    static public class AppendIterator<T> implements Iterator<T> {
    
        // The base set of iterators to Append
        private Iterator<? extends Iterator<? extends T>> _base;

        // The current iterator being consumed
        private Iterator<? extends T> _crnt = new EmptyIterator<T>();

        // A tri state flag that controls advancement of the current/base iterators
        private Boolean _hasNext = null;
    
        /**
         * Builds a Append.Iterator for the given base iterators
         */

        public AppendIterator(Iterator<? extends T>... iters) {
            this(ArrayUtils.iterate(iters));
        }
        
        /**
         * Builds a Append.Iterator for the given base iterators
         */

        public AppendIterator(Iterator<? extends Iterator<? extends T>> iters) {
            _base = iters;
        }
        
        // - - - - - - - - - - -
        // Iterator<T> interface
        // - - - - - - - - - - -
        
        public boolean hasNext() {
            if (_hasNext == null)
                advance();
            
            return _hasNext;
        }

        public T next() {
            if (_hasNext == null)
                advance();

            _hasNext = null;
            return _crnt.next();
        }

        public void remove() {
            _crnt.remove();
        }

        // - - - - - - - - - - - - -
        
        private void advance() {
            if (_hasNext != null)
                return;
            
            if (_crnt.hasNext()) {
                _hasNext = Boolean.TRUE;
                return;
            }
            
            while (!_crnt.hasNext() && _base.hasNext()) {
                _crnt = _base.next();
            }
            
            _hasNext = _crnt.hasNext();
        }
    }
}
