// ============================================================================
// $Id: Compare.java,v 1.2 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2007  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not
// use this file except in compliance with the License.
//
// From time to time, the license steward (initially Sun Microsystems,
// Inc.), and may publish revised and/or new versions of this License.
// You may not use, distribute, or otherwise make this file available
// under subsequent versions of the License.
//
// You should have received a copy of the the License along with this
// file: if not, a copy of the License is available at
//
// http://www.sun.com/cddl/cddl.html
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// ============================================================================

package net.sf.jga.algorithms;

import java.io.Serializable;
import java.util.Comparator;
import java.util.Iterator;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.comparison.EqualTo;
import net.sf.jga.util.ComparableComparator;

import static net.sf.jga.util.ArrayUtils.*;

/**
 * Algorithms and utilities that compare two sets of input.  The inputs may be arrays, collections,
 * or iterations (although, in the case of iterations, the comparison also consumes elements from
 * the iteration).
 * <p>
 * Copyright &copy; 2007  David Hall
 */

public class Compare {
    
    // ============
    // Arrays
    // ============
    
    /**
     * Returns true if the two arrays are equal, using the equals() method to compare
     * the elements of each array
     * @return true if the two arrays are equal
     */
    static public <T> boolean equal(T[] a1, T[] a2) {
        return equal(iterate(a1), iterate(a2), new EqualTo<T>());
    }

    
    /**
     * Returns true if the two arrays are equal, using the given comparator to compare the
     * elements in each array
     * @return true if the two arrays are equal.
     */
    static public <T> boolean equal(T[] a1, T[] a2, Comparator<T> comp) {
        return equal(iterate(a1), iterate(a2), comp);
    }

    
   /**
    * Returns true if the two arrays are equal, using the given functor to compare the elements
    * in each arrays.  The functor is expected to evaluate its two argments and return true if
    * they are "equal", therefore this method returns true if the arrays contain the same number
    * of elements and if the functor returns true for all pairs of elements.
    * @return true if the two collections are equal
    */
    static public <T> boolean equal(T[] a1, T[] a2, BinaryFunctor<T,T,Boolean> eq) {
        return equal(iterate(a1), iterate(a2), eq);
    }

    
    /**
     * Returns true if the first array is lexically less than the second, using the default
     * comparison operation to compare the elements in each array.
     * @return true if c1 < c2
     */
    static public <T extends Comparable<? super T>> boolean lessThan(T[] a1, T[] a2) {
        return lessThan(iterate(a1), iterate(a2));
    }

    
    /**
     * Returns true if the first array is lexically less than the second,
     * using the given comparator to compare the elements in each array.
     * @return true if c1 < c2
     */
    static public <T> boolean lessThan(T[] a1, T[] a2, Comparator<T> comp) {
        return lessThan(iterate(a1), iterate(a2), comp);
    }

    
    /**
     * Returns true if the first array is lexically less than the second,
     * using the given operator to compare the elements in each collection.  The
     * first is less than the second if it is not longer than the second and if
     * the first corresponding element that is not equal is less.
     * @return true if c1 < c2
     */
    static public <T> boolean lessThan(T[] a1, T[] a2, BinaryFunctor<T,T,Boolean> lt) {
        return lessThan(iterate(a1), iterate(a2), lt);
    }

    
    /**
     * Compares the contents of the two arrays using a default comparator.  An element from each
     * array is passed to the comparator, and if they are not equal, the resource that produced the
     * lesser of the two elements is determined to be the lesser of the two resources.  If the
     * elements are equal, then the next pair of elements if passed.  If one of the two arrays is
     * exhausted before a pair of unequal elements is found, then it is considered to be the lesser.
     * <p>
     * @return a negative integer if a1 < a2, a positive integer if a1 > a2, 0 if a1 == a2
     */
    static public <T extends Comparable<? super T>> int compare(T[] a1, T[] a2) {
        return compare(iterate(a1), iterate(a2));
    }
    
    /**
     * Compares the contents of the two arrays using a binary predicate that returns TRUE when its
     * first argument is less than its second.  An element from each array is compared, and if they
     * are not equal, the resource that produced the lesser of the two elements is determined to be
     * the lesser.  If the elements are equal, then the next pair of elements is passed.  If one of
     * the two arrays is exhausted before a pair of unequal elements is found, then it is considered
     * to be the lesser.
     * <p>
     * @return a negative integer if a1 < a2, a positive integer if a1 > a2, 0 if a1 == a2 
     */
    static public <T> int compare(T[] a1, T[] a2, BinaryFunctor<T,T,Boolean> bf) {
        return compare(iterate(a1), iterate(a2), bf);
    }
    
    /**
     * Compares the contents of the two arrays using the given comparator.  An element from each
     * array is passed to the given comparator, and if they are not equal, the array that produced
     * the lesser of the two elements is determined to be the lesser.  If the elements are equal,
     * then the next pair of elements if passed.  If one of the two arrays is exhausted before a
     * pair of unequal elements is found, then it is considered to be the lesser.
     * <p>
     * @return a negative integer if a1 < a2, a positive integer if a1 > a2, 0 if a1 == a2
     */
    static public <T> int compare(T[] a1, T[] a2, Comparator<T> comp) {
        return compare(iterate(a1), iterate(a2), comp);
    }
    
    // ============
    // Iterables
    // ============

    /**
     * Returns true if the two resources are equal, using the equals() method to compare
     * the elements of each resource
     * @return true if the two resources are equal
     */
    static public <T> boolean equal(Iterable<? extends T> c1, Iterable<? extends T> c2) {
        return equal(c1.iterator(), c2.iterator(), new EqualTo<T>());
    }

    
    /**
     * Returns true if the two resources are equal, using the given comparator to compare the
     * elements in each resource
     * @return true if the two resources are equal.
     */
    static public <T> boolean
    equal(Iterable<? extends T> c1, Iterable<? extends T> c2, Comparator<T> comp) {
        return equal(c1.iterator(), c2.iterator(), comp);
    }

    
   /**
    * Returns true if the two resources are equal, using the given functor to compare the elements
    * in each resources.  The functor is expected to evaluate its two argments and return true if
    * they are "equal", therefore this method returns true if the resources contain the same number
    * of elements and if the functor returns true for all pairs of elements.
    * @return true if the two collections are equal
    */
    static public <T> boolean
    equal(Iterable<? extends T> c1, Iterable<? extends T> c2, BinaryFunctor<T,T,Boolean> eq) {
        return equal(c1.iterator(), c2.iterator(), eq);
    }

    
    /**
     * Returns true if the first resource is lexically less than the second, using the default
     * comparison operation to compare the elements in each resource.
     * @return true if c1 < c2
     */
    static public <T extends Comparable<? super T>> boolean
    lessThan(Iterable<? extends T> c1, Iterable<? extends T> c2) {
        return lessThan(c1.iterator(), c2.iterator());
    }

    
    /**
     * Returns true if the first resource is lexically less than the second,
     * using the given comparator to compare the elements in each resource.
     * @return true if c1 < c2
     */
    static public <T> boolean
    lessThan(Iterable<? extends T> c1, Iterable<? extends T> c2, Comparator<T> comp) {
        return lessThan(c1.iterator(), c2.iterator(), comp);
    }

    
    /**
     * Returns true if the first resource is lexically less than the second,
     * using the given operator to compare the elements in each collection.  The
     * first is less than the second if it is not longer than the second and if
     * the first corresponding element that is not equal is less.
     * @return true if c1 < c2
     */
    static public <T> boolean
    lessThan(Iterable<? extends T> c1, Iterable<? extends T> c2, BinaryFunctor<T,T,Boolean> lt) {
        return lessThan(c1.iterator(), c2.iterator(), lt);
    }

    
    /**
     * Compares the contents of the two iterable resources using a default comparator.  An element
     * from each resource is passed to the comparator, and if they are not equal, the resource that
     * produced the lesser of the two elements is determined to be the lesser. If the elements are
     * equal, then the next pair of elements if passed.  If one of the two resources is exhausted
     * before a pair of unequal elements is found, then it is considered to be the lesser.
     * <p>
     * @return a negative integer if c1 < c2, a positive integer if c1 > c2, 0 if c1 == c2
     */
    static public <T extends Comparable<? super T>> int
    compare(Iterable<? extends T> c1, Iterable<? extends T> c2) {
        return compare(c1.iterator(), c2.iterator());
    }

    
    /**
     * Compares the contents of the two iterable resources using a binary predicate that returns
     * TRUE when its first argument is less than its second.  An element from each resource is
     * passed to the given comparator, and if they are not equal, the resource that produced the
     * lesser of the two elements is determined to be the lesser.  If the elements are equal, then
     * the next pair of elements if passed.  If one of the two resources is exhausted before a pair
     * of unequal elements is found, then it is considered to be the lesser.
     * <p>
     * @return a negative integer if c1 < c2, a positive integer if c1 > c2, 0 if c1 == c2
     */
    static public <T> int
    compare(Iterable<? extends T> c1, Iterable<? extends T> c2, BinaryFunctor<T,T,Boolean> bf) {
        return compare(c1.iterator(), c2.iterator(), bf);
    }

    
    /**
     * Compares the contents of the two iterable resources.  An element from each resources is
     * passed to the given comparator, and if they are not equal, the resource that produced the
     * lesser of the two elements is determined to be the lesser.  If the elements are equal, then
     * the next pair of elements if passed.  If one of the two resources is exhausted before a pair
     * of unequal elements is found, then it is considered to be the lesser.
     * <p>
     * @return a negative integer if c1 < c2, a positive integer if c1 > c2, 0 if c1 == c2
     */
    static public <T> int
    compare(Iterable<? extends T> c1, Iterable<? extends T> c2, Comparator<? super T> comp) {
        return compare(c1.iterator(), c2.iterator(), comp);
    }
    
    // ============
    // Iterators
    // ============
    
    /**
     * Returns true if the two iterations are equal, using the Comparable
     * interface to compare elements in the iterations.
     * @return true if the two iterations are equal
     */
    static public <T extends Comparable<? super T>> boolean
    equal(Iterator<? extends T> i1, Iterator<? extends T> i2)
    {
        return equal(i1, i2, new ComparableComparator<T>());
    }

    
    /**
     * Returns true if the two iterations are equal, using the given Comparator
     * to compare elements in the iterations.
     * @return true if the two iterations are equal
     */
    static public <T> boolean
    equal(Iterator<? extends T> i1, Iterator<? extends T> i2, Comparator<T> comp)
    {
        return compare(i1, i2, comp) == 0;
    }

    
    /**
     * Returns true if the two iterations are equal, using the given
     * BinaryFunctor to compare elements in the iterations.
     * @return true if the two iterations are equal
     */
    static public <T> boolean
    equal(Iterator<? extends T> i1, Iterator<? extends T> i2, BinaryFunctor<T,T,Boolean> eq)
    {
        while (i1.hasNext() && i2.hasNext()) {
            if (!eq.fn(i1.next(),i2.next()).booleanValue())
                return false;
        }

        return i1.hasNext() == i2.hasNext();
    }

    
    /**
     * Returns true if the first iterator is lexically less than the second,
     * using the default comparison operation to compare the elements in each
     * iterator.
     * @return true if the first iteration is less than the second
     */
    static public <T extends Comparable<? super T>> boolean
    lessThan(Iterator<? extends T> i1, Iterator<? extends T> i2)
    {
        return compare(i1, i2, new ComparableComparator<T>()) < 0;
    }

    
    /**
     * Returns true if the first iterator is lexically less than the second,
     * using the given comparator to compare the elements in each iterator.
     * @return true if the first iteration is less than the second
     */
    static public <T> boolean
    lessThan(Iterator<? extends T> i1, Iterator<? extends T> i2, Comparator<T> comp)
    {
        return compare(i1, i2, comp) < 0;
    }

    
    /**
     * Returns true if the first iterator is lexically less than the second,
     * using the given operator to compare the elements in each iterator.  The
     * first is less than the second if it is not longer than the second and if
     * the first corresponding element that is not equal is less.
     * @return true if the first iteration is less than the second
     */
    static public <T> boolean
    lessThan(Iterator<? extends T> i1, Iterator<? extends T> i2, final BinaryFunctor<T,T,Boolean> lt)
    {
        int result = compare(i1, i2, new Comparator<T>() {
            public int compare(T x,T y) {
                return lt.fn(x,y).booleanValue() ? -1 :
                       lt.fn(y,x).booleanValue() ? 1 : 0;
            }
        });

        return result < 0;
    }
    
    /**
     * Compares the contents of the two iterations using a default comparator.  An element from each
     * iteration is passed to the comparator, and if they are not equal, the iteration that produced
     * the lesser of the two elements is determined to be the lesser. If the elements are equal,
     * then the next pair of elements if passed.  If one of the two iterations is exhausted before a
     * pair of unequal elements is found, then it is considered to be the lesser.  If both iterations
     * are exhausted before a pair of unequal elements is found, then the iterations are determined
     * to be equal.
     * <p>
     * Neither iteration will be advanced past the point where an unequal pair is found, but the
     * elements that were unequal will have been consumed.
     * 
     * @return a negative integer if i1 < i2, a positive integer if i1 > i2, 0 if i1 == i2
     */
    static public <T extends Comparable<? super T>> int
    compare(Iterator<? extends T> i1, Iterator<? extends T> i2) {
        return compare(i1, i2, new ComparableComparator<T>());
    }
    
    /**
     * Compares the contents of the two iterations using a predicate that returns TRUE if its first
     * argument is less than its second.  An element from each iteration is passed to the comparator,
     * and if they are not equal, the iteration that produced the lesser of the two elements is
     * determined to be the lesser.  If the elements are equal, then the next pair of elements if
     * passed.  If one of the two iterations is exhausted before a pair of unequal elements is found,
     * then it is considered to be the lesser.  If both iterations are exhausted before a pair of
     * unequal elements is found, then the iterations are determined to be equal.
     * <p>
     * Neither iteration will be advanced past the point where an unequal pair is found, but the
     * elements that were unequal will have been consumed.
     * 
     * @return a negative integer if i1 < i2, a positive integer if i1 > i2, 0 if i1 == i2
     */
    static public <T> int
    compare(Iterator<? extends T> i1, Iterator<? extends T> i2, final BinaryFunctor<T,T,Boolean> bf)
    {
        return compare(i1, i2, new Comparator<T>() {
            public int compare(T i1, T i2) {
                return bf.fn(i1, i2) ? -1 : bf.fn(i2, i1) ? 1 : 0;
            }
        });
    }
    
    /**
     * Compares the contents of the two iterations using the given comparator.  An element from each
     * iteration is passed to the comparator, and if they are not equal, the iteration that produced
     * the lesser of the two elements is determined to be the lesser.  If the elements are equal,
     * then the next pair of elements if passed.  If one of the two iterations is exhausted before a
     * pair of unequal elements is found, then it is considered to be the lesser.  If both iterations
     * are exhausted before a pair of unequal elements is found, then the iterations are determined
     * to be equal.
     * <p>
     * Neither iteration will be advanced past the point where an unequal pair is found, but the
     * elements that were unequal will have been consumed.
     * 
     * @return -1 if i1 < i2, 1 if i1 > i2, 0 if i1 == i2
     */
    static public <T> int
    compare(Iterator<? extends T> i1, Iterator<? extends T> i2, Comparator<? super T> comp)
    {
        while (true) {
            if (i1.hasNext()) {
                if (i2.hasNext()) {
                    int i = comp.compare(i1.next(), i2.next());
                    if( i != 0) {
                        return i;
                    }
                }
                else
                    return 1;
            }
            else 
                return i2.hasNext() ? -1 : 0;
        }
    }
    
    // ========================
    // IteratorComparator class
    // ========================
    
    static public class IteratorComparator<T>
        implements Comparator<Iterator<? extends T>>, Serializable {

        static final long serialVersionUID = 1110218880817939318L;
        
        // The comparator used to compare the elements of the two argument comparators
        private Comparator<? super T> _comp;
        
        /**
         * Builds an IteratorComparator that uses the given Comparator to compare
         * the elements of an iteration.
         */
        public IteratorComparator(Comparator<? super T> comp) {
            _comp = comp;
        }
        
        public int compare(Iterator<? extends T> i1, Iterator<? extends T> i2) {
            return Compare.compare(i1, i2, _comp);
        }
    }
}
