// ============================================================================
// $Id: Filter.java,v 1.9 2008/12/20 21:07:50 davidahall Exp $
// Copyright (c) 2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.algorithms;

import java.util.Collection;
import java.util.Iterator;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.fn.comparison.Equality;
import net.sf.jga.fn.comparison.NotEqualTo;
import net.sf.jga.fn.logical.UnaryNegate;
import net.sf.jga.util.FindIterator;

import static net.sf.jga.util.ArrayUtils.*;
import static net.sf.jga.util.CollectionUtils.*;

/**
 * Algorithms that return a subset of its input.  The input may be an array, collection, or
 * iteration.  There are two general algorithms included: filter and remove.  The filter
 * methods return elements from the input that pass the selection criteria, while the remove
 * methods removes elements from the input stream for which the selection criteria is true.
 * Note that the actual input is not modified in any of these methods.
 * <p>
 * Copyright &copy; 2006  David A. Hall
 */

public class Filter {
    
    // ============
    // Arrays
    // ============
    
    /**
     * Returns the elements of the array that meet the given selection criteria.
     */
    static public <T> Iterable<T> filter(T[] ts, UnaryFunctor<T,Boolean> pred) {
        return new FilterIterable<T>(iterable(ts), pred);
    }
    
    /**
     * Returns the elements of the array that are not equal to the given value, using the value's
     * {@link #equals(Object)} method
     */
    static public <T> Iterable<T> remove(T[] ts, T value) {
        return filter(ts, new NotEqualTo<T>().bind2nd(value));
    }

    /**
     * Returns the elements of the array that are not equal to the given value, using the given
     * functor to compare for equality
     */
    static public <T> Iterable<T> remove(T[] ts, T value, Equality<T> eq) {
        return filter(ts, new UnaryNegate<T>(eq.bind2nd(value)));
    }

    /**
     * Returns the elements of the array that do not meet the given selection criteria
     */
    static public <T> Iterable<T> remove(T[] ts, UnaryFunctor<T,Boolean> pred) {
        return filter(ts, new UnaryNegate<T>(pred));
    }

    // ============
    // Iterables
    // ============
    
    /**
     * Returns the contents of the input that meet the given selection criteria.
     */
    static public <T> Iterable<T> filter(Iterable<? extends T> i, UnaryFunctor<T,Boolean> pred) {
        return new FilterIterable<T>(i, pred);
    }

    /**
     * Returns the elements of the input that are not equal to the given value, using the value's
     * {@link #equals(Object)} method
     */
    static public <T> Iterable<T> remove(Iterable<? extends T> i, T value) {
        return filter(i, new NotEqualTo<T>().bind2nd(value));
    }

    /**
     * Returns the elements of the input that are not equal to the given value, using the given
     * functor to compare for equality
     */
    static public <T> Iterable<T> remove(Iterable<? extends T> i, T value, Equality<T> eq) {
        return filter(i, new UnaryNegate<T>(eq.bind2nd(value)));
    }

    /**
     * Returns the elements of the input that do not meet the given selection criteria
     */
    static public <T> Iterable<T> remove(Iterable<? extends T> i, UnaryFunctor<T,Boolean> pred) {
        return filter(i, new UnaryNegate<T>(pred));
    }
    
    // ============
    // Iterators
    // ============
    
    /**
     * Returns the contents of the iterator that meet the given selection criteria.
     */
    static public <T> Iterator<T> filter(Iterator<? extends T> iter, UnaryFunctor<T,Boolean> pred) {
        return new FilterIterator<T>(iter, pred);
    }
    
    /**
     * Returns the elements of the iterator that are not equal to the given value, using the value's
     * {@link #equals(Object)} method
     */
    static public <T> Iterator<T> remove(Iterator<? extends T> iter, T value) {
        return filter(iter, new NotEqualTo<T>().bind2nd(value));
    }

    /**
     * Returns the elements of the iterator that are not equal to the given value, using the given
     * functor to compare for equality
     */
    static public <T> Iterator<T> remove(Iterator<? extends T> iter, T value, Equality<T> eq) {
        return filter(iter, new UnaryNegate<T>(eq.bind2nd(value)));
    }

    /**
     * Returns the elements of the iterator that do not meet the given selection criteria
     */
    static public <T> Iterator<T> remove(Iterator<? extends T> iter,UnaryFunctor<T,Boolean> pred){
        return filter(iter, new UnaryNegate<T>(pred));
    }
    
    // ============
    // IterableCopy
    // ============
    
    /**
     * Appends the contents of the input that meet the given selection criteria to the output.
     */
    static public <T, TCollection extends Collection<? super T>> TCollection
    filter(Iterable<? extends T> cin, UnaryFunctor<T,Boolean> pred, TCollection cout) {
        return append(cout, filter(cin.iterator(), pred));
    }
    

    /**
     * Returns the elements of the input that are not equal to the given value, using the value's
     * {@link #equals(Object)} method
     */
    static public <T, TCollection extends Collection<? super T>> TCollection
    remove(Iterable<? extends T> cin, T value, TCollection cout) {
        return append(cout, remove(cin.iterator(), value));
    }

    
    /**
     * Returns the elements of the input that are not equal to the given value, using the given
     * functor to compare for equality
     */
    static public <T, TCollection extends Collection<? super T>> TCollection
    remove(Iterable<? extends T> cin, T value, Equality<T> eq, TCollection cout) {
        return append(cout, remove(cin.iterator(), value, eq));
    }

    
    /**
     * Returns the elements of the input that do not meet the given selection criteria
     */
    static public <T, TCollection extends Collection<? super T>> TCollection
    remove(Iterable<? extends T> cin, UnaryFunctor<T,Boolean> pred, TCollection cout) {
        return append(cout, remove(cin.iterator(), pred));
    }

    // ==============
    // FilterIterable
    // ==============
    
    /**
     * Produces Iterators that only return elements that meet a given condition.
     */

    // Suggested by sourceforge user Jerry Vos:
    // https://sourceforge.net/tracker/index.php?func=detail&aid=1276988&group_id=49942&atid=458041

    static public class FilterIterable<T> implements Iterable<T> {
    
        private Iterable<? extends T> _delegate;

        // predicate used to test elements of the base iterator.
        private UnaryFunctor<T,Boolean> _filter;

        /**
         * Builds a FilterIterator that will return only qualifying elements of
         * the given iterable.
         */
        public FilterIterable(Iterable<? extends T> iter,
                              UnaryFunctor<T,Boolean> pred)
        {
            _delegate = iter;
            _filter = pred;
        }
    
        // - - - - - - - - - - -
        // Iterable<T> interface
        // - - - - - - - - - - -

        public Iterator<T> iterator() {
            return new FilterIterator<T>(_delegate.iterator(), _filter);
        }
    }

    
    /**
     * Iterator that only returns elements that meet the given selection criteria.
     **/

    static public class FilterIterator<T> implements Iterator<T> {

        // base iterator
        private FindIterator<T> _base;

        // predicate used to test elements of the base iterator.
        private UnaryFunctor<T,Boolean> _filter;

        // flag indicating that the base iterator has been advanced to the next
        // qualifying element (or off the end if no remaining element qualifies)
        private boolean _testedNext;

        /**
         * Builds a FilterIterator that will return only qualifying elements of
         * the given iterator.
         */
        public FilterIterator(Iterator<? extends T> iter, UnaryFunctor<T,Boolean> pred) {
            _base = new FindIterator<T>(iter);
            _filter = pred;
        }

        // - - - - - - - - - - - - -
        // Iterator<T> interface
        // - - - - - - - - - - - - -
    
        public boolean hasNext(){
            _testedNext = true;
            return _base.findNext(_filter);
        }

        public T next() {
            if (!_testedNext) {
                hasNext();
            }
        
            _testedNext = false;
            return _base.next();
        }

        public void remove() { throw new UnsupportedOperationException(); }
    }
}
