// ============================================================================
// $Id: Find.java,v 1.6 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.algorithms;

import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.fn.algorithm.ElementOf;
import net.sf.jga.fn.comparison.EqualTo;
import net.sf.jga.fn.comparison.Equality;
import net.sf.jga.fn.comparison.NotEqualTo;
import net.sf.jga.util.EmptyIterator;
import net.sf.jga.util.FindIterator;
import net.sf.jga.util.LookAheadIterator;

import static net.sf.jga.fn.comparison.ComparisonFunctors.*;
import static net.sf.jga.util.ArrayUtils.*;

/**
 * Algorithms that return an iterator pointing at the first/next element of the input
 * that meets some condition.
 * <p>
 * <b>NOTE</b> To find and return a value in the input, see the various flavors of
 * {@link Summarize#lookup(Iterable,Object)}
 */

public class Find {
    
    // ============
    // Arrays
    // ============

    /**
     * Finds an arbitrary value in the array using the equals() method.
     * @return an iterator whose next() [if it hasNext()]
     * will return the first instance of the value.  If the value is not in the
     * array, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> find(T[] ts, T value) {
        return find(iterate(ts), equalTo(value));
    }

    /**
     * Finds an arbitrary value in the array using the given Comparator.
     * @return an iterator whose next() [if it hasNext()]
     * will return the first instance of the value.  If the value is not in the
     * array, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> find(T[] ts, T value, Comparator<? super T> comp){
        return find(iterate(ts), equalTo(comp, value));
    }

    
    /**
     * Finds an arbitrary value in the array using the given Equality operator.
     * @return an iterator whose next() [if it hasNext()]
     * will return the first instance of the value.  If the value is not in the
     * array, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> find(T[] ts, T value, Equality<T> eq){
        return find(iterate(ts), equalTo(eq, value));
    }

    
    /**
     * Finds a value in the input for which the given function returns TRUE.
     * @return an iterator whose next() [if it hasNext()]
     * will return the first instance of such a value.  If no such value is in the
     * array, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> find(T[] ts, UnaryFunctor<T,Boolean> fn) {
        return find(iterate(ts), fn);
    }
 
    /**
     * Finds first/next element in the array that is also in the values array.
     * @return an iterator whose next() [if it hasNext()] will return the first instance of any
     * element found in the array of values.  If no such element is found then the returned
     * iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> findElement(T[] ts, T[] values) {
        return find(iterate(ts), elementOf(values));
    }

    
    /**
     * Finds first/next element in the array that is also in the collection of values, using
     * the collection's contains() method.
     * @return an iterator whose next() [if it hasNext()] will return the first instance of any
     * element found in the collection of values.  If no such element is found  then the returned
     * iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> findElement(T[] ts, Collection<? extends T> values) {
        return find(iterate(ts), elementOf(values) );
    }

    /**
     * Finds first/next element in the array that is also in the values array, using
     * the given Comparator to compare values.
     * @return an iterator whose next() [if it hasNext()] will return the first instance of any
     * element found in the array of values.  If no such element is found then the returned
     * iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> findElement(T[] ts, T[] values, Comparator<? super T> comp) {
        return find(iterate(ts), elementOf(values, comp));
    }

    
    /**
     * Finds first/next element in the array that is also in the values collection, using
     * the given functor to compare values.
     * @return an iterator whose next() [if it hasNext()] will return the first instance of any
     * element found in collection of values.  If no such value is found then the returned
     * iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findElement(T[] ts, Collection<? extends T> values, Comparator<? super T> comp) {
        return find(iterate(ts), elementOf(values, comp));
    }
    
    /**
     * Finds first/next element in the array that is also in the values array, using
     * the given functor to compare values.
     * @return an iterator whose next() [if it hasNext()] will return the first instance of any
     * element found in the array of values.  If no such element is found then the returned
     * iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> findElement(T[] ts, T[] values, BinaryFunctor<T,T,Boolean> bf) {
        return find(iterate(ts), elementOf(values, bf));
    }

    
    /**
     * Finds first/next element in the input that is in the values collection, using
     * the given functor to compare values.
     * @return an iterator whose next() [if it hasNext()] will return the first instance of any
     * element found in collection of values.  If no such value is found then the returned
     * iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findElement(T[] ts, Collection<? extends T> values, BinaryFunctor<T,T,Boolean> bf) {
        return find(iterate(ts), elementOf(values, bf));
    }
    

    /**
     * Locates the first/next pair of adjacent elements in the array that
     * are the same value, using the equals() method.
     * @return an iterator whose next() [if it hasNext()] points to the first of
     * a pair of adjacent equivalent values.  If no such pair exists, then the
     * iterator's hasNext() will be false.
     */
    static public <T> Iterator<T> findAdjacent(T[] ts) {
        return findAdjacent(iterate(ts), new EqualTo<T>());
    }

    
    /**
     * Locates the first/next pair of adjacent elements in the array that
     * are the same value, using the given comparator.
     * @return an iterator whose next() [if it hasNext()] points to the first of
     * a pair of adjacent equivalent values.  If no such pair exists, then the
     * iterator's hasNext() will be false.
     */
    static public <T> Iterator<T> findAdjacent(T[] ts, Comparator<? super T> comp) {
        return findAdjacent(iterate(ts), new EqualTo<T>(comp));
    }

    
    /**
     * Locates the first/next pair of adjacent elements in the array that
     * are the same value, using the given functor.  The functor should return
     * TRUE when two values are to be considered equal.
     * @return an iterator whose next() [if it hasNext()] points to the first of
     * a pair of adjacent equivalent values.  If no such pair exists, then the
     * iterator's hasNext() will be false.
     */
    static public <T> Iterator<T> findAdjacent(T[] ts, BinaryFunctor<T,T,Boolean> eq) {
        return findAdjacent(iterate(ts), eq);
    }

    
    /**
     * Finds first/next arbitrary length run of a given value in the input.  Runs of length zero
     * are well-defined: every input begins with a run of length zero of all possible values.
     * @return an iterator based on the given iterator whose next() [if it hasNext()] will return
     * the first of n adjacent instances of value.  If no run of values of the requested length
     * exist in the iteration, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> findRepeated (T[] ts, int count, T value) {
        return findRepeated(iterate(ts), count, equalTo(value));
    }
    
    /**
     * Finds first/next arbitrary length run of a given value in the input using the given
     * comparator.  Runs of length zero are well-defined: every input begins with
     * a run of length zero of all possible values.
     * @return an iterator based on the given iterator whose next() [if it hasNext()] will return
     * the first of n adjacent instances of value.  If no run of values of the requested length
     * exist in the iteration, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findRepeated (T[] ts, int count, T value, Comparator<? super T> comp) {
        return findRepeated(iterate(ts), count, equalTo(comp, value));
    }
    
    /**
     * Finds first/next arbitrary length run of a given value in the input using the given
     * equality function.  Runs of length zero are well-defined: every input begins with
     * a run of length zero of all possible values.
     * @return an iterator based on the given iterator whose next() [if it hasNext()] will return
     * the first of n adjacent instances of value.  If no run of values of the requested length
     * exist in the iteration, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> findRepeated (T[] ts, int count, T value, Equality<T> eq) {
        return findRepeated(iterate(ts), count, equalTo(eq, value));
    }
    
    /**
     * Finds first/next arbitrary length run of elements in the input for which the
     * given function returns TRUE.  Runs of length zero are well-defined: every
     * input begins with a run of length zero of all possible values.
     * @return an iterator based on the given iterator whose next() [if it
     * hasNext()] will return the first of n adjacent instances of value.  If no
     * run of values of the requested length exist in the iteration, then the
     * returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> findRepeated (T[] ts, int count, UnaryFunctor<T,Boolean> eq) {
        return findRepeated(iterate(ts), count, eq);
    }

    /**
     * Finds the given pattern in the input using the equals method.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T> findSequence(T[] ts, T[] pattern) {
        return findSequence(iterate(ts), Arrays.asList(pattern), new EqualTo<T>());
    }

    
    /**
     * Finds the given pattern in the input using the equals method.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T> findSequence(T[] ts, Collection<? extends T> pattern) {
        return findSequence(iterate(ts), pattern, new EqualTo<T>());
    }
    
    
    /**
     * Finds the given pattern in the input using the given comparator to determine equivalence.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T> findSequence(T[] ts, T[] pattern, Comparator<? super T> comp) {
        return findSequence(iterate(ts), Arrays.asList(pattern), new EqualTo<T>(comp));
    }
    
    
    /**
     * Finds the given pattern in the input using the given comparator to determine equivalence.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T>
    findSequence(T[] ts, Collection<? extends T> pattern, Comparator<T> comp) {
        return findSequence(iterate(ts), pattern, new EqualTo<T>(comp));
    }

    
    /**
     * Finds the given pattern in the input using the given functor to determine equivalence.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T> findSequence(T[] ts, T[] pattern, BinaryFunctor<T,T,Boolean> fn) {
        return findSequence(iterate(ts), Arrays.asList(pattern), fn);
    }
    
    
    /**
     * Finds the given pattern in the input using the given functor to determine equivalence.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T>
    findSequence(T[] ts, Collection<? extends T> pattern, BinaryFunctor<T,T,Boolean> fn) {
        return findSequence(iterate(ts), pattern, fn);
    }

    
    /**
     * Finds the point at which the array differs from the pattern.  
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the array
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the array but is shorter, the returned iterator will point at the next element in the array
     * after the length of the pattern.
     */
    static public <T> Iterator<T> findMismatch(T[] ts, T[] pattern){
        return findMismatch(iterate(ts), Arrays.asList(pattern), new NotEqualTo<T>());        
    }
    
    /**
     * Finds the point at which the array differs from the pattern.
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the array
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the array but is shorter, the returned iterator will point at the next element in the array
     * after the length of the pattern.
     */
    static public <T> Iterator<T> findMismatch(T[] ts, Collection<? extends T> pattern) {
        return findMismatch(iterate(ts), pattern, new NotEqualTo<T>());
    }
    
    /**
     * Finds the point at which the array differs from the pattern, using the given comparator to
     * determine the mismatch.  
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the array
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the array but is shorter, the returned iterator will point at the next element in the array
     * after the length of the pattern.
     */
    static public <T> Iterator<T> findMismatch(T[] ts, T[] pattern, Comparator<? super T> comp) {
        return findMismatch(iterate(ts), Arrays.asList(pattern), new NotEqualTo<T>(comp));        
    }
    
    /**
     * Finds the point at which the array differs from the pattern, using the given comparator to
     * determine the mismatch.  
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the array
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the array but is shorter, the returned iterator will point at the next element in the array
     * after the length of the pattern.
     */
    static public <T> Iterator<T>
    findMismatch(T[] ts, Collection<? extends T> pattern, Comparator<? super T> comp) {
        return findMismatch(iterate(ts), pattern, new NotEqualTo<T>(comp));
    }
    
    /**
     * Finds the point at which the array differs from the pattern, using the given functor to
     * determine the mismatch.  The functor should return TRUE if a given pair of elements does
     * not match.
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the array
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the array but is shorter, the returned iterator will point at the next element in the array
     * after the length of the pattern.
     */
    static public <T> Iterator<T> findMismatch(T[] ts, T[] pattern, BinaryFunctor<T,T,Boolean> neq) {
        return findMismatch(iterate(ts), Arrays.asList(pattern), neq);        
    }
    
    /**
     * Finds the point at which the array differs from the pattern, using the given functor to
     * determine the mismatch.  The functor should return TRUE if a given pair of elements does
     * not match.
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the array
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the array but is shorter, the returned iterator will point at the next element in the array
     * after the length of the pattern.
     */
    static public <T> Iterator<T>
    findMismatch(T[] ts, Collection<? extends T> pattern, BinaryFunctor<T,T,Boolean> neq) {
        return findMismatch(iterate(ts), pattern, neq);
    }                                       

    // ============
    // Iterables
    // ============
    
    /**
     * Finds an arbitrary value in the input using the equals() method.
     * @return an iterator whose next() [if it hasNext()]
     * will return the first instance of the value.  If the value is not in the
     * input, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> find(Iterable<? extends T> iter, T value) {
        return find(iter, equalTo(value));
    }

    /**
     * Finds an arbitrary value in the input using the given Comparator.
     * @return an iterator whose next() [if it hasNext()]
     * will return the first instance of the value.  If the value is not in the
     * input, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    find(Iterable<? extends T> iter, T value, Comparator<? super T> comp) {
        return find(iter, equalTo(comp, value));
    }

    
    /**
     * Finds an arbitrary value in the input using the given Equality operator.
     * @return an iterator whose next() [if it hasNext()]
     * will return the first instance of the value.  If the value is not in the
     * input, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> find(Iterable<? extends T> iter, T value, Equality<T> eq) {
        return find(iter, equalTo(eq, value));
    }

    
    /**
     * Finds a value in the input for which the given function returns TRUE.
     * @return an iterator whose next() [if it hasNext()]
     * will return the first instance of the value.  If the value is not in the
     * input, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> find(Iterable<? extends T> iter, UnaryFunctor<T,Boolean> fn) {
        return find(iter.iterator(), fn);
    }

    
    /**
     * Finds first/next value in the input that is an element of the given array.
     * @return an iterator whose next() [if it hasNext()] will return
     * the first instance of any element found in the array of values.  If no such element is
     * found then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> findElement(Iterable<? extends T> iter, T[] values) {
        return find(iter.iterator(), elementOf(values));
    }

    
    /**
     * Finds first/next value in the input that is an element of the given collection, using
     * the collection's contains() method.
     * @return an iterator whose next() [if it hasNext()] will return
     * the first instance of any element found in the collection of values.  If no such element
     * is found  then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findElement(Iterable<? extends T> iter, Collection<? extends T> values) {
        return find(iter.iterator(), elementOf(values) );
    }

    /**
     * Finds first/next value in the input that is an element of the given array, using
     * the given Comparator to compare values.
     * @return an iterator whose next() [if it hasNext()] will return
     * the first instance of any element found in the array of values.  If no such element is
     * found then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findElement(Iterable<? extends T> iter, T[] values, Comparator<? super T> comp) {
        return find(iter.iterator(), elementOf(values, comp));
    }

    
    /**
     * Finds first/next value in the input that is an element of the given collection, using
     * the given functor to compare values.
     * @return an iterator whose next() [if it hasNext()] will return
     * the first instance of any element found in collection of values.  If no such value is
     * found then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findElement(Iterable<? extends T> iter, Collection<? extends T> values,
                Comparator<? super T> comp) {
        return find(iter.iterator(), elementOf(values, comp));
    }
    
    /**
     * Finds first/next value in the input that is an element of the given array, using
     * the given functor to compare values.
     * @return an iterator whose next() [if it hasNext()] will return
     * the first instance of any element found in the array of values.  If no such element is
     * found then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findElement(Iterable<? extends T> iter, T[] values, BinaryFunctor<T,T,Boolean> bf) {
        return find(iter.iterator(), elementOf(values, bf));
    }

    
    /**
     * Finds first/next value in the input that is an element of the given collection, using
     * the given functor to compare values.
     * @return an iterator whose next() [if it hasNext()] will return
     * the first instance of any element found in collection of values.  If no such value is
     * found then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findElement(Iterable<? extends T> iter, Collection<? extends T> values,
                BinaryFunctor<T,T,Boolean> bf)
    {
        return find(iter.iterator(), elementOf(values, bf));
    }
    
    
    /**
     * Locates the first/next pair of adjacent elements in the input that
     * are the same value, using the equals() method.
     * @return an iterator whose next() [if it hasNext()] points to the first of
     * a pair of adjacent equivalent values.  If no such pair exists, then the
     * iterator's hasNext() will be false.
     */
    static public <T> Iterator<T> findAdjacent(Iterable<? extends T> iter) {
        return findAdjacent(iter, new EqualTo<T>());
    }

    
    /**
     * Locates the first/next pair of adjacent elements in the input that
     * are the same value, using the given comparator.
     * @return an iterator whose next() [if it hasNext()] points to the first of
     * a pair of adjacent equivalent values.  If no such pair exists, then the
     * iterator's hasNext() will be false.
     */
    static public <T> Iterator<T>
    findAdjacent(Iterable<? extends T> iter, Comparator<? super T> comp) {
        return findAdjacent(iter, new EqualTo<T>(comp));
    }

    
    /**
     * Locates the first/next pair of adjacent elements in the input that
     * are the same value, using the given functor.  The functor should return
     * TRUE when two values are to be considered equal.
     * @return an iterator whose next() [if it hasNext()] points to the first of
     * a pair of adjacent equivalent values.  If no such pair exists, then the
     * iterator's hasNext() will be false.
     */
    static public <T> Iterator<T>
    findAdjacent(Iterable<? extends T> iter, BinaryFunctor<T,T,Boolean> eq) {
        return findAdjacent(iter.iterator(), eq);
    }
    
    /**
     * Finds first/next arbitrary length run of a given value in the input.  Runs of length zero
     * are well-defined: every input begins with a run of length zero of all possible values.
     * @return an iterator based on the given iterator whose next() [if it hasNext()] will return
     * the first of n adjacent instances of value.  If no run of values of the requested length
     * exist in the iteration, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> findRepeated (Iterable<? extends T> iter, int count, T value) {
        return findRepeated(iter.iterator(), count, equalTo(value));
    }
    
    /**
     * Finds first/next arbitrary length run of a given value in the input using the given
     * comparator.  Runs of length zero are well-defined: every input begins with
     * a run of length zero of all possible values.
     * @return an iterator based on the given iterator whose next() [if it hasNext()] will return
     * the first of n adjacent instances of value.  If no run of values of the requested length
     * exist in the iteration, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findRepeated (Iterable<? extends T> iter, int count, T value, Comparator<? super T> comp) {
        return findRepeated(iter.iterator(), count, equalTo(comp, value));
    }
    
    /**
     * Finds first/next arbitrary length run of a given value in the input using the given
     * equality function.  Runs of length zero are well-defined: every input begins with
     * a run of length zero of all possible values.
     * @return an iterator based on the given iterator whose next() [if it hasNext()] will return
     * the first of n adjacent instances of value.  If no run of values of the requested length
     * exist in the iteration, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findRepeated (Iterable<? extends T> iter, int count, T value, Equality<T> eq) {
        return findRepeated(iter.iterator(), count, equalTo(eq, value));
    }
    
    /**
     * Finds first/next arbitrary length run of elements in the input for which the
     * given function returns TRUE.  Runs of length zero are well-defined: every
     * input begins with a run of length zero of all possible values.
     * @return an iterator based on the given iterator whose next() [if it
     * hasNext()] will return the first of n adjacent instances of value.  If no
     * run of values of the requested length exist in the iteration, then the
     * returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findRepeated (Iterable<? extends T> iter, int count, UnaryFunctor<T,Boolean> eq) {
        return findRepeated(iter.iterator(), count, eq);
    }


    /**
     * Finds the given pattern in the input using the equals method.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T>
    findSequence(Iterable<? extends T> iter, T[] pattern) {
        return findSequence(iter.iterator(), Arrays.asList(pattern), new EqualTo<T>());
    }

    
    /**
     * Finds the given pattern in the input using the equals method.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T>
    findSequence(Iterable<? extends T> iter, Collection<? extends T> pattern) {
        return findSequence(iter.iterator(), pattern, new EqualTo<T>());
    }
    
    
    /**
     * Finds the given pattern in the input using the given comparator to determine equivalence.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T>
    findSequence(Iterable<? extends T> iter, T[] pattern, Comparator<? super T> comp) {
        return findSequence(iter.iterator(), Arrays.asList(pattern), new EqualTo<T>(comp));
    }
    
    
    /**
     * Finds the given pattern in the input using the given comparator to determine equivalence.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T>
    findSequence(Iterable<? extends T> iter, Collection<? extends T> pattern, Comparator<T> comp) {
        return findSequence(iter.iterator(), pattern, new EqualTo<T>(comp));
    }

    
    /**
     * Finds the given pattern in the input using the given functor to determine equivalence.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T>
    findSequence(Iterable<? extends T> iter, T[] pattern, BinaryFunctor<T,T,Boolean> fn) {
        return findSequence(iter.iterator(), Arrays.asList(pattern), fn);
    }
    
    
    /**
     * Finds the given pattern in the input using the given functor to determine equivalence.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T>
    findSequence(Iterable<? extends T> iter, Collection<? extends T> pattern,
                 BinaryFunctor<T,T,Boolean> fn)
    {
        return findSequence(iter.iterator(), pattern, fn);
    }

    
    /**
     * Finds the point at which the input differs from the pattern.  
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the array
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the array but is shorter, the returned iterator will point at the next element in the array
     * after the length of the pattern.
     */
    static public <T> Iterator<T> findMismatch(Iterable<? extends T> iter, T[] pattern) {
        return findMismatch(iter.iterator(), Arrays.asList(pattern), new NotEqualTo<T>());        
    }
    
    /**
     * Finds the point at which the input differs from the pattern.
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the input
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the input but is shorter, the returned iterator will point at the next element in the input
     * after the length of the pattern.
     */
    static public <T> Iterator<T>
    findMismatch(Iterable<? extends T> iter, Collection<? extends T> pattern) {
        return findMismatch(iter.iterator(), pattern, new NotEqualTo<T>());
    }
    
    /**
     * Finds the point at which the input differs from the pattern, using the given comparator to
     * determine the mismatch.  
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the array
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the array but is shorter, the returned iterator will point at the next element in the array
     * after the length of the pattern.
     */
    static public <T> Iterator<T>
    findMismatch(Iterable<? extends T> iter, T[] pattern, Comparator<? super T> comp) {
        return findMismatch(iter.iterator(), Arrays.asList(pattern), new NotEqualTo<T>(comp));        
    }
    
    /**
     * Finds the point at which the input differs from the pattern, using the given comparator to
     * determine the mismatch.  
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the input
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the input but is shorter, the returned iterator will point at the next element in the input
     * after the length of the pattern.
     */
    static public <T> Iterator<T>
    findMismatch(Iterable<? extends T> iter, Collection<? extends T> pattern,
                 Comparator<? super T> comp) {
        return findMismatch(iter.iterator(), pattern, new NotEqualTo<T>(comp));
    }
    
    /**
     * Finds the point at which the input differs from the pattern, using the given functor to
     * determine the mismatch.  The functor should return TRUE if a given pair of elements does
     * not match.
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the array
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the array but is shorter, the returned iterator will point at the next element in the array
     * after the length of the pattern.
     */
    static public <T> Iterator<T>
    findMismatch(Iterable<? extends T> iter, T[] pattern, BinaryFunctor<T,T,Boolean> neq) {
        return findMismatch(iter.iterator(), Arrays.asList(pattern), neq);        
    }
    
    /**
     * Finds the point at which the input differs from the pattern, using the given functor to
     * determine the mismatch.  The functor should return TRUE if a given pair of elements does
     * not match.
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the array
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the array but is shorter, the returned iterator will point at the next element in the array
     * after the length of the pattern.
     */
    static public <T> Iterator<T>
    findMismatch(Iterable<? extends T> iter, Collection<? extends T> pattern,
                 BinaryFunctor<T,T,Boolean> neq) {
        return findMismatch(iter.iterator(), pattern, neq);
    }                                       

    // ============
    // Iterators
    // ============

    /**
     * Finds an arbitrary value in the input using the equals() method.
     * @return an iterator whose next() [if it hasNext()]
     * will return the first instance of the value.  If the value is not in the
     * input, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> find(Iterator<? extends T> iterator, T value) {
        return find(iterator, equalTo(value));
    }

    /**
     * Finds an arbitrary value in the input using the given Comparator.
     * @return an iterator whose next() [if it hasNext()]
     * will return the first instance of the value.  If the value is not in the
     * input, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    find(Iterator<? extends T> iterator, T value, Comparator<? super T> comp) {
        return find(iterator, equalTo(comp, value));
    }

    
    /**
     * Finds an arbitrary value in the input using the given Equality operator.
     * @return an iterator whose next() [if it hasNext()]
     * will return the first instance of the value.  If the value is not in the
     * input, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> find(Iterator<? extends T> iterator, T value, Equality<T> eq) {
        return find(iterator, equalTo(eq, value));
    }

    
    /**
     * Finds a value in the input for which the given function returns TRUE.
     * @return an iterator whose next() [if it hasNext()]
     * will return the first instance of such a value.  If no such value is in the
     * input, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> find(Iterator<? extends T> iterator, UnaryFunctor<T,Boolean> eq) {
        FindIterator<T> finder = finder(iterator);
        finder.findNext(eq);
        return finder;
    }

    
    /**
     * Finds first/next value in the input that is an element of the given array.
     * @return an iterator whose next() [if it hasNext()] will return
     * the first instance of any element found in the array of values.  If no such element is
     * found then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T> findElement(Iterator<? extends T> iterator, T[] values) {
        return find(iterator, elementOf(values));
    }

    
    /**
     * Finds first/next value in the input that is an element of the given collection, using
     * the collection's contains() method.
     * @return an iterator whose next() [if it hasNext()] will return
     * the first instance of any element found in the collection of values.  If no such element
     * is found  then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findElement(Iterator<? extends T> iterator, Collection<? extends T> values) {
        return find(iterator, elementOf(values));
    }

    /**
     * Finds first/next value in the input that is an element of the given array, using
     * the given Comparator to compare values.
     * @return an iterator whose next() [if it hasNext()] will return
     * the first instance of any element found in the array of values.  If no such element is
     * found then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findElement(Iterator<? extends T> iterator, T[] values, Comparator<? super T> comp) {
        return find(iterator, elementOf(values, comp));
    }

    
    /**
     * Finds first/next value in the input that is an element of the given collection, using
     * the given functor to compare values.
     * @return an iterator whose next() [if it hasNext()] will return
     * the first instance of any element found in collection of values.  If no such value is
     * found then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findElement(Iterator<? extends T> iterator, Collection<? extends T> values,
                Comparator<? super T> comp) {
        return find(iterator, elementOf(values, comp));
    }
    
    /**
     * Finds first/next value in the input that is an element of the given array, using
     * the given functor to compare values.
     * @return an iterator whose next() [if it hasNext()] will return
     * the first instance of any element found in the array of values.  If no such element is
     * found then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findElement(Iterator<? extends T> iterator, T[] values, BinaryFunctor<T,T,Boolean> bf) {
        return find(iterator, elementOf(values, bf));
    }

    
    /**
     * Finds first/next value in the input that is an element of the given collection, using
     * the given functor to compare values.
     * @return an iterator whose next() [if it hasNext()] will return
     * the first instance of any element found in collection of values.  If no such value is
     * found then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findElement(Iterator<? extends T> iterator, Collection<? extends T> values,
                BinaryFunctor<T,T,Boolean> bf) {
        return find(iterator, elementOf(values, bf));
    }
    
    /**
     * Locates the first/next pair of adjacent elements in an iteration that
     * are the same value, using the equals() method.
     * @return an iterator whose next() [if it hasNext()] points to the first of
     * a pair of adjacent equivalent values.  If no such pair exists, then the
     * iterator's hasNext() will be false.
     */
    static public <T> Iterator<T> findAdjacent(Iterator<? extends T> iterator) {
        return findAdjacent(iterator, new EqualTo<T>());
    }

    
    /**
     * Locates the first/next pair of adjacent elements in an iteration that
     * are the same value, using the given comparator.
     * @return an iterator whose next() [if it hasNext()] points to the first of
     * a pair of adjacent equivalent values.  If no such pair exists, then the
     * iterator's hasNext() will be false.
     */
    static public <T> Iterator<T>
    findAdjacent(Iterator<? extends T> iterator, Comparator<? super T> comp) {
        return findAdjacent(iterator, new EqualTo<T>(comp));
    }

    
    /**
     * Locates the first/next pair of adjacent elements in an iteration that
     * are the same value, using the given functor.  The functor should return
     * TRUE when two values are to be considered equal.
     * @return an iterator whose next() [if it hasNext()] points to the first of
     * a pair of adjacent equivalent values.  If no such pair exists, then the
     * iterator's hasNext() will be false.
     */
    static public <T> Iterator<T>
    findAdjacent(Iterator<? extends T> iterator, BinaryFunctor<T,T,Boolean> eq) {
        // return early If the input iterator is finished, 
        if (!iterator.hasNext()) {
            return lookAhead(iterator, 1);
        }
        
        LookAheadIterator<T> lai = lookAhead(iterator, 2);
        while (lai.hasNextPlus(2)) {
            T arg1 = lai.peek(1);
            T arg2 = lai.peek(2);
            if (eq.fn(arg1, arg2)) {
                return lai;
            }

            lai.next();
        }

        // didn't find anything, so we advance our working iterator off the end
        // and return it.
        lai.next();
        return lai;
    }

    
    /**
     * Finds first/next arbitrary length run of a given value in an iteration.  Runs of length zero
     * are well-defined: every iteration begins with a run of length zero of all possible values.
     * @return an iterator based on the given iterator whose next() [if it hasNext()] will return
     * the first of n adjacent instances of value.  If no run of values of the requested length
     * exist in the iteration, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findRepeated (Iterator<? extends T> iterator, int count, T value) {
        return findRepeated(iterator, count, equalTo(value));
    }
    
    /**
     * Finds first/next arbitrary length run of a given value in an iteration using the given
     * comparator.  Runs of length zero are well-defined: every iteration begins with
     * a run of length zero of all possible values.
     * @return an iterator based on the given iterator whose next() [if it hasNext()] will return
     * the first of n adjacent instances of value.  If no run of values of the requested length
     * exist in the iteration, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findRepeated (Iterator<? extends T> iterator, int count, T value, Comparator<? super T> comp) {
        return findRepeated(iterator, count, equalTo(comp, value));
    }
    
    /**
     * Finds first/next arbitrary length run of a given value in an iteration using the given
     * equality function.  Runs of length zero are well-defined: every iteration begins with
     * a run of length zero of all possible values.
     * @return an iterator based on the given iterator whose next() [if it hasNext()] will return
     * the first of n adjacent instances of value.  If no run of values of the requested length
     * exist in the iteration, then the returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findRepeated (Iterator<? extends T> iterator, int count, T value, Equality<T> eq) {
        return findRepeated(iterator, count, equalTo(eq, value));
    }
    
    /**
     * Finds first/next arbitrary length run of elements in an iteration for which the
     * given function returns TRUE.  Runs of length zero are well-defined: every
     * iteration begins with a run of length zero of all possible values.
     * @return an iterator based on the given iterator whose next() [if it
     * hasNext()] will return the first of n adjacent instances of value.  If no
     * run of values of the requested length exist in the iteration, then the
     * returned iterator's hasNext() will report false.
     */
    static public <T> Iterator<T>
    findRepeated (Iterator<? extends T> iterator, int count, UnaryFunctor<T,Boolean> eq) {
        if (!iterator.hasNext() || count == 0) {
            return new LookAheadIterator<T>(iterator, 1);
        }

        LookAheadIterator<T> lai = lookAhead(iterator, count);
        
    OUTER:
        while (lai.hasNextPlus(count)) {

            // ... examine the contents of the look ahead buffer ...
            for (int i = 1; i <= count; ++i) {
                T arg = lai.peek(i);
                
                // ... and if we find something in the buffer that isn't
                // 'equal', then we'll advance past that point in the iterator
                // and try again
                if ( ! eq.fn(arg)) {
                    for (int j = i; j > 0; --j) {
                        lai.next();
                    }

                    continue OUTER;
                }
            }

            // If we safely got off the end of the INNER loop, then we must
            // have found the point we're looking for.
            return lai;
        }

        // didn't find anything, make an iterator that will return false.
        return new LookAheadIterator<T>(new EmptyIterator<T>(), 1);
    }

    /**
     * Finds the given pattern in the input using the equals method.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T> findSequence(Iterator<? extends T> iterator, T[] pattern) {
        return findSequence(iterator, Arrays.asList(pattern), new EqualTo<T>());
    }

    
    /**
     * Finds the given pattern in the input using the equals method.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T>
    findSequence(Iterator<? extends T> iterator, Collection<? extends T> pattern) {
        return findSequence(iterator, pattern, new EqualTo<T>());
    }
    
    
    /**
     * Finds the given pattern in the input using the given comparator to determine equivalence.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T>
    findSequence(Iterator<? extends T> iterator, T[] pattern, Comparator<? super T> comp) {
        return findSequence(iterator, Arrays.asList(pattern), new EqualTo<T>(comp));
    }
    
    
    /**
     * Finds the given pattern in the collection using the given comparator to determine equivalence.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T>
    findSequence(Iterator<? extends T> iterator, Collection<? extends T> pattern,
                 Comparator<T> comp) {
        return findSequence(iterator, pattern, new EqualTo<T>(comp));
    }

    
    /**
     * Finds the given pattern in the collection using the given functor to determine equivalence.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T>
    findSequence(Iterator<? extends T> iterator, T[] pattern, BinaryFunctor<T,T,Boolean> fn) {
        return findSequence(iterator, Arrays.asList(pattern), fn);
    }
    
    
    /**
     * Finds the given pattern in the collection using the given functor to determine equivalence.
     * @return an iterator whose next() [if it hasNext()] will return the first element of a
     * sequence that matches the pattern.  If no such match is found, then the returned iterator's
     * hasNext() will report false.  If the pattern is empty, then the iterator will not be advanced.
     */
    static public <T> Iterator<T>
    findSequence(Iterator<? extends T> iterator, Collection<? extends T> pattern,
                 BinaryFunctor<T,T,Boolean> fn) {
        // return early if the input iterator is already finished, 
        int length = pattern.size();
        if (!iterator.hasNext() || length == 0) {
            return lookAhead(iterator, 1);
        }
        
        LookAheadIterator<T> lai = lookAhead(iterator, length);
        
        // So long as the LookAhead has enough room for the repeat count to
        // possibly fit, ...
        
    OUTER:
        while (lai.hasNextPlus(length)) {
            int idx = 1;
            
            // ... examine the contents of the look ahead buffer ...
            for (T obj : pattern) {
                
                // ... and if we find something in the buffer that isn't
                // 'equal', then advance the look ahead
                if (!fn.fn(obj, lai.peek(idx))) {
                    lai.next();
                    continue OUTER;
                }

                ++idx;
            }
            
            // If we safely got off the end of the INNER loop, then we must
            // have found the point we're looking for.
            return lai;
        }

        // didn't find anything, make an iterator that will return false.
        return new LookAheadIterator<T>(new EmptyIterator<T>(), 1);
    }

    
    /**
     * Finds the point at which the input differs from the pattern.  
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the array
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the array but is shorter, the returned iterator will point at the next element in the array
     * after the length of the pattern.  If the pattern is empty, no input will  be consumed.
     */
    static public <T> Iterator<T> findMismatch(Iterator<? extends T> iterator, T[] pattern) {
        return findMismatch(iterator, Arrays.asList(pattern), new NotEqualTo<T>());        
    }
    
    /**
     * Finds the point at which the input differs from the pattern.
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the input
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the input but is shorter, the returned iterator will point at the next element in the input
     * after the length of the pattern.  If the pattern is empty, no input will be consumed.
     */
    static public <T> Iterator<T>
    findMismatch(Iterator<? extends T> iterator, Collection<? extends T> pattern) {
        return findMismatch(iterator, pattern, new NotEqualTo<T>());
    }
    
    /**
     * Finds the point at which the input differs from the pattern, using the given comparator to
     * determine the mismatch.  
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the array
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the array but is shorter, the returned iterator will point at the next element in the array
     * after the length of the pattern.  If the pattern is empty, no input will  be consumed.
     */
    static public <T> Iterator<T>
    findMismatch(Iterator<? extends T> iterator, T[] pattern, Comparator<? super T> comp) {
        return findMismatch(iterator, Arrays.asList(pattern), new NotEqualTo<T>(comp));        
    }
    
    /**
     * Finds the point at which the input differs from the pattern, using the given comparator to
     * determine the mismatch.  
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the input
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the input but is shorter, the returned iterator will point at the next element in the input
     * after the length of the pattern.  If the pattern is empty, no input will be consumed.
     */
    static public <T> Iterator<T>
    findMismatch(Iterator<? extends T> iterator, Collection<? extends T> pattern,
                 Comparator<? super T> comp) {
        return findMismatch(iterator, pattern, new NotEqualTo<T>(comp));
    }
    
    /**
     * Finds the point at which the input differs from the pattern, using the given functor to
     * determine the mismatch.  The functor should return TRUE if a given pair of elements does
     * not match.
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the array
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the array but is shorter, the returned iterator will point at the next element in the array
     * after the length of the pattern.  If the pattern is empty, no input will  be consumed.
     */
    static public <T> Iterator<T>
    findMismatch(Iterator<? extends T> iterator, T[] pattern, BinaryFunctor<T,T,Boolean> neq) {
        return findMismatch(iterator, Arrays.asList(pattern), neq);        
    }
    
    /**
     * Finds the point at which the input differs from the pattern, using the given functor to
     * determine the mismatch.  The functor should return TRUE if a given pair of elements does
     * not match.
     * @return an iterator based whose next() [if it hasNext()] will return the first element which
     * does not match the corresponding element in the pattern.  If the pattern matches the array
     * but is longer, the returned iterator's hasNext() will report false.  If the pattern matches
     * the array but is shorter, the returned iterator will point at the next element in the array
     * after the length of the pattern.  If the pattern is empty, no input will  be consumed.
     */
    static public <T> Iterator<T>
    findMismatch(Iterator<? extends T> iterator, Collection<? extends T> pattern,
                 BinaryFunctor<T,T,Boolean> neq) {
        LookAheadIterator<? extends T> patternIter = new LookAheadIterator<T>(pattern.iterator());
        LookAheadIterator<T> lai = lookAhead(iterator, 1);
        while (lai.hasNextPlus(1) && patternIter.hasNextPlus(1)) {
            T arg1 = lai.peek(1);
            T arg2 = patternIter.peek(1);
            if (neq.fn(arg1,arg2)) {
                return lai;
            }
            
            lai.next();
            patternIter.next();
        }

        return lai;
    }                                       

    // ==============
    // Implementation
    // ==============

    static public <T> UnaryFunctor<T,Boolean> elementOf(T[] values) {
        return new ElementOf<T>().bind2nd(Arrays.asList(values));
    }
    
    static public <T> UnaryFunctor<T,Boolean> elementOf(Collection<? extends T> values) {
        return new ElementOf<T>().bind2nd(values);
    }
    
    static public <T> UnaryFunctor<T,Boolean> elementOf(T[] values, Comparator<? super T> comp) {
        return new ElementOf<T>(new EqualTo<T>(comp)).bind2nd(Arrays.asList(values));
    }
    
    static public <T> UnaryFunctor<T,Boolean>
    elementOf(Collection<? extends T> values, Comparator<? super T> comp) {
        return new ElementOf<T>(new EqualTo<T>(comp)).bind2nd(values);
    }
    
    static public <T> UnaryFunctor<T,Boolean> elementOf(T[] values, BinaryFunctor<T,T,Boolean> bf) {
        return new ElementOf<T>(bf).bind2nd(Arrays.asList(values));
    }
    
    static public <T> UnaryFunctor<T,Boolean>
    elementOf(Collection<? extends T> values, BinaryFunctor<T,T,Boolean> bf) {
        return new ElementOf<T>(bf).bind2nd(values);
    }
    
    @SuppressWarnings("unchecked")
    // The cast from Iterator<? extends T> to FindIterator<T>
    static private <T> FindIterator<T> finder(Iterator<? extends T> iterator) {
        if (iterator instanceof FindIterator)
            return (FindIterator<T>) iterator;

        return new FindIterator<T>(iterator);
    }

    @SuppressWarnings("unchecked")
    // The cast from Iterator<? extends T> to LookAheadIterator<T>
    static private <T> LookAheadIterator<T> lookAhead(Iterator<? extends T> iterator, int size) {
        if (iterator instanceof LookAheadIterator) {
            LookAheadIterator<T> lai = (LookAheadIterator<T>)(Iterator<T>) iterator;
            if (lai.getMaxPeekSize() >= size)
                return lai;
        }
        
        return new LookAheadIterator<T>(iterator, size);
    }
}
