// ============================================================================
// $Id: ListAlgorithms.java,v 1.4 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html 
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================
package net.sf.jga.algorithms;

import java.util.List;
import java.util.ListIterator;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.fn.adaptor.ConstantUnary;
import net.sf.jga.fn.comparison.EqualTo;
import net.sf.jga.fn.comparison.Equality;

// import net.sf.jga.fn.logical.UnaryNegate;

import static net.sf.jga.fn.comparison.ComparisonFunctors.*;
import static net.sf.jga.fn.logical.LogicalFunctors.*;

/**
 * Algorithms that modify a given list.  Unlike the static algorithm classes
 * within this package, this class will change the contents of its input.
 * <p>
 * Copyright &copy; 2007  David A. Hall
 */

public class ListAlgorithms {
    /**
     * Removes all elements from the list that do not meet the given selection criteria.
     * @return the modified list
     */
    static public <T, LT extends List<? extends T>> LT filter(LT list, UnaryFunctor<T,Boolean> uf) {
        ListIterator<? extends T> liter = list.listIterator();
        while (liter.hasNext()) {
            if (!uf.fn(liter.next()).booleanValue()) {
                liter.remove();
            }
        }

        return list;
    }

    
    /**
     * removes all elements equal to the given value from the list, 
     * @return the modified list
     */
    static public <T, LT extends List<? extends T>> LT remove(LT list, T value) {
        return filter(list, notEqualTo(value));
    }

    
    /**
     * removes elements equal to the given value from the list, using the given functor
     * to compare for equality.
     * @return the modified list
     */
    static public <T, LT extends List<? extends T>> LT remove(LT list, T value, Equality<T> eq) {
        return filter(list, unaryNegate(eq.bind2nd(value)));
    }

    
    /**
     * removes elements from the list that do meet the given selection criteria
     * @return the modified list
     */
    static public <T, LT extends List<? extends T>> LT remove(LT list, UnaryFunctor<T,Boolean> pred){
        return filter(list, unaryNegate(pred));
    }


    /**
     * Applies the UnaryFunctor to each element in the list, replacing each element with the result.
     * @return the modified list
     */
    static public <T, LT extends List<T>> LT transform(LT list, UnaryFunctor<T,T> uf) {
        ListIterator<T> liter = list.listIterator();
        while(liter.hasNext()) {
            liter.set(uf.fn(liter.next()));
        }

        return list;
    }

    
    /**
     * Replaces elements equal to the given value with the given replacement value.  
     * @return the modified list
     */
    static public <T, LT extends List<T>> LT replace(LT list, T value, T replacement) {
        return replace(list, equalTo(value), new ConstantUnary<T,T>(replacement));
    }


    /**
     * Tests each element in the list, and replaces elements that pass with the given value.  
     * @return the modified list
     */
    static public <T, LT extends List<T>> LT
    replace(LT list, UnaryFunctor<T,Boolean> test, T replacement)
    {
        return replace(list, test, new ConstantUnary<T,T>(replacement));
    }


    /**
     * Tests each element in the list, and replaces elements that pass with the result of passing
     * the element to the given replacement function.  
     * @return the modified list
     */
    static public <T, LT extends List<T>> LT
    replace(LT list, UnaryFunctor<T,Boolean> test, UnaryFunctor<T,T> fn)
    {
        ListIterator<T> liter = list.listIterator();
        while (liter.hasNext()) {
            T value = liter.next();
            if (test.fn(value).booleanValue()) {
                liter.set(fn.fn(value));
            }
        }

        return list;
    }

    /**
     * Reverses the order of the list.
     * @return the modified list
     */

    static public <T, LT extends List<T>> LT reverse(LT list) {
        ListIterator<T> front = list.listIterator();
        ListIterator<T> back  = list.listIterator(list.size());
        int i = 0, mid = (list.size() + 1) / 2;
        while(i++ < mid) {
            T t = front.next();
            front.set(back.previous());
            back.set(t);            
        }

        return list;
    }


    /**
     * Removes all adjacent duplicate values in the given list, using equals() to compare
     * adjacent elements.
     * @return the modified list
     */
    static public <T, LT extends List<? extends T>> LT unique(LT list) {
        return unique(list, new EqualTo<T>());
    }

    
    /**
     * Removes all adjacent duplicate values in the given list, using the given functor to compare
     * adjacent elements
     * @return the modified list
     * @deprecated use {@link Unique#unique(Iterable,BinaryFunctor)}
     */
    static public <T, LT extends List<? extends T>> LT unique(LT list,BinaryFunctor<T,T,Boolean> eq){
        ListIterator<? extends T> liter = list.listIterator();

        T last = null;
        // skip the first element, if there is one
        if (liter.hasNext())
            last = liter.next();

        while (liter.hasNext()) {
            T next = liter.next();
            if (eq.fn(last, next).booleanValue()) {
                liter.remove();
            }
            else {
                last = next;
            }
        }

        return list;
    }
}
