// ============================================================================
// $Id: Merge.java,v 1.9 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2006  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.algorithms;

import static net.sf.jga.util.ArrayUtils.iterable;

import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.NoSuchElementException;

import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.comparison.LessEqual;
import net.sf.jga.util.CollectionUtils;
import net.sf.jga.util.ComparableComparator;
import net.sf.jga.util.LookAheadIterator;

/**
 * Algorithms that combine the contents of two inputs.  The inputs may arrays, collections, or
 * iterations.  The append methods that appear here are historical -- the original implementation
 * of append was a Merge where the functor was constant TRUE (always choosing the contents of the
 * first input until it is exhausted).  Now, the append methods delegate to the Append class, but 
 * due to backwards compatability (and a desire to shield the user from "unchecked" warnings caused
 * by the use of a varargs array of generic types in the Append signatures), these methods will
 * remain to support the most common use-case.
 * <p>
 * Copyright &copy; 2006  David A. Hall
 */

public class Merge {

    // ============
    // Arrays
    // ============
    
    /**
     * Returns all elements of each array
     */
    @SuppressWarnings("unchecked") // generated by the varargs call to Append
    static public <T> Iterable<T> append(T[] t0, T[] t1) {
        return Append.append(t0, t1);
    }

 
    /**
     * 	Returns all elements of each array, always choosing the least of the current elements.
     */
    static public <T extends Comparable<? super T>> Iterable<T> merge(T[] t0, T[] t1) {
        return merge(t0, t1, new LessEqual.Comparable<T>());
    }

 
    /**
     * Returns all elements of both arrays, always choosing the lesser of the two current elements
     * using the given comparator.
     */
    static public <T> Iterable<T> merge(T[] ts1, T[] ts2, Comparator<T> comp) {
        return new MergeIterable<T>(iterable(ts1), iterable(ts2), comp);
    }

    
    /**
     * Returns all elements of both arrays, always choosing the lesser of the two current elements
     * by using the given predicate.  When the predicate returns TRUE, an element is taken from the
     * first array: when FALSE, an element is taken from the second array.
     */
    static public <T> Iterable<T> merge(T[] ts1, T[] ts2, BinaryFunctor<T,T,Boolean> pred){
        return new MergeIterable<T>(iterable(ts1), iterable(ts2), pred); 
    }

    // ============
    // Iterables
    // ============
    
    /**
     * Returns an Iterable implementation that 'contains' all elements of first iterable followed by
     * all elements of the second iterable
     */
    @SuppressWarnings("unchecked")
    static public <T> Iterable<? extends T>
    append(Iterable<? extends T> input0, Iterable<? extends T> input1) {
        return Append.append(input0, input1);
    }


    /**
     * Returns an Iterable implementation that 'contains' all elements of both iterables, always
     * choosing the lesser of the two current elements.
     */
    static public <T extends Comparable<? super T>> Iterable<? extends T>
    merge(Iterable<? extends T> input0, Iterable<? extends T> input1) {
        return merge(input0, input1, new LessEqual.Comparable<T>());
    }


    /**
     * Returns all elements of both iterables, always choosing the lesser of the two current
     * elements.
     */
    static public <T> Iterable<T>
    merge(Iterable<? extends T> i1, Iterable<? extends T> i2, Comparator<T> comp) {
        return new MergeIterable<T>(i1, i2, comp);
    }

    /**
     * Returns all elements of both iterables, always choosing the lesser of the two current
     * elements by using the given predicate.  When the predicate returns TRUE, an element is
     * taken from the first array: when FALSE, an element is taken from the second array.
     */
    static public <T> Iterable<T>
    merge(Iterable<? extends T> i1, Iterable<? extends T> i2, BinaryFunctor<T,T,Boolean> pred) {
        return new MergeIterable<T>(i1, i2, pred);
    }
    
    // ============
    // Iterators
    // ============
    
     /**
      * Returns an Iterator that will return elements from the first input iterator followed by
      * elements from the second input iterator.
      */
     @SuppressWarnings("unchecked") // generated by the varargs call to Append
     static public <T> Iterator<? extends T> append(Iterator<? extends T> iter1, Iterator<? extends T> iter2) {
         return Append.append(iter1, iter2);
     }

    
     /**
      * Returns an Iterator that will return elements from both input iterators, choosing the
      * lesser of the two current values.  Neither input iterator is advanced in this method.
      */
     static public <T extends Comparable<? super T>> Iterator<T>
     merge(Iterator<? extends T> iter1, Iterator<? extends T> iter2) {
         return new MergeIterator<T>(iter1, iter2, new ComparableComparator<T>());
     }

    
    /**
     * Returns an Iterator that will return elements from both input iterators, choosing the
     * lesser of the two current values using the given comparator.  Neither input iterator
     * is advanced in this method.
     */
    static public <T> Iterator<T>
    merge(Iterator<? extends T> iter1, Iterator<? extends T> iter2, Comparator<T> comp) {
        return new MergeIterator<T>(iter1, iter2, comp);
    }

    
    /**
     * Returns an Iterator that will return elements from both input iterators, choosing the
     * lesser of the two current values by using the given predicate.  When the predicate
     * returns TRUE, an element is taken from the first array: when FALSE, an element is taken
     * from the second iterator.  Neither input iterator is advanced in this method.
     */
    static public <T> Iterator<T>
    merge(Iterator<? extends T> iter1, Iterator<? extends T> iter2, BinaryFunctor<T,T,Boolean> pred){
        return new MergeIterator<T>(iter1, iter2, pred);
    }

    // ============
    // IterableCopy
    // ============
    
    /**
     * Appends the values of the first and second collection to the output collection
     * @return the output collection
     */
    @SuppressWarnings("unchecked") // generated by the varargs call to Append
    static public <T, TCollection extends Collection<? super T>> TCollection
    append(Iterable<? extends T> cin1, Iterable<? extends T> cin2, TCollection cout) {
        return CollectionUtils.append(cout, Append.append(cin1, cin2));
    }

    
    /**
     * Merges two collections, appending values to the output collection.
     * @return the output collection
     */
    static public <T extends Comparable<? super T>, TCollection extends Collection<? super T>> TCollection
    merge(Iterable<? extends T> cin1, Iterable<? extends T> cin2, TCollection cout) {
        return mergeAppend(cout, new LessEqual.Comparable<T>(), cin1, cin2);
    }

    
    /**
     * Merges two collections using the given comparator, appending values to the output collection.
     * @return the output collection
     */
    static public <T, TCollection extends Collection<? super T>> TCollection
    merge(Iterable<? extends T> cin1, Iterable<? extends T> cin2, Comparator<T> comp,
          TCollection cout)
    {
        return mergeAppend(cout, new LessEqual<T>(comp), cin1, cin2);
    }

    /**
     * Merges two collections using the given comparator, appending values to the output collection.
     * @return the output collection
     */

    static public <T, TCollection extends Collection<? super T>> TCollection
    merge(Iterable<? extends T> cin1, Iterable<? extends T> cin2, BinaryFunctor<T,T,Boolean> pred,
          TCollection cout)
    {
        return mergeAppend(cout, pred, cin1, cin2);
    }

   /**
    * Merges two collections using the given comparator, appending values to the output collection.
    * @return the output collection
    */
    static public <T, TCollection extends Collection<? super T>> TCollection
    mergeAppend(TCollection cout, BinaryFunctor<T,T,Boolean> pred, Iterable<? extends T> cin0, Iterable<? extends T> cin1)
    {
        return CollectionUtils.append(cout, merge(cin0, cin1, pred));
    }
    
    /**
     * Produces iterators that return all of the merged contents of two inputs
     */
    static public class MergeIterable<T> implements Iterable<T> {

        // The contents
        private Iterable<? extends T> _delegate1;
        private Iterable<? extends T> _delegate2;
    
        // The test
        private BinaryFunctor<T,T,Boolean> _pred;

        /**
         * Builds a Merge for the given base iterators that uses the given Comparator to compare
         * corresponding elements.  The Comparator will be used with a LessEqualComp predicate.
         * @throws IllegalArgumentException if either argument is null
         */
      
        public MergeIterable(Iterable<? extends T> base1, Iterable<? extends T> base2,
                             Comparator<T> comp)
        {
            this(base1, base2, new LessEqual<T>(comp));
        }

        /**
         * Builds a Merge for the given base iterators that uses the given predicate to compare
         * corresponding elements.  The predicate should return TRUE if its first argument is less
         * than or equal to the second.
         * @throws IllegalArgumentException if either base iterable is null
         */
    
        public MergeIterable(Iterable<? extends T> base1, Iterable<? extends T> base2,
                             BinaryFunctor<T,T,Boolean> pred)
        {
            if (base1 == null || base2 == null) {
                String msg = "two base iterables are required";
                throw new IllegalArgumentException(msg);
            }
        
            if (pred == null) {
                String msg = "functor is required";
                throw new IllegalArgumentException(msg);
            }
        
            _delegate1 = base1;
            _delegate2 = base2;
            _pred = pred;
        }

        // - - - - - - - - - - -
        // Iterable<T> interface
        // - - - - - - - - - - -

        public Iterator<T> iterator() {
            return new MergeIterator<T>(_delegate1.iterator(), _delegate2.iterator(), _pred);
        }
    }
    
    /**
     * Iterator that merges the contents of two input iterators.
     */
    static public class MergeIterator<T> implements Iterator<T> {
        // The base iterators
        private LookAheadIterator<? extends T> _base1;
        private LookAheadIterator<? extends T> _base2;

        // The test
        private BinaryFunctor<T,T,Boolean> _pred;

        /**
         * Builds a Merge.Iterator for the given base iterators that uses the given
         * Comparator to compare corresponding elements.  The Comparator will be
         * used with a LessEqualComp predicate.
         * @throws IllegalArgumentException if either argument is null
         */

        public MergeIterator(Iterator<? extends T> base1, Iterator<? extends T> base2,
                             Comparator<T> comp)
        {
            this(base1, base2, new LessEqual<T>(comp));
        }
        
        /**
         * Builds a Merge.Iterator for the given base iterators that uses the given
         * predicate to compare corresponding elements.  The predicate should return
         * TRUE if its first argument is less than or equal to the second.
         * @throws IllegalArgumentException if either argument is null
         */
        
        public MergeIterator(Iterator<? extends T> base1, Iterator<? extends T> base2,
                             BinaryFunctor<T,T,Boolean> pred)
        {
            if (base1 == null || base2 == null) {
                String msg = "two base iterators are required";
                throw new IllegalArgumentException(msg);
            }
            if (pred == null) {
                String msg = "functor is required";
                throw new IllegalArgumentException(msg);
            }

            _base1 = new LookAheadIterator<T>(base1, 1);
            _base2 = new LookAheadIterator<T>(base2, 1);
            _pred = pred;
        }

    
        // - - - - - - - - - - -
        // Iterator<T> interface
        // - - - - - - - - - - -
    
        public boolean hasNext() {
            return _base1.hasNextPlus(1) || _base2.hasNextPlus(1);
        }

        public T next() {
            if (_base1.hasNextPlus(1)) 
                if (_base2.hasNextPlus(1))
                    if (_pred.fn(_base1.peek(1), _base2.peek(1)))
                        return _base1.next();           
                    else
                        return _base2.next();
                else
                    return _base1.next();
            else
                if (_base2.hasNextPlus(1)) 
                    return _base2.next();
                else
                    throw new NoSuchElementException();
        }

        public void remove() { throw new UnsupportedOperationException(); }
    }
}
