// ============================================================================
// $Id: Reverse.java,v 1.2 2008/02/08 04:22:34 davidahall Exp $
// Copyright (c) 2007  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.algorithms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

import static net.sf.jga.util.ArrayUtils.*;
import static net.sf.jga.util.CollectionUtils.*;

/**
 * Reverses the order of the input
 * <p>
 * Copyright &copy; 2007  David A. Hall
 */

public class Reverse {

    // ============
    // Arrays
    // ============
    
    /**
     * Returns an iterable object over the reversed contents of the array.
     * The array will not be modified.
     */
    static public <T> Iterable<T> reverse(T[] ts) {
        return new ReverseArrayIterable<T>(ts);
    }

    
    // ============
    // Iterables
    // ============
    
    /**
     * Returns an iterable object over the reversed contents of the input.  The input itself
     * will not be modified.
     */
    static public <T> Iterable<T> reverse(Iterable<T> i) {
        return new ReverseIterable<T>(i);
    }

    
    // ============
    // Iterators
    // ============
    
    /**
     * Returns an iterator object over the reversed contents of the input.  The input itself
     * will not be modified.
     */
    static public <T> Iterator<T> reverse(Iterator<T> i) {
        return new ReverseIterable<T>(i).iterator();
    }


    // ============
    // IterableCopy
    // ============
    
    /**
     * Appends the reversed contents of the input to the output.  The input will not be modified.
     */
    static public <T, TCollection extends Collection<? super T>>
    TCollection reverse(Iterable<? extends T> i, TCollection co)
    {
        return append(co, reverse(i));
    }

    
    // ==================
    // ReverseIterable(s)
    // ==================

    static public class ReverseArrayIterable<T> extends ArrayIterable<T> {
        public ReverseArrayIterable(T[] ts) {
            super(ts);
        }

        protected Iterator<T> buildIterator(T[] delegate) {
            return new ReverseIterator<T>(new ArrayIterator<T>(delegate, delegate.length));
        }
    }
    
    static public class ReverseIterable<T> implements Iterable<T> {
        List<T> _list;

        public ReverseIterable(Iterable<T> ts) {
            _list = append(new ArrayList<T>(), ts);
        }

        public ReverseIterable(List<T> ts) {
            _list = ts;
        }

        public ReverseIterable(Iterator<T> ts) {
            _list = append(new ArrayList<T>(), ts);
        }

        public Iterator<T> iterator() {
            return new ReverseIterator<T>(_list.listIterator(_list.size()));
        }
    }

    
    // ===============
    // ReverseIterator
    // ===============

    static public class ReverseIterator<T> implements ListIterator<T> {
        ListIterator<T> _delegate;

        public ReverseIterator(ListIterator<T> iter) {
            _delegate = iter;
        }

        public boolean hasNext() { return _delegate.hasPrevious(); }
        public boolean hasPrevious() { return _delegate.hasNext(); }

        public T next() { return _delegate.previous(); }
        public T previous() { return _delegate.next(); }

        public int nextIndex() { return _delegate.previousIndex(); }
        public int previousIndex() { return _delegate.nextIndex(); }
        
        public void remove() { _delegate.remove(); }
        public void set (T arg) { _delegate.set(arg); }
        public void add (T arg) { _delegate.add(arg); }
    }
}
