// ============================================================================
// $Id: Sort.java,v 1.6 2006/12/16 16:48:58 davidahall Exp $
// Copyright (c) 2006  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.algorithms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;

import static net.sf.jga.util.ArrayUtils.*;
import static net.sf.jga.util.CollectionUtils.*;

/**
 * Adapts standard java Sort capabilities to the interfaces common to jga.
 * <p>
 * Copyright &copy; 2006  David A. Hall
 */

public class Sort {

    // ============
    // Arrays
    // ============
    
    /**
     * Returns an iterable object over the sorted contents of the array.
     * The array will not be modified.
     */
    static public <T extends Comparable<? super T>> Iterable<T> sort(T[] ts) {
        return sort(iterable(ts));
    }

    /**
     * Returns an iterable object over the sorted contents of the array, using the given comparator
     * to determine ordering.  The array itself will not be modified.  .
     */
    static public <T> Iterable<T> sort(T[] ts, Comparator<? super T> comp) {
        // NOTE: it might be faster to clone the array and use ArrayUtils.sort, but we'll need to
        // check that cloning the array can be done in a typesafe manner.
        return sort(iterable(ts), comp);
    }
    
    // ============
    // Iterables
    // ============
    
    /**
     * Returns an iterable object over the sorted contents of the input.  The input itself
     * will not be modified.
     */
    static public <T extends Comparable<? super T>> Iterable<T> sort(Iterable<T> i) {
        ArrayList<T> list = append(new ArrayList<T>(), i.iterator());
        Collections.sort(list);
        return list;
    }

    
    /**
     * Returns an iterable object over the sorted contents of the input, using the given comparator
     * to determine ordering.  The input will not be modified.
     */
    static public <T> Iterable<T> sort(Iterable<T> i, Comparator<? super T> comp) {
        ArrayList<T> list = append(new ArrayList<T>(), i.iterator());
        Collections.sort(list, comp);
        return list;
    }

    // ============
    // Iterators
    // ============
    
    /**
     * Returns an iterator object over the sorted contents of the input.  The input itself
     * will not be modified.
     */
    static public <T extends Comparable<? super T>> Iterator<T> sort(Iterator<T> i) {
        ArrayList<T> list = append(new ArrayList<T>(), i);
        Collections.sort(list);
        return list.iterator();
    }


    /**
     * Returns an iterator object over the sorted contents of the input, using the given comparator
     * to determine ordering.  The input will not be modified.
     */
    static public <T> Iterator<T> sort(Iterator<T> i, Comparator<? super T> comp) {
        ArrayList<T> list = append(new ArrayList<T>(), i);
        Collections.sort(list, comp);
        return list.iterator();
    }
    
    // ============
    // IterableCopy
    // ============
    
    /**
     * Appends the sorted contents of the input to the output.  The input will not be modified.
     */
    static public <T extends Comparable<? super T>, TCollection extends Collection<? super T>>
    TCollection sort(Iterable<? extends T> i, TCollection co)
    {
        ArrayList<T> list = append(new ArrayList<T>(), i.iterator());
        Collections.sort(list);
        co.addAll(list);
        return co;
    }

    
    /**
     * Appends the sorted contents of the input to the output, using the given comparator
     * to determine ordering.  The input will not be modified.
     */
    static public <T, TCollection extends Collection<? super T>> TCollection
    sort(Iterable<T> i, Comparator<? super T> comp, TCollection co) {
        ArrayList<T> list = append(new ArrayList<T>(), i.iterator());
        Collections.sort(list, comp);
        co.addAll(list);
        return co;
    }

}
