// ============================================================================
// $Id: Summarize.java,v 1.8 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2006  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.algorithms;

import java.text.MessageFormat;
import java.util.Comparator;
import java.util.Iterator;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.fn.arithmetic.Arithmetic;
import net.sf.jga.fn.arithmetic.ArithmeticFactory;
import net.sf.jga.fn.arithmetic.Average;
import net.sf.jga.fn.arithmetic.Plus;
import net.sf.jga.fn.comparison.Equality;
import net.sf.jga.fn.comparison.Max;
import net.sf.jga.fn.comparison.Min;

import static net.sf.jga.fn.comparison.ComparisonFunctors.*;
import static net.sf.jga.util.ArrayUtils.*;

/**
 * Algorithms that consume input and produce a single value.  The input may be
 * an array, collection, or iteration.
 * <p>
 * Copyright &copy; 2006  David A. Hall
 */

public class Summarize {

    // -------------------
    // counting algorithms
    // -------------------
    
    /**
     * Returns the number of times that the given value appears in the array
     */
    static public <T> long count(T[] ts, T value) {
        return count(iterate(ts), equalTo(value));
    }


    /**
     * Returns the number of times that the given value appears in the array, using
     * the given Comparator
     */
    static public <T> long count(T[] ts, T value, Comparator<? super T> comp) {
        return count(iterate(ts), equalTo(comp, value));
    }

        
    /**
     * Returns the number of times that the given value appears in the array, using
     * the given equality operator
     * @deprecated - switching the order of arguments to be consistent with Find
     */
    static public <T> long count(T[] ts, Equality<T> eq, T value) {
        return count(ts, value, eq);
    }


    /**
     * Returns the number of times that the given value appears in the array, using
     * the given equality operator
     */
    static public <T> long count(T[] ts, T value, Equality<T> eq) {
        return count(iterate(ts), equalTo(eq, value));
    }

        
    /**
     * Returns the number of elements in the array for which the predicate is true
     */
    static public <T> long count(T[] ts, UnaryFunctor<T,Boolean> pred) {
        return count(iterate(ts), pred);
    }

    
    /**
     * Returns the number of elements in the input.
     */
    static public <T> long count(Iterable<? extends T> input) {
        return count(input.iterator());
    }

    
    /**
     * Returns the number of times that the given value appears in the input.
     */
    static public <T> long count(Iterable<? extends T> input, T value) {
        return count(input.iterator(), value);
    }
    

    /**
     * Returns the number of times that the given value appears in the input, using
     * the given Comparator.
     */
    static public <T> long count(Iterable<? extends T> input, T value, Comparator<? super T> comp) {
        return count(input.iterator(), equalTo(comp, value));
    }

    
    /**
     * Returns the number of times that the given value appears in the input, using
     * the given equality operator.
     * @deprecated - switching the order of arguments to be consistent with Find
     */
    static public <T> long count(Iterable<? extends T> input, Equality<T> eq, T value) {
        return count(input, value, eq);
    }

    
    /**
     * Returns the number of times that the given value appears in the input, using
     * the given equality operator.
     */
    static public <T> long count(Iterable<? extends T> input, T value, Equality<T> eq) {
        return count(input.iterator(), equalTo(eq, value));
    }

    
    /**
     * Returns the number of elements in the input for which the predicate is true.
     */
    static public <T> long count(Iterable<? extends T> input, UnaryFunctor<T,Boolean> pred) {
        return count(input.iterator(), pred);
    }

    
    /**
     * Returns the number of elements in the iterator.  The iterator is consumed entirely
     * by this function.
     */
    static public <T> long count(Iterator<? extends T> iter) {
        long ct = 0;
        while(iter.hasNext()) {
            iter.next();
            ++ct;
        }
        
        return ct;
    }

    
    /**
     * Returns the number of times that the given value appears in the iterator.  The
     * iterator is consumed entirely by this function.
     */
    static public <T> long count(Iterator<? extends T> iter, T value) {
        return count(Filter.filter(iter, equalTo(value)));
    }

    
    /**
     * Returns the number of times that the given value appears in the iterator, using
     * the given equality operator.  The iterator is consumed entirely by this function.
     */
    static public <T> long count(Iterator<? extends T> iter, T value, Comparator<? super T> comp) {
        return count(Filter.filter(iter, equalTo(comp, value)));
    }

    
    /**
     * Returns the number of times that the given value appears in the iterator, using
     * the given equality operator.  The iterator is consumed entirely by this function.
     * @deprecated - switching the order of arguments to be consistent with Find
     */
    static public <T> long count(Iterator<? extends T> iter, Equality<T> eq, T value) {
        return count(iter, value, eq);
    }

    
    /**
     * Returns the number of times that the given value appears in the iterator, using
     * the given equality operator.  The iterator is consumed entirely by this function.
     */
    static public <T> long count(Iterator<? extends T> iter, T value, Equality<T> eq) {
        return count(Filter.filter(iter, equalTo(eq, value)));
    }

    /**
     * Returns the number of elements in the iterator for which the predicate is true.  The
     * iterator is consumed entirely by this function.
     */
    static public <T> long count(Iterator<? extends T> iter, UnaryFunctor<T,Boolean> pred) {
        return count(Filter.filter(iter, pred));
    }

    // -------------------
    // minimum value
    // -------------------
    
    /**
     * Returns the smallest T value in the input, as determined by the given comparator.
     */
    static public <T> T min(T[] ts, Comparator<? super T> comp) {
        return min(iterate(ts), new Min<T>(comp));
    }

    /**
     * Returns the smallest T value in the input.
     */
    static public <T extends Comparable<? super T>> T min(T[] ts) {
        return min(iterate(ts), new Min.Comparable<T>());
    }


    /**
     * Returns the smallest T value in the input, as determined by the given functor.
     * The functor is expected to return the smaller of its two arguments.
     */
    static public <T> T min(T[] ts, BinaryFunctor<T,T,T> fn) {
        return min(iterate(ts), fn);
    }


    /**
     * Returns the smallest T value in the input, as determined by the given comparator.
     */
    static public <T> T min(Iterable<? extends T> ts, Comparator<? super T> comp) {
        return min(ts.iterator(), new Min<T>(comp));
    }


    /**
     * Returns the smallest T value in the input.
     */
    static public <T extends Comparable<? super T>> T min(Iterable<? extends T> ts) {
        return min(ts.iterator(), new Min.Comparable<T>());
    }


    /**
     * Returns the smallest T value in the input, as determined by the given functor.
     * The functor is expected to return the smaller of its two arguments.
     */
    static public <T> T min(Iterable<? extends T> ts, BinaryFunctor<T,T,T> fn) {
        return min(ts.iterator(), fn);
    }


    /**
     * Returns the smallest T value in the input, as determined by the given comparator.
     * The input iterator is consumed.
     */
    static public <T> T min(Iterator<? extends T> ts, Comparator<? super T> comp) {
        return min(ts, new Min<T>(comp));
    }


    /**
     * Returns the smallest T value in the input.  The input iterator is consumed.
     */
    static public <T extends Comparable<? super T>>  T min(Iterator<? extends T> ts) {
        return min(ts, new Min.Comparable<T>());
    }


    /**
     * Returns the smallest T value in the input, as determined by the given functor.
     * The functor is expected to return the smaller of its two arguments.
     * The input iterator is consumed.
     */
    static public <T> T min(Iterator<? extends T> ts, BinaryFunctor<T,T,T> fn) {
        return accumulate(ts, fn);
    }

    
    // -------------------
    // maximum value
    // -------------------

    
    /**
     * Returns the largest T value in the input, as determined by the given comparator.
     */
    static public <T> T max(T[] ts, Comparator<? super T> comp) {
        return max(iterate(ts), new Max<T>(comp));
    }


    /**
     * Returns the largest T value in the input.
     */
    static public <T extends Comparable<? super T>>  T max(T[] ts) {
        return max(iterate(ts), new Max.Comparable<T>());
    }


    /**
     * Returns the largest T value in the input, as determined by the given functor.
     * The functor is expected to return the smaller of its two arguments.
     */
    static public <T> T max(T[] ts, BinaryFunctor<T,T,T> fn) {
        return max(iterate(ts), fn);
    }


    /**
     * Returns the largest T value in the input, as determined by the given comparator.
     */
    static public <T> T max(Iterable<? extends T> ts, Comparator<? super T> comp) {
        return max(ts.iterator(), new Max<T>(comp));
    }


    /**
     * Returns the largest T value in the input.
     */
    static public <T extends Comparable<? super T>> T max(Iterable<? extends T> ts) {
        return max(ts.iterator(), new Max.Comparable<T>());
    }


    /**
     * Returns the largest T value in the input, as determined by the given functor.
     * The functor is expected to return the smaller of its two arguments.
     */
    static public <T> T max(Iterable<? extends T> ts, BinaryFunctor<T,T,T> fn) {
        return max(ts.iterator(), fn);
    }


    /**
     * Returns the largest T value in the input, as determined by the given comparator.
     * The input iterator is consumed.
     */
    static public <T> T max(Iterator<? extends T> ts, Comparator<? super T> comp) {
        return max(ts, new Max<T>(comp));
    }


    /**
     * Returns the largest T value in the input.  The input iterator is consumed.
     */
    static public <T extends Comparable<? super T>>  T max(Iterator<? extends T> ts) {
        return max(ts, new Max.Comparable<T>());
    }


    /**
     * Returns the largest T value in the input, as determined by the given functor.
     * The functor is expected to return the smaller of its two arguments.
     * The input iterator is consumed.
     */
    static public <T> T max(Iterator<? extends T> ts, BinaryFunctor<T,T,T> fn) {
        return accumulate(ts, fn);
    }


    // -------------------
    // total value
    // -------------------

    /**
     * Returns the sum of the values
     */
    @SuppressWarnings("unchecked")
    // Warning generated by the cast of the array's component type class to Class<T>
    static public <T extends Number> T sum(T[] ts) {
        return sum((Class<T>) ts.getClass().getComponentType(), iterate(ts));
    }

    
    /**
     * Returns the sum of the values
     */
    static public <T extends Number> T sum(Class<T> type, T[] ts) {
        return sum(type, iterate(ts));
    }

    
    /**
     * Returns the sum of the values
     */
    static public <T extends Number> T sum(Class<T> type, Iterable<? extends T> ts) {
        return sum(type, ts.iterator());
    }

    
    /**
     * Returns the sum of the values.  The iterator is entirely consumed.
     */
    static public <T extends Number> T sum(Class<T> type, Iterator<? extends T> ts) {
        Arithmetic<T> math = ArithmeticFactory.getArithmetic(type);
        if (math == null) {
            String msg = "No implementation of Arithmetic registered for {0}";
            throw new IllegalArgumentException(MessageFormat.format(msg, type));
        }
        
        return accumulate(ts, math.zero(), new Plus<T>(type));
    }

    // -------------------
    // accumulated value
    // -------------------
                                   

    static public <T> T accumulate(T[] ts, BinaryFunctor<T,T,T> fn) {
        return accumulate(iterate(ts), fn);
    }

    
    static public <T> T accumulate(T[] ts, T initial, BinaryFunctor<T,T,T> fn) {
        return accumulate(iterate(ts), initial, fn);
    }
    
    
    static public <T> T accumulate(Iterable<? extends T> values, BinaryFunctor<T,T,T> fn) {
        return accumulate(values.iterator(), fn);
    }

    
    static public <T> T accumulate(Iterable<? extends T> values, T init, BinaryFunctor<T,T,T> fn) {
        return accumulate(values.iterator(), init, fn);
    }
    

    static public <T> T accumulate(Iterator<? extends T> iterator, BinaryFunctor<T,T,T> fn) {
        return accumulate(iterator, iterator.hasNext() ? iterator.next() : null, fn);
    }
                                   
                                   
    static public <T> T
    accumulate(Iterator<? extends T> iterator, T init, BinaryFunctor<T,T,T> fn) {
        T value = init;
        while(iterator.hasNext()) {
            value = fn.fn(value, iterator.next());
        }
        
        return value;
    }
    

    // -------------------
    // unary evaluation
    // -------------------

    static public <T,R> UnaryFunctor<T,R> forEach(T[] ts, UnaryFunctor<T,R> fn) {
        return forEach(iterate(ts), fn);
    }

    
    static public <T,R> UnaryFunctor<T,R> forEach(Iterable<? extends T> c, UnaryFunctor<T,R> fn) {
        return forEach(c.iterator(), fn);
    }

    
    static public <T,R> UnaryFunctor<T,R> forEach(Iterator<? extends T> iter, UnaryFunctor<T,R> fn) {
        while(iter.hasNext()) {
            fn.fn(iter.next());
        }

        return fn;
    }
                                   
    // --------------
    // Average value
    // --------------
    
    /**
     * Returns the average of the values
     */
    @SuppressWarnings("unchecked")
    // Warning generated by the cast of the array's component type class to Class<T>
    static public <T extends Number> T average(T[] ts) {
        return average((Class<T>) ts.getClass().getComponentType(), iterate(ts));
    }

    
    /**
     * Returns the average of the values
     */
    static public <T extends Number> T average(Class<T> type, T[] ts) {
        return average(type, iterate(ts));
    }

    
    /**
     * Returns the average of the values
     */
    static public <T extends Number> T average(Class<T> type, Iterable<T> ts) {
        return average(type, ts.iterator());
    }


    /**
     * Returns the average of the values.  The iterator is entirely consumed.
     */
    static public <T extends Number> T average(Class<T> type, Iterator<T> ts) {
        return new Average<T>(type).fn(ts);
    }

    // --------------
    // lookup value
    // --------------

    // UNTESTED
    
    /**
     * Returns the first item in the array that equals the given value
     * using the equals() method or null if no such item exists
     */
    static public <T> T lookup(T[] ts, T value) {
        return lookup(iterate(ts), equalTo(value));
    }

    
    /**
     * Returns the first item in the array that equals the given value
     * using the given comparator or null if no such item exists
     */
    static public <T> T lookup(T[] ts, T value, Comparator<? super T> comp){
        return lookup(iterate(ts), equalTo(comp,value));
    }

    
    /**
     * Returns the first item in the array that equals the given value
     * using the given Equality operator or null if no such item exists
     */
    static public <T> T lookup(T[] ts, T value, Equality<T> eq) {
        return lookup(iterate(ts), equalTo(eq, value));
    }

    
    /**
     * Returns the first item that in the array for which the given functor
     * returns TRUE, or null if no such item exists
     */
    static public <T> T lookup(T[] ts, UnaryFunctor<T,Boolean> fn) {
        return lookup(iterate(ts), fn);
    }

    
    /**
     * Returns the first item in the input that equals the given value
     * using the equals() method or null if no such item exists
     */
    static public <T> T lookup(Iterable<? extends T> ts, T value){
        return lookup(ts.iterator(), equalTo(value));
    }

    
    /**
     * Returns the first item in the input that equals the given value
     * using the given comparator or null if no such item exists
     */
    static public <T> T lookup(Iterable<? extends T> ts, T value, Comparator<? super T> comp){
        return lookup(ts.iterator(), equalTo(comp,value));
    }

    
    /**
     * Returns the first item in the input that equals the given value
     * using the given Equality operator or null if no such item exists
     */
    static public <T> T lookup(Iterable<? extends T> ts, T value, Equality<T> eq) {
        return lookup(ts.iterator(), equalTo(eq, value));
    }

    
    /**
     * Returns the first item that in the input for which the given functor
     * returns TRUE, or null if no such item exists
     */
    static public <T> T lookup(Iterable<? extends T> ts, UnaryFunctor<T,Boolean> fn) {
        return lookup(ts.iterator(), fn);
    }

    
    /**
     * Returns the first item in the input that equals the given value
     * using the equals() method or null if no such item exists
     */
    static public <T> T lookup(Iterator<? extends T> ts, T value){
        return lookup(ts, equalTo(value));
    }

    
    /**
     * Returns the first item in the input that equals the given value
     * using the given comparator or null if no such item exists
     */
    static public <T> T lookup(Iterator<? extends T> ts, T value, Comparator<? super T> comp){
        return lookup(ts, equalTo(comp,value));
    }

    
    /**
     * Returns the first item in the input that equals the given value
     * using the given Equality operator or null if no such item exists
     */
    static public <T> T lookup(Iterator<? extends T> ts, T value, Equality<T> eq) {
        return lookup(ts, equalTo(eq, value));
    }

    
    /**
     * Returns the first value that meets the selection criteria, or null
     */
    static public <T> T lookup(Iterator<? extends T> ts, UnaryFunctor<T,Boolean> fn) {
        Iterator<? extends T> iter = Find.find(ts, fn);
        return (iter.hasNext()) ? iter.next() : null;
    }    
}
