// ============================================================================
// $Id: Transform.java,v 1.8 2008/02/08 04:22:34 davidahall Exp $
// Copyright (c) 2002-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.algorithms;

import static net.sf.jga.fn.comparison.ComparisonFunctors.equalTo;
import static net.sf.jga.util.ArrayUtils.iterable;
import static net.sf.jga.util.CollectionUtils.append;

import java.util.Collection;
import java.util.Iterator;

import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.fn.adaptor.ConditionalUnary;
import net.sf.jga.fn.adaptor.ConstantUnary;
import net.sf.jga.fn.adaptor.Identity;
import net.sf.jga.fn.comparison.EqualTo;
import net.sf.jga.util.LookAheadIterator;

/**
 * Algorithms that process the elements of a collection, iteration, or iterable resource,
 * and present the results.
 * <p>
 * Copyright &copy; 2002-2006  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/
public class Transform {

    static private <T> UnaryFunctor<T,T>
    conditional(UnaryFunctor<T,Boolean> test, UnaryFunctor<T,T> xform)
    {
        return new ConditionalUnary<T,T>(test, xform, new Identity<T>());
    }

    // ============
    // Arrays
    // ============
    
    /**
     * Produces an iterable over the contents of the array, such that all instances of the
     * given value are replaced with the given replacement value.
     */
    static public <T> Iterable<T> replace(T[] ts, T value, T replacement) {
        return transform(ts, conditional(equalTo(value), new ConstantUnary<T,T>(replacement)));
    }

    
    /**
     * Produces an iterable over the contents of the array, such that all elements for which
     * the given test returns TRUE are replaced with the given replacement value.
     */
    static public <T> Iterable<T> replace(T[] ts, UnaryFunctor<T,Boolean> test, T replacement) {
        return transform(ts, conditional(test, new ConstantUnary<T,T>(replacement)));
    }

    
    /**
     * Produces an iterable over the contents of the array, such that all elements for which
     * the given test returns TRUE are replaced with the results of passing that element to
     * the given replacement functor.
     */
    static public <T> Iterable<T>
    replace(T[] ts, UnaryFunctor<T,Boolean> test, UnaryFunctor<T,T> fn)
    {
        return transform(ts, conditional(test, fn));
    }

        
    /**
     * Passes each element in the array to the functor, iterating over the results
     */
    static public <T,R> Iterable<R> transform(T[] ts, UnaryFunctor<T,R> fn) {
        return new TransformIterable<T,R>(iterable(ts), fn);
    }

    
    /**
     * Passes successive pairs of elements in the array to the functor, iterating over the results
     */
    static public <T,R> Iterable<R> transform(T[] ts, BinaryFunctor<T,T,R> fn) {
        return new AdjacentIterable<T,R>(iterable(ts), fn);
    }

    
    /**
     * Passes corresponding pairs of elements in the arrays to the functor,
     * iterating over the results
     */
    static public <T1,T2,R> Iterable<R> transform(T1[] ts1, T2[] ts2, BinaryFunctor<T1,T2,R> fn) {
        return new BinaryIterable<T1,T2,R>(iterable(ts1), iterable(ts2), fn);
    }
    
    // ============
    // Iterables
    // ============
    
    /**
     * Produces an iterable over the contents of the input, such that all instances of the
     * given value are replaced with the given replacement value.
     */
    static public <T> Iterable<T> replace(Iterable<T> ts, T value, T replacement) {
        return transform(ts, conditional(new EqualTo<T>().bind2nd(value),
                                         new ConstantUnary<T,T>(replacement)));
    }

        
    /**
     * Produces an iterable over the contents of the input, such that all elements for which
     * the given test returns TRUE are replaced with the given replacement value.
     */
    static public <T> Iterable<T>
    replace(Iterable<T> ts, UnaryFunctor<T,Boolean> test, T replacement)
    {
        return transform(ts, conditional(test, new ConstantUnary<T,T>(replacement)));
    }

    
    /**
     * Produces an iterable over the contents of the input, such that all elements for which
     * the given test returns TRUE are replaced with the results of passing that element to
     * the given replacement functor.
     */
    static public <T> Iterable<T>
    replace(Iterable<T> ts, UnaryFunctor<T,Boolean> test, UnaryFunctor<T,T> fn)
    {
        return transform(ts, conditional(test, fn));
    }
    
    
    /**
     * Passes each element in the iterable resource to the functor, iterating over the results
     */
    static public <T,R> Iterable<R> transform(Iterable<? extends T> i, UnaryFunctor<T,R> fn) {
        return new TransformIterable<T,R>(i, fn);
    }

    
    /**
     * Passes successive pairs of elements in the iterable resource to the functor,
     * iterating over the results
     */
    static public <T,R> Iterable<R> transform(Iterable<? extends T> i, BinaryFunctor<T,T,R> fn) {
        return new AdjacentIterable<T,R>(i, fn);
    }

    
    /**
     * Passes corresponding pairs of elements in the iterable resources to the functor,
     * iterating over the results
     */
    static public <T1,T2,R> Iterable<R>
    transform(Iterable<? extends T1> i1, Iterable<? extends T2> i2, BinaryFunctor<T1,T2,R> fn)
    {
        return new BinaryIterable<T1,T2,R>(i1, i2, fn);
    }
    
    // ============
    // Iterators
    // ============
    
    /**
     * Produces an iterator over the contents of the given iterator, such that all instances of the
     * given value are replaced with the given replacement value.
     */
    static public <T> Iterator<T> replace(Iterator<? extends T> ts, T value, T replacement) {
        return transform(ts, conditional(new EqualTo<T>().bind2nd(value),
                                         new ConstantUnary<T,T>(replacement)));
    }

        
    /**
     * Produces an iterator over the contents of the given iterator, such that all elements for which
     * the given test returns TRUE are replaced with the given replacement value.
     */
    static public <T> Iterator<T>
    replace(Iterator<? extends T> ts, UnaryFunctor<T,Boolean> test, T replacement)
            {
        return transform(ts, conditional(test, new ConstantUnary<T,T>(replacement)));
    }

    
    /**
     * Produces an iterator over the contents of the given iterator, such that all elements for which
     * the given test returns TRUE are replaced with the results of passing that element to
     * the given replacement functor.
     */
    static public <T> Iterator<T>
    replace(Iterator<? extends T> ts, UnaryFunctor<T,Boolean> test, UnaryFunctor<T,T> fn)
    {
        return transform(ts, conditional(test, fn));
    }

    /**
     * Passes each element in the iterator to the functor, iterating over the results
     */
    static public <T,R> Iterator<R> transform(Iterator<? extends T> iter, UnaryFunctor<T,R> fn) {
        return new TransformIterator<T,R>(iter, fn);
    }

    
    /**
     * Passes successive pairs of elements in the iterator to the functor, iterating over the results
     */
    static public <T,R> Iterator<R> transform(Iterator<? extends T> iter, BinaryFunctor<T,T,R> fn) {
        return new AdjacentIterator<T,R>(iter, fn);
    }

    
    /**
     * Passes corresponding pairs of elements in the iterators to the functor,
     * iterating over the results
     */
    static public <T1,T2,R> Iterator<R>
    transform(Iterator<? extends T1> iter1, Iterator<? extends T2> iter2, BinaryFunctor<T1,T2,R> fn)
    {
        return new BinaryIterator<T1,T2,R>(iter1, iter2, fn);
    }

    // ============
    // IterableCopy
    // ============
    
    /**
     * Appends the contents of the input to the output, such that all instances of the
     * given value are replaced with the given replacement value
     */
    static public <T, TCollection extends Collection<? super T>> TCollection
    replace(Iterable<? extends T> cin, T value, T repl, TCollection cout)
    {
        return append(cout, replace(cin.iterator(), value, repl));
    }

    
    /**
     * Appends the contents of the input to the output, such that all elements for which
     * the given test returns TRUE are replaced with the given replacement value.
     */
    static public <T, TCollection extends Collection<? super T>> TCollection
    replace(Iterable<? extends T> cin, UnaryFunctor<T,Boolean> test, T repl, TCollection cout)
    {
        return append(cout, replace(cin.iterator(), test, repl));
    }

    
    /**
     * Appends the contents of the input to the output, such that all elements for which
     * the given test returns TRUE are replaced with the results of passing that element to
     * the given replacement functor.
     */
    static public <T, TCollection extends Collection<? super T>> TCollection
    replace(Iterable<? extends T> cin, UnaryFunctor<T,Boolean> test, UnaryFunctor<T,T> fn, TCollection cout)
    {
        return append(cout, replace(cin.iterator(), test, fn));
    }
    
    
    /**
     * Appends the transformed contents of the input to the output.
     */
    static public <T, RCollection extends Collection<? super R>, R> RCollection
    transform(Iterable<? extends T> cin, UnaryFunctor<T,R> fn, RCollection cout) {
        return append(cout, transform(cin.iterator(), fn));
    }
    
    
    /**
     * Appends the results of transforming successive pairs of elements from the input to the output.
     */
    static public <T, RCollection extends Collection<? super R>, R> RCollection
    transform(Iterable<? extends T> cin, BinaryFunctor<T,T,R> fn, RCollection cout) {
        return append(cout, transform(cin.iterator(), fn));
    }
    
    
    /**
     * Appends the results of transforming corresponding pairs of elements from the
     * input sto the output.
     */
    static public <T1, T2, RCollection extends Collection<? super R>, R> RCollection
    transform(Iterable<? extends T1> cin1, Iterable<? extends T2> cin2, BinaryFunctor<T1,T2,R> fn, RCollection cout)
    {
        return append(cout, transform(cin1.iterator(), cin2.iterator(), fn));
    }
    
    // =================
    // TransformIterable
    // =================
    
    /**
     * Produces iterators that will process elements of an iterable resource and present
     * the results.
     * <p>
     * Copyright &copy; 2005  David A. Hall
     */

    static public class TransformIterable<T,R> implements Iterable<R> {
        // The contents
        private Iterable<? extends T> _delegate;
    
        // The process to be applied
        private UnaryFunctor<T,R> _fn;

        /**
         * Builds a TransformIterable for the given base.
         * @throws IllegalArgumentException if the base iterable is null
         */
        public TransformIterable(Iterable<? extends T> base, UnaryFunctor<T,R> process) {
            _delegate = base;
            _fn = process;
        }


        // - - - - - - - - - - -
        // Iterable<T> interface
        // - - - - - - - - - - -

        public Iterator<R> iterator() {
            return new TransformIterator<T,R>(_delegate.iterator(), _fn);
        }
    }
    
    // =================
    // TransformIterable
    // =================
    
    /*
     * Iterator that returns the results of applying the given functor to the
     * elements of the given iterator.
     */
    static public class TransformIterator<T,R> implements Iterator<R> {

        // The base iterator
        private Iterator<? extends T> _base;
        
        // Functor applied to each element of the base iterator
        private UnaryFunctor<T,R> _fn;

        /**
         * Builds a TransformIterator that applies the given functor to each element
         * of the given base iterator.
         */
        public TransformIterator(Iterator<? extends T> iter, UnaryFunctor<T,R> fn) {
            _base = iter;
            _fn = fn;
        }

        // - - - - - - - - - - -
        // Iterator<R> interface
        // - - - - - - - - - - -
    
        public boolean hasNext(){
            return _base.hasNext();
        }

        public R next() {
            return _fn.fn(_base.next());
        }

        public void remove() { throw new UnsupportedOperationException(); }
    }


    /**
     * Produces iterators that will process successive pairs of elements of an iterable resource
     * and present the results.
     */

    static public class AdjacentIterable<T,R> implements Iterable<R> {
        // The contents
        private Iterable<? extends T> _delegate;
    
        // The process to be applied
        private BinaryFunctor<T,T,R> _fn;

        /**
         * Builds a TransformIterable for the given base.
         * @throws IllegalArgumentException if the base iterable is null
         */
        public AdjacentIterable(Iterable<? extends T> base, BinaryFunctor<T,T,R> process) {
            _delegate = base;
            _fn = process;
        }

        // - - - - - - - - - - -
        // Iterable<R> interface
        // - - - - - - - - - - -

        public Iterator<R> iterator() {
            return new AdjacentIterator<T,R>(_delegate.iterator(), _fn);
        }
    }
    
    
    /**
     * Iterator that applies a given BinaryFunctor to successive pairs of elements
     * from a given iterator, returning the results as elements.
     */
    static public class AdjacentIterator<T,R> implements Iterator<R> {

        // The base iterator
        private LookAheadIterator<? extends T> _base;

        // Functor applied to adjacent values in the base iterator
        private BinaryFunctor<T,T,R> _fn;

        /**
         * Builds a TransormAdjacentIterator that applies the given functor to
         * adjacent elements of the given base iterator.
         */
        public AdjacentIterator(Iterator<? extends T> iter, BinaryFunctor<T,T,R> fn) {
            _base = new LookAheadIterator<T>(iter, 2);
            _fn = fn;
        }

        // - - - - - - - - - - -
        // Iterator<R> interface
        // - - - - - - - - - - -
    
        public boolean hasNext(){
            return _base.hasNextPlus(2);
        }

        public R next() {
            T next = _base.next();
            return _fn.fn(next,_base.peek(1));
        }

        public void remove() { throw new UnsupportedOperationException(); }
    }


    /**
     * Produces iterators that applies a given BinaryFunctor to corresponding pairs of elements
     * from a pair of iterators, returning the results.
     */

    static public class BinaryIterable<T1,T2,R> implements Iterable<R> {
        // The contents
        private Iterable<? extends T1> _delegate1;
        private Iterable<? extends T2> _delegate2;
    
        // The process to be applied
        private BinaryFunctor<T1,T2,R> _fn;

        /**
         * Builds a TransformIterable for the given base.
         * @throws IllegalArgumentException if the base iterable is null
         */
        public BinaryIterable(Iterable<? extends T1> base1, Iterable<? extends T2> base2,
                              BinaryFunctor<T1,T2,R> process)
        {
            _delegate1 = base1;
            _delegate2 = base2;
            _fn = process;
        }

        // - - - - - - - - - - -
        // Iterable<R> interface
        // - - - - - - - - - - -

        public Iterator<R> iterator() {
            return new BinaryIterator<T1,T2,R>(_delegate1.iterator(), _delegate2.iterator(), _fn);
        }
    }
    
    
    /**
     * Iterator that applies a given BinaryFunctor to corresponding pairs of elements
     * from a pair of iterators, returning the results.
     */
    static public class BinaryIterator<T1,T2,R> implements Iterator<R> {
        // The two base iterators
        private Iterator<? extends T1> _i1;
        private Iterator<? extends T2> _i2;
    
        // Functor applied to corresponding elements of the base iterator
        private BinaryFunctor<T1,T2,R> _bf;
        
        /**
         * Builds a TransformBinaryIterator that applies the given functor to
         * corresponding elements of the given base iterators.
         */
        public BinaryIterator(Iterator<? extends T1> i1, Iterator<? extends T2> i2, BinaryFunctor<T1,T2,R> bf) {
            _i1 = i1;
            _i2 = i2;
            _bf = bf;
        }

        // - - - - - - - - - - -
        // Iterator<R> interface
        // - - - - - - - - - - -
    
        public boolean hasNext() {
            return _i1.hasNext() && _i2.hasNext();
        }

        public R next() {
            return _bf.fn(_i1.next(), _i2.next());
        }
        
        public void remove() { throw new UnsupportedOperationException(); }
    }
}
