// ============================================================================
// $Id: Unique.java,v 1.7 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2006  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not
// use this file except in compliance with the License.
//
// From time to time, the license steward (initially Sun Microsystems,
// Inc.), and may publish revised and/or new versions of this License.
// You may not use, distribute, or otherwise make this file available
// under subsequent versions of the License.
//
// You should have received a copy of the the License along with this
// file: if not, a copy of the License is available at
//
// http://www.sun.com/cddl/cddl.html
//
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// ============================================================================

package net.sf.jga.algorithms;

import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.NoSuchElementException;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.comparison.EqualTo;

import static net.sf.jga.util.ArrayUtils.*;
import static net.sf.jga.util.CollectionUtils.*;

/**
 * Algorithms that eliminate successive duplicate entries from a collection, iteration,
 * or iterable resource.
 * <p>
 * Copyright &copy; 2006  David A. Hall
 */

public class Unique {
    
    // ============
    // Arrays
    // ============
    
    /**
     * Returns the elements of the array, skipping duplicate entries.  This version uses T.equals()
     * to test for equality.
     */
    static public <T> Iterable<T> unique(T[] ts) {
        return new UniqueIterable<T>(iterable(ts));
    }

    
    /**
     * Returns the elements of the array, skipping duplicate entries as determined by the given
     * comparator.
     */
    static public <T> Iterable<T> unique(T[] ts, Comparator<? super T> comp) {
        return new UniqueIterable<T>(iterable(ts), comp);
    }

    
    /**
     * Returns the elements of the array, skipping duplicate entries as determined by the given
     * predicate.  The predicate should return true when the adjacent items are the same.
     */
    static public <T> Iterable<T> unique(T[] ts, BinaryFunctor<T,T,Boolean> eq) {
        return new UniqueIterable<T>(iterable(ts), eq);
    }
    
    // ============
    // Iterables
    // ============
    
    /**
     * Returns the contents of the input, skipping duplicate entries.  This version uses T.equals()
     * to test for equality.
     */
    static public <T> Iterable<T> unique(Iterable<? extends T> i) {
        return new UniqueIterable<T>(i);
    }

    
    /*
     * Returns the contents of the input, skipping duplicate entries as determined by the given
     * comparator.
     */
    static public <T> Iterable<T> unique(Iterable<? extends T> i, Comparator<? super T> comp) {
        return new UniqueIterable<T>(i, comp);
    }

    
    /*
     * Returns the contents of the input, skipping duplicate entries as determined by the given
     * predicate. The predicate should return true when the adjacent items are the same.
     */
    static public <T> Iterable<T> unique(Iterable<? extends T> i, BinaryFunctor<T,T,Boolean> eq) {
        return new UniqueIterable<T>(i, eq);
    }
    
    // ============
    // Iterators
    // ============
    
    /**
     * Returns the contents of the iterator, skipping duplicate entries.  This version uses
     * T.equals() to test for equality.
     */
    
    static public <T> Iterator<T> unique(Iterator<? extends T> iterator) {
        return new UniqueIterator<T>(iterator);
    }

    
    /**
     * Returns the contents of the iterator, skipping duplicate entries as determined by the given
     * comparator. 
     */
    
    static public <T> Iterator<T>
    unique(Iterator<? extends T> iterator, Comparator<? super T> comp) {
        return new UniqueIterator<T>(iterator, comp);
    }
    

    /**
     * Returns the contents of the iterator, skipping duplicate entries as determined by the given
     * predicate. The predicate should return true when the adjacent items are the same.
     */
    
    static public <T> Iterator<T>
    unique(Iterator<? extends T> iterator, BinaryFunctor<T,T,Boolean> eq) {
        return new UniqueIterator<T>(iterator, eq);
    }

    // ============
    // IterableCopy
    // ============
    
    /**
     * Appends the contents of the input (skipping duplicate entries) to the output.  This version
     * uses T.equals() to test for equality.
     */

    static public <T, TCollection extends Collection<? super T>> TCollection
    unique(Iterable<? extends T> lin, TCollection cout) {
        return append(cout, unique(lin.iterator()));
    }


    /**
     * Appends the contents of the input (skipping duplicate entries as determined by the given
     * comparator) to the output.
     */

    static public <T, TCollection extends Collection<? super T>> TCollection
    unique(Iterable<? extends T> lin, Comparator<? super T> comp, TCollection cout) {
        return append(cout, unique(lin.iterator(), comp));
    }


    /**
     * Appends the contents of the input (skipping duplicate entries as determined by the given
     * predicate) to the output.
     */

    static public <T, TCollection extends Collection<? super T>> TCollection
    unique(Iterable<? extends T> lin, BinaryFunctor<T,T,Boolean> eq, TCollection cout) {
        return append(cout, unique(lin.iterator(), eq));
    }

    
    /**
     * Produces iterators that will not return the same element twice in succession.
     * <p>
     * Copyright &copy; 2005  David A. Hall
     */

    static public class UniqueIterable<T> implements Iterable<T> {

        // The contents
        private Iterable<? extends T> _delegate;
    
        // The test
        private BinaryFunctor<T,T,Boolean> _pred;

        /**
         * Builds a UniqueIterable for the given resource.  It will use an EqualTo
         * predicate by default.
         * @throws IllegalArgumentException if the base iterable is null
         */
        public UniqueIterable(Iterable<? extends T> base) {
            this(base, new EqualTo<T>());
        }

        /**
         * Builds a UniqueIterable for the given base iterator, that uses the given Comparator
         * @throws IllegalArgumentException if either argument is null
         */
        public UniqueIterable(Iterable<? extends T> base, Comparator<? super T> comp) {
            this(base, new EqualTo<T>(comp));
        }


        /**
         * Builds a UniqueIterable for the given base iterator that uses the given
         * predicate to compare successive elements.  The predicate should return
         * TRUE if its first argument is equal to the second.
         * @throws IllegalArgumentException if the base iterable is null
         */

        public UniqueIterable(Iterable<? extends T> base, BinaryFunctor<T,T,Boolean> pred) {
            if (base == null) 
                throw new IllegalArgumentException( "a base iterable is required");
        
            _delegate = base;
            _pred = pred;
        }

        // - - - - - - - - - - -
        // Iterable<T> interface
        // - - - - - - - - - - -

        public Iterator<T> iterator() {
            return new UniqueIterator<T>(_delegate.iterator(), _pred); }
    
    }

    
    /**
     * Iterator that will not return the same element twice in succession.
     */

    static public class UniqueIterator<T> implements Iterator<T> {
        // The base iterator
        private Iterator<? extends T> _base;

        // The test for equality
        private BinaryFunctor<T,T,Boolean> _eq;

        // null at startup, TRUE if hasNext() has been called since last next()
        private Boolean _testedNext;

        // flag indicating that there is at least new element remaining
        private boolean _hasNext;
    
        // previous item returned
        private T _lastValue;

        // next item to return
        private T _nextValue;

        /**
         * Builds a UniqueIterator for the given base iterator, using a test based on the equals()
         * method of type T.
         */
      
        public UniqueIterator(Iterator<? extends T> base) {
            this(base, new EqualTo<T>());
        }

        /**
         * Builds a UniqueIterator for the given base iterator, using the given comparator.
         */
      
        public UniqueIterator(Iterator<? extends T> base, Comparator<? super T> comp) {
            this(base, new EqualTo<T>(comp));
        }

        /**
         * Builds a UniqueIterator for the given base iterator that uses the given predicate to
         * compare adjacent elements.  The predicate should return true if the adjacent elements
         * are the same.
         * @throws IllegalArgumentException if either argument is null
         */

        public UniqueIterator(Iterator<? extends T> base, BinaryFunctor<T,T,Boolean> eq) {
            if (base == null) 
                throw new IllegalArgumentException("base iterator is required");

            if (eq == null) 
                throw new IllegalArgumentException("functor is required");

            _base = base;
            _eq = eq;
        }

        // - - - - - - - - - - -
        // Iterator<T> interface
        // - - - - - - - - - - -
    
        public boolean hasNext() {
            // make sure this isn't the first time through
            if (_testedNext != null) {

                // make sure that we haven't already tested hasNext()
                if (!_testedNext.booleanValue()) { // autounbox
                
                    // first test since next() called
                    _testedNext = Boolean.TRUE;
                    _hasNext = false; // in case we walk off the end
                    while (_base.hasNext()) {
                        _nextValue = _base.next();
                        if (_eq.fn(_lastValue, _nextValue).booleanValue()) { // autounbox
                            continue; // skip this value, try the next one
                        }
                    
                        _hasNext = true;
                        break;
                    }
                }
            }
            else {
                // this is the first time through -- can't call the functor with the
                // first value (the first value may be null, so we can't take the
                // chance that it equals the not yet initialized _lastValue member);
            
                _testedNext = Boolean.TRUE;
                _hasNext = _base.hasNext();
                if (_hasNext) {
                    _nextValue = _base.next();
                }
            }
        
            return _hasNext;

        }

        public T next() {
            // make sure that hasNext() has been called
            if (_testedNext == null || !_testedNext.booleanValue()) { // autounbox
                hasNext();
            }

            // just in case the result of hasNext() has been ignored
            if (!_hasNext)
                throw new NoSuchElementException();

            //stash off the value being returned for future testing
            _testedNext = Boolean.FALSE;
            _lastValue = _nextValue;
            return _nextValue;
        }

        public void remove() { throw new UnsupportedOperationException(); }
    }
}
