// ============================================================================
// $Id: AdaptorVisitor.java,v 1.8 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn;

import net.sf.jga.fn.adaptor.*;
import net.sf.jga.fn.logical.All;
import net.sf.jga.fn.logical.Any;
import java.util.Iterator;

/**
 * Visitor that performs a walk of compound functor structures.  This visitor
 * implements the Visitor interface associated with all of the compounding
 * functors in the <a href="adaptor/package-summary.html">net.sf.jga.fn.adaptor</a>
 * package, as well as the two that are in the
 * <a href="logical/package-summary.html">net.sf.jga.fn.logical</a> package.
 * <p>
 * Basing visitors on this base class will allow most implementations to ignore
 * the tree structure, and implement visiting the leaf node functors that are
 * of interest.  When used in this way, the tree nodes will be ignored by the
 * visitor (exception that the visit walks through them).  If the tree nodes
 * are to be considered during the visit, then the implementation can override
 * methods contained in this class: depending on where in the overridden
 * implementation the call to super() occurs, either breadth-first, depth-first,
 * or in-line traversal can be supported.
 * <p>
 * Copyright &copy; 2005  David A. Hall
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

public class AdaptorVisitor extends AbstractVisitor
        implements
            All.Visitor,
            AndBinary.Visitor,
            AndGenerator.Visitor,
            AndUnary.Visitor,
            Any.Visitor,
            ApplyBinary.Visitor,
            ApplyGenerator.Visitor,
            ApplyUnary.Visitor,
            Bind.Visitor,
            Bind1st.Visitor,
            Bind2nd.Visitor,
            ChainBinary.Visitor,
            ChainUnary.Visitor,
            ComposeBinary.Visitor,
            ComposeUnary.Visitor,
            CompoundBinary.Visitor,
            CompoundGenerator.Visitor,
            CompoundUnary.Visitor,
            ConditionalBinary.Visitor,
            ConditionalGenerator.Visitor,
            ConditionalUnary.Visitor,
            Distribute.Visitor,
            Generate.Visitor,
            Generate1st.Visitor,
            Generate2nd.Visitor,
            GenerateBinary.Visitor,
            GenerateUnary.Visitor,
            OrBinary.Visitor,
            OrGenerator.Visitor,
            OrUnary.Visitor
{
    @SuppressWarnings("unchecked")
    // The type of the Iterator would be <Iterator<UnaryFunctor<?,Boolean>>.  The eclipse
    // compiler cannot convert from Iterator<UnaryFunctor<capture#1-of ?,Boolean>> to 
    // Iterator<UnaryFunctor<?,Boolean>, or even Iterator<UnaryFunctor>
    public void visit(All<?> host) {
        for(Iterator iter = host.branches(); iter.hasNext(); ) {
            ((UnaryFunctor) iter.next()).accept(this);
        }
    }
    
    public void visit(AndBinary<?,?> host) {
        host.getFirstFunctor().accept(this);
        host.getSecondFunctor().accept(this);
    }
        
    public void visit(AndGenerator host) {
        host.getFirstFunctor().accept(this);
        host.getSecondFunctor().accept(this);
    }
        
    public void visit(AndUnary<?> host) {
        host.getFirstFunctor().accept(this);
        host.getSecondFunctor().accept(this);
    }
        
    @SuppressWarnings("unchecked")
    // The type of the Iterator would be <Iterator<UnaryFunctor<?,Boolean>>.  The eclipse
    // compiler cannot convert from Iterator<UnaryFunctor<capture#1-of ?,Boolean>> to 
    // Iterator<UnaryFunctor<?,Boolean>, or even Iterator<UnaryFunctor>
    public void visit(Any<?> host) {
        for(Iterator iter = host.branches(); iter.hasNext(); ) {
            ((UnaryFunctor) iter.next()).accept(this);
        }
    }
    
    public void visit(ApplyBinary<?,?> host) {
        BinaryFunctor<?,?,?>[] fns = host.getFunctors();
        for (int i = 0; i < fns.length; ++i)
            fns[i].accept(this);
    }

    public void visit(ApplyGenerator host) {
        Generator<?>[] fns = host.getGenerators();
        for (int i = 0; i < fns.length; ++i)
            fns[i].accept(this);
    }
        
    public void visit(ApplyUnary<?> host) {
        UnaryFunctor<?,?>[] fns = host.getFunctors();
        for (int i = 0; i < fns.length; ++i)
            fns[i].accept(this);
    }
        
    public void visit(Bind<?,?> host) {
        host.getFunctor().accept(this);
    }
        
    public void visit(Bind1st<?,?,?> host) {
        host.getFunctor().accept(this);
    }
        
    public void visit(Bind2nd<?,?,?> host) {
        host.getFunctor().accept(this);
    }
        
    public void visit(ChainBinary<?,?,?,?> host) {
        host.getInnerFunctor().accept(this);
        host.getOuterFunctor().accept(this);
    }
        
    public void visit(ChainUnary<?,?,?> host) {
        host.getInnerFunctor().accept(this);
        host.getOuterFunctor().accept(this);
    }
        
    public void visit(ComposeBinary<?,?,?,?,?> host) {
        host.getFirstInnerFunctor().accept(this);
        host.getSecondInnerFunctor().accept(this);
        host.getOuterFunctor().accept(this);
    }
        
    public void visit(ComposeUnary<?,?,?,?> host) {
        host.getFirstInnerFunctor().accept(this);
        host.getSecondInnerFunctor().accept(this);
        host.getOuterFunctor().accept(this);
    }
        
    public void visit(CompoundBinary<?,?,?> host) {
        host.getFirstFunctor().accept(this);
        host.getSecondFunctor().accept(this);
    }
        
    public void visit(CompoundGenerator<?> host) {
        host.getFirstFunctor().accept(this);
        host.getSecondFunctor().accept(this);
    }
        
    public void visit(CompoundUnary<?,?> host) {
        host.getFirstFunctor().accept(this);
        host.getSecondFunctor().accept(this);
    }
        
    public void visit(ConditionalBinary<?,?,?> host) {
        host.getCondition().accept(this);
        host.getTrueFunctor().accept(this);
        host.getFalseFunctor().accept(this);
    }
        
    public void visit(ConditionalGenerator<?> host) {
        host.getCondition().accept(this);
        host.getTrueFunctor().accept(this);
        host.getFalseFunctor().accept(this);
    }
        
    public void visit(ConditionalUnary<?,?> host) {
        host.getCondition().accept(this);
        host.getTrueFunctor().accept(this);
        host.getFalseFunctor().accept(this);
    }
        
    public void visit(Distribute<?,?,?,?,?> host) {
        host.getFirstInnerFunctor().accept(this);
        host.getSecondInnerFunctor().accept(this);
        host.getOuterFunctor().accept(this);
    }
        
    public void visit(Generate<?,?> host) {
        host.getFunctor().accept(this);
        host.getGenerator().accept(this);
    }
        
    public void visit(Generate1st<?,?,?> host) {
        host.getFunctor().accept(this);
        host.getGenerator().accept(this);
    }
        
    public void visit(Generate2nd<?,?,?> host) {
        host.getFunctor().accept(this);
        host.getGenerator().accept(this);
    }
        
    public void visit(GenerateUnary<?,?> host) {
        host.getGenerator().accept(this);
    }
        
    public void visit(GenerateBinary<?,?,?> host) {
        host.getGenerator().accept(this);
    }

    public void visit(OrBinary<?,?> host) {
        host.getFirstFunctor().accept(this);
        host.getSecondFunctor().accept(this);
    }
        
    public void visit(OrGenerator host) {
        host.getFirstFunctor().accept(this);
        host.getSecondFunctor().accept(this);
    }
        
    public void visit(OrUnary<?> host) {
        host.getFirstFunctor().accept(this);
        host.getSecondFunctor().accept(this);
    }
}
