// ============================================================================
// $Id: BinaryFunctor.java,v 1.23 2009/09/06 22:30:06 davidahall Exp $
// Copyright (c) 2002-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn;

import java.io.Serializable;
import net.sf.jga.fn.adaptor.Bind1st;
import net.sf.jga.fn.adaptor.Bind2nd;
import net.sf.jga.fn.adaptor.ComposeBinary;
import net.sf.jga.fn.adaptor.ComposeUnary;
import net.sf.jga.fn.adaptor.Distribute;
import net.sf.jga.fn.adaptor.Generate1st;
import net.sf.jga.fn.adaptor.Generate2nd;
import net.sf.jga.fn.adaptor.Identity;

/**
 * A Function Object that takes two arguments and returns a result.  The two
 * arguments are of type <code>T1</code> and <code>T2</code>, and
 * the result is of type <code>R</code>
 * <p>
 * Copyright &copy; 2002-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

abstract public class BinaryFunctor<T1,T2,R> extends Functor<R> implements Serializable,Visitable {
    /**
     * Executes the function and returns the result.
     */
    abstract public R fn(T1 arg1, T2 arg2);

    /**
     * Factory method that creates a UnaryFunctor that binds a constant value to
     * this' first argument.  Given argument <b>x</b>, the new functor will
     * return <code>fn<sub>this</sub>(value, x)</code>
     */
    public UnaryFunctor<T2,R> bind1st (T1 value){
        return new Bind1st<T1,T2,R>(value, this);
    }

    /**
     * Factory method that creates a UnaryFunctor that binds a constant value to
     * this' second argument.  Given argument <b>x</b>, the new functor will
     * return <code>fn<sub>this</sub>(x, value)</code>
     */
    public UnaryFunctor<T1,R> bind2nd (T2 value){
        return new Bind2nd<T1,T2,R>(value, this);
    }

    /**
     * Factory method that creates a Generator that binds constant values to
     * this' arguments.  The new generator will return
     * <code>fn<sub>this</sub>(value1, value2)</code>
     */
    public Generator<R> bind (T1 value1, T2 value2) {
        return bind1st(value1).bind(value2);
    }

    /**
     * FactoryMethod that creates a UnaryFunctor that passes its argument to
     * each of the given functors, and uses the results as the arguments to this
     * function. Given argument <b>x</b>, the new functor will return
     * <code>fn<sub>this</sub>(g(x), h(x))</code>
     */
    public <F> UnaryFunctor<F,R> compose (UnaryFunctor<F,T1> g, UnaryFunctor<F,T2> h) {
        return new ComposeUnary<F,T1,T2,R>(g,h,this);
    }

    /**
     * FactoryMethod that creates a BinaryFunctor that passes its argument to
     * each of the given functors, and uses the results as the arguments to this
     * function. Given arguments <b>x</b> and <b>y</b>, the new functor will
     * return <code>fn<sub>this</sub>(g(x,y), h(x,y))</code>
     */
    public <F1,F2> BinaryFunctor<F1,F2,R>
    compose (BinaryFunctor<F1,F2,T1> g, BinaryFunctor<F1,F2,T2> h)
    {
        return new ComposeBinary<F1,F2,T1,T2,R>(g,h,this);
    }

    /**
     * FactoryMethod that creates a BinaryFunctor that passes each of its two
     * arguments to a pair of UnaryFunctors, then uses the results as arguments
     * to this function.  Given arguments <b>x</b> and <b>y</b>, the new functor
     * will return <code>fn<sub>this</sub>(g(x), h(y))</code>.  Note: this
     * method cannot be called compose, as it is only distinct from the unary
     * form of compose in its param types and its return type
     */
    public <F1,F2> BinaryFunctor<F1,F2,R> distribute (UnaryFunctor<F1,T1> g, UnaryFunctor<F2,T2> h)
    {
        return new Distribute<F1,F2,T1,T2,R>(g,h,this);
    }

    /**
     * FactoryMethod that converts this functor to a UnaryFunctor by using the
     * given generator to produce the first argument.  Given argument <b>y</b>,
     * the new functor will return <code>fn<sub>this</sub>(fn(), y)</code>.
     */
    
    public UnaryFunctor<T2,R> generate1st(Generator<T1> gen) {
        return new Generate1st<T1,T2,R>(this, gen);
    } 

    /**
     * FactoryMethod that converts this functor to a UnaryFunctor by using the
     * given generator to produce the first argument.  Given argument <b>x</b>,
     * the new functor will return <code>fn<sub>this</sub>(x, fn())</code>.
     */
    
    public UnaryFunctor<T1,R> generate2nd(Generator<T2> gen) {
        return new Generate2nd<T1,T2,R>(this, gen);
    } 

    /**
     * FactoryMethod that converts this functor to a Generator by using the two
     * given generators to produce the arguments.  The new functor will return
     * <code>fn<sub>this</sub>(gen1(), gen2())</code>.
     */ 
    
    public Generator<R> generate(Generator<T1> gen1, Generator<T2> gen2) {
        return generate1st(gen1).generate(gen2);
    } 

    // -----------------
    // Functor interface
    // -----------------
    
    /**
     * Executes the function and returns the result.
     * @throw ArrayIndexOutOfBoundsExcetpsion if fewer than 2 arguments are passed. 
     */
    @SuppressWarnings("unchecked")
    // This generates two unchecked cast warnings: we're crossing the line from
    // unsafe rawform code to typesafe generic code.  This interface is known
    // to be unsafe, and documented to the user as such.   
    public R eval(Object... args) {
        return fn((T1) args[0], (T2)args[1]);
    }
    
    /**
     * Binds the given values to this functor, starting from the first argument.
     * @param arg0
     * @param args
     * @return
     */
    @Override
    public Functor<R> bind0th(Object arg0, Object... args) {
        return bindNth(0, arg0, args);
    }
    
    /**
     * Binds the given values to this functor, starting from the given index.
     * @param idx
     * @param arg0
     * @param args
     * @return
     */
    @Override
    @SuppressWarnings("unchecked")
    // This generates several unchecked cast warnings: we're crossing the line from
    // unsafe rawform code to typesafe generic code.   
    public Functor<R> bindNth(int idx, Object arg0, Object... args) {
        if (idx < 0) {
            throw new IllegalArgumentException("Index must be non-negative");
        }
            
        if (idx > 1)
            return this; // args bound out of bounds would be ignored anyway
        
        if (idx == 1)
            return bind2nd((T2) arg0);
        
        if (args == null || args.length == 0)
            return bind1st((T1) arg0);

        return bind((T1) arg0, (T2) args[0]);
    }
    
    /**
     * Generates the argument at the given index to this functor 
     * @param idx
     * @param arg0
     * @param args
     * @return
     */
    @Override
    @SuppressWarnings("unchecked")
    // This generates unchecked cast warnings: we're crossing the line from
    // unsafe rawform code to typesafe generic code.   
    public Functor<R> generateNth(int idx, Generator<?> gen) {
        if (idx < 0) {
            throw new IllegalArgumentException("Index must be non-negative");
        }
            
        if (gen == null) {
            throw new IllegalArgumentException("Must supply a Generator to supply a runtime argument");
        }
        
        if (idx > 1) {
            return this; // args bound out of bounds would be ignored anyway
        }
        
        if (idx == 1) {
            return generate2nd((Generator<T2>) gen);
        }
        
        return generate1st((Generator<T1>) gen);
    }


    /**
     * Modifies the argument at the given index using the given functor 
     * @param idx
     * @param arg0
     * @param args
     * @return
     */
    @Override
    @SuppressWarnings("unchecked")
    // This generates unchecked cast warnings: we're crossing the line from
    // unsafe rawform code to typesafe generic code.   
    public Functor<R> composeNth(int idx, UnaryFunctor<?,?> uf) {
        if (idx < 0) {
            throw new IllegalArgumentException("Index must be non-negative");
        }
            
        if (uf == null) {
            throw new IllegalArgumentException("Must supply a UnaryFunctor to modify a runtime argument");
        }
        
        if (idx > 1) {
            return this; // args bound out of bounds would be ignored anyway
        }
        
        if (idx == 1) 
            return distribute(new Identity<T1>(), (UnaryFunctor<?,T2>) uf);
        
        return distribute((UnaryFunctor<?,T1>) uf, new Identity<T2>());
    }
} 
