// ============================================================================
// $Id: Functor.java,v 1.5 2009/09/06 22:30:06 davidahall Exp $
// Copyright (c) 2006  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn;

import java.io.Serializable;

import net.sf.jga.fn.adaptor.BindNth;
import net.sf.jga.fn.adaptor.ComposeBinaryNth;
import net.sf.jga.fn.adaptor.ComposeUnaryNth;
import net.sf.jga.fn.adaptor.GenerateNth;

/**
 * A Function object that accepts an arbitrary number of arguments and returns
 * a value of type R.  The Functor interface is generally not typesafe.  It is
 * typesafe with respect to the return type (R), but the compiler cannot ensure
 * that the number of arguments or their types are appropriate for any given
 * implementation of the Functor interface.  However, three direct subclasses,
 * Generator<R>, UnaryFunctor<T,R>, and BindaryFunctor<T1,T2,R> are typesafe,
 * provided their interfaces are used in leiu of this interface.
 * <p>
 * In short, if you want the compiler to enforce type safetype with respect to
 * the number and type of arguments, then do not use the method(s) defined by
 * this class: use the methods defined by the typesafe derived classes.
 * <p>
 * Copyright &copy; 2006  David A. Hall
 */

abstract public class Functor<R> implements Serializable, Visitable {

    /**
     * A non-typesafe evaluation method for function objects.  The number and
     * types of arguments is not checked at compile time: an invalid number of
     * arguments may cause IndexOutOfBoundsException (or cause extraneous
     * arguments to be ignored); an argument of an invalid type may cause a
     * ClassCastException.
     */
    abstract public R eval(Object... args);
    
    /**
     * Binds the given values to this functor, starting from the first argument.
     * @param arg0
     * @param args
     * @return
     */
    public Functor<R> bind0th(Object arg0, Object... args) {
        return new BindNth<R>(this, 0, arg0, args);
    }
    
    /**
     * Binds the given values to this functor, starting from the given index.
     * @param idx
     * @param arg0
     * @param args
     * @return
     */
    public Functor<R> bindNth(int idx, Object arg0, Object... args) {
        return new BindNth<R>(this, idx, arg0, args);
    }

    /**
     * Uses the given generator to insert an argument at the given index
     * @param idx
     * @param arg0
     * @param args
     * @return
     */
    public Functor<R> generateNth(int idx, Generator<?> gen) {
        return new GenerateNth<R>(this, idx, gen);
    }

    /**
     * Uses the given functor to replace an argument at the given index
     * @param idx
     * @param arg0
     * @param args
     * @return
     */
    public Functor<R> composeNth(int idx, UnaryFunctor<?,?> uf) {
        return new ComposeUnaryNth<R>(this, idx, uf);
    }

    /**
     * Uses the given functor to replace two arguments at the given index with a single value
     * @param idx
     * @param arg0
     * @param args
     * @return
     */
    public Functor<R> composeNth(int idx, BinaryFunctor<?,?,?> bf) {
        return new ComposeBinaryNth<R>(this, idx, bf);
    }

    /**
     * default implementation of Visitable interface.
     */
    public void accept(Visitor v) {
        v.visit(this);
    }
}
