// ============================================================================
// $Id: UnaryFunctor.java,v 1.21 2009/09/06 22:30:06 davidahall Exp $
// Copyright (c) 2002-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn;

import java.io.Serializable;

import net.sf.jga.fn.adaptor.Bind;
import net.sf.jga.fn.adaptor.ChainBinary;
import net.sf.jga.fn.adaptor.ChainFunctor;
import net.sf.jga.fn.adaptor.ChainUnary;
import net.sf.jga.fn.adaptor.Generate;

/**
 * A Function Object that takes one argument and returns a result.  The
 * argument is of type <code>T</code>, and the result is of type <code>R</code>.
 * <p>
 * Copyright &copy; 2002-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

abstract public class UnaryFunctor<T, R> extends Functor<R> implements Serializable, Visitable {
    /**
     * Executes the function and returns the result.
     */
    abstract public R fn(T arg);

    /**
     * FactoryMethod that creates a UnaryFunctor that passes its argument to
     * the given functor, and uses the result as the argument to this
     * function. Given argument <b>x</b>, the new functor will return
     * <code>fn<sub>this</sub>(f(x)))</code>
     */
    public <F> UnaryFunctor<F,R> compose(UnaryFunctor<F,T> f) {
        return new ChainUnary<F,T,R>(this,f);
    }

    /**
     * FactoryMethod that creates a BinaryFunctor that passes its arguments to
     * the given functor, and uses the result as the argument to this
     * function. Given arguments <b>x</b> and <b>y</b>, the new functor will
     * return <code>fn<sub>this</sub>(f(x,y)))</code>
     */
    public <F1,F2> BinaryFunctor<F1,F2,R> compose(BinaryFunctor<F1,F2,T> f) {
        return new ChainBinary<F1,F2,T,R>(this, f);
    }

    /**
     * FactoryMethod that creates a BinaryFunctor that passes its arguments to
     * the given functor, and uses the result as the argument to this
     * function. Given arguments <b>x<sub>0</sub>, x<sub>1</sub>, ...</b>, the 
     * new functor will return <code>fn<sub>this</sub>(f(x<sub>0</sub>,x<sub>0</sub>,...)))</code>
     */
    public Functor<R> compose(Functor<T> f) {
        return new ChainFunctor<T,R>(this, f);
    }

    /**
     * FactoryMethod that creates a Generator to create the argument to this
     * function.  The new functor will return
     * <code>fn<sub>this</sub>(fn())</code>.
     */

    public Generator<R> generate(Generator<T> gen) {
        return new Generate<T,R>(this, gen);
    }
    
    /**
     * FactoryMethod that binds the argument arguments to this function to a specific
     * value.  The new functor will return
     * <code>fn<sub>this</sub>(val)</code>.
     */

    public Generator<R> bind(T val) {
        return new Bind<T,R>(val, this);
    }
    
    // -----------------
    // Functor interface
    // -----------------
    
    /**
     * Executes the function and returns the result.
     * @throw ArrayIndexOutOfBoundsExcetpsion if fewer than 2 arguments are passed. 
     */
    @SuppressWarnings("unchecked")
    // This generates an unchecked cast warning: we're crossing the line from
    // unsafe rawform code to typesafe generic code.  This interface is known
    // to be unsafe, and documented to the user as such.        
    public R eval(Object... args) {
        return fn((T) args[0]);
    }
    /**
     * Binds the given values to this functor, starting from the first argument.
     * @param arg0
     * @param args
     * @return
     */
    @Override
    @SuppressWarnings("unchecked")
    // This generates unchecked cast warnings: we're crossing the line from
    // unsafe rawform code to typesafe generic code.   
    public Functor<R> bind0th(Object arg0, Object... args) {
        return bind((T) arg0);
    }
    
    /**
     * Binds the given values to this functor, starting from the given index.
     * @param idx
     * @param arg0
     * @param args
     * @return
     */
    @Override
    @SuppressWarnings("unchecked")
    // This generates unchecked cast warnings: we're crossing the line from
    // unsafe rawform code to typesafe generic code.   
    public Functor<R> bindNth(int idx, Object arg0, Object... args) {
        if (idx < 0) {
            throw new IllegalArgumentException("Index must be non-negative");
        }
            
        if (idx > 0)
            return this; // args bound out of bounds would be ignored anyway
        
        return bind((T) arg0);
    }
    
    /**
     * Generates the argument at the given index to this functor 
     * @param idx
     * @param arg0
     * @param args
     * @return
     */
    @Override
    @SuppressWarnings("unchecked")
    // This generates unchecked cast warnings: we're crossing the line from
    // unsafe rawform code to typesafe generic code.   
    public Functor<R> generateNth(int idx, Generator<?> gen) {
        if (idx < 0) {
            throw new IllegalArgumentException("Index must be non-negative");
        }
            
        if (gen == null) {
            throw new IllegalArgumentException("Must supply a Generator to supply a runtime argument");
        }
        
        if (idx > 0) {
            return this; // args generated out of bounds would be ignored anyway
        }
        
        return generate((Generator<T>) gen);
    }
    
    /**
     * Modifies the argument at the given index to this functor 
     * @param idx
     * @param arg0
     * @param args
     * @return
     */
    @Override
    @SuppressWarnings("unchecked")
    // This generates unchecked cast warnings: we're crossing the line from
    // unsafe rawform code to typesafe generic code.   
    public Functor<R> composeNth(int idx, UnaryFunctor<?,?> uf) {
        if (idx < 0) {
            throw new IllegalArgumentException("Index must be non-negative");
        }
            
        if (uf == null) {
            throw new IllegalArgumentException("Must supply a UnaryFunctor to modify a runtime argument");
        }
        
        if (idx > 0) {
            return this; // args bound out of bounds would be ignored anyway
        }
        
        return compose((UnaryFunctor<?,T>) uf);
    }

    /**
     * Computes the argument at the given index from the pair of arguments starting at the given index 
     * @param idx
     * @param arg0
     * @param args
     * @return
     */
    @Override
    @SuppressWarnings("unchecked")
    // This generates unchecked cast warnings: we're crossing the line from
    // unsafe rawform code to typesafe generic code.   
    public Functor<R> composeNth(int idx, BinaryFunctor<?,?,?> bf) {
        if (idx < 0) {
            throw new IllegalArgumentException("Index must be non-negative");
        }
            
        if (bf == null) {
            throw new IllegalArgumentException("Must supply a BinaryFunctor to modify a runtime argument");
        }
        
        if (idx > 0) {
            return this; // args bound out of bounds would be ignored anyway
        }
        
        return compose((BinaryFunctor<?,?,T>) bf);
    }
} 

