// ============================================================================
// $Id: AdaptorFunctors.java,v 1.3 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2006  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.adaptor;

import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.Generator;
import net.sf.jga.fn.UnaryFunctor;

/**
 * Static factory methods for selected functors in the Adaptor package.
 * <p>
 * Copyright &copy; 2006  David A. Hall
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

public final class AdaptorFunctors {
    
    // Stateless functors for which only one instance is really necessary
    static private Identity<?> _id = new Identity<Object>();

    @SuppressWarnings("unchecked")
    static private Project1st _proj1 = new Project1st();
    
    @SuppressWarnings("unchecked")
    static private Project2nd _proj2 = new Project2nd();

    /**
     * Returns a Boolean Generator that produces TRUE when both input generators return TRUE.
     */
    static public Generator<Boolean> and(Generator<Boolean> g1, Generator<Boolean> g2) {
        return new AndGenerator(g1, g2);
    }
    
    /**
     * Returns a Boolean Functor that produces TRUE when both input functors return TRUE for
     * same given argument.
     */
    static public <T> UnaryFunctor<T,Boolean>
    and(UnaryFunctor<T,Boolean> uf1, UnaryFunctor<T,Boolean> uf2)
    {
        return new AndUnary<T>(uf1, uf2);
    }
    
    /**
     * Returns a Boolean Functor that produces TRUE when both input functors return TRUE for
     * same given pair of arguments.
     */
    static public <T1,T2> BinaryFunctor<T1,T2,Boolean>
    and(BinaryFunctor<T1,T2,Boolean> uf1, BinaryFunctor<T1,T2,Boolean> uf2)
    {
        return new AndBinary<T1,T2>(uf1, uf2);
    }


    /**
     * Returns a generator that evaluates the <tt>test</tt> and, when true, evaluates the
     * <tt>trueFn</tt> and returns the results.  When the <tt>test</tt> is false, the
     * generator returns <tt>null</tt>.
     */
    static public <R> Generator<R> conditional(Generator<Boolean> test, Generator<R> trueFn) {
        return new ConditionalGenerator<R>(test, trueFn, new Constant<R>(null));
    }

    /**
     * Returns a generator that evaluates the <tt>test</tt> and, when true, evaluates the
     * <tt>trueFn</tt> and returns the results.  When the <tt>test</tt> is false, the
     * generator evaluates <tt>falseFn</tt> and the returns the results
     */
    static public <R> Generator<R>
    conditional(Generator<Boolean> test, Generator<R> trueFn, Generator<R> falseFn)
    {
        return new ConditionalGenerator<R>(test, trueFn, falseFn);
    }

    /**
     * Returns a functor that evaluates the <tt>test</tt> and, when true, evaluates the
     * <tt>trueFn</tt> and returns the results.  When the <tt>test</tt> is false, the
     * functor returns <tt>null</tt>.  Both functors receiv the same argument when evaluated.
     */
    static public <T,R> UnaryFunctor<T,R>
    conditional(UnaryFunctor<T,Boolean> test, UnaryFunctor<T,R> trueFn)
    {
        return new ConditionalUnary<T,R>(test, trueFn, new ConstantUnary<T,R>(null));
    }

    /**
     * Returns a functor that evaluates the <tt>test</tt> and, when true, evaluates the
     * <tt>trueFn</tt> and returns the results.  When the <tt>test</tt> is false, the
     * functor evaluates <tt>falseFn</tt> and the returns the results.  Both functors
     * that are evaluated receive the same pair of arguments.
     */
    static public <T,R> UnaryFunctor<T,R>
    conditional(UnaryFunctor<T,Boolean> test, UnaryFunctor<T,R> trueFn, UnaryFunctor<T,R> falseFn)
    {
        return new ConditionalUnary<T,R>(test, trueFn, falseFn);
    }


    /**
     * Returns a functor that evaluates the <tt>test</tt> and, when true, evaluates the
     * <tt>trueFn</tt> and returns the results.  When the <tt>test</tt> is false, the
     * functor returns <tt>null</tt>.  Both functors receive the same pair of arguments when
     * valuated.
     */
    static public <T1,T2,R> BinaryFunctor<T1,T2,R>
    conditional(BinaryFunctor<T1,T2,Boolean> test, BinaryFunctor<T1,T2,R> trueFn)
    {
        return new ConditionalBinary<T1,T2,R>(test, trueFn, new ConstantBinary<T1,T2,R>(null));
    }

    /**
     * Returns a functor that evaluates the <tt>test</tt> and, when true, evaluates the
     * <tt>trueFn</tt> and returns the results.  When the <tt>test</tt> is false, the
     * functor evaluates <tt>falseFn</tt> and the returns the results.  Both functors
     * that are evaluated receive the same pair of arguments.
     */
    static public <T1,T2,R> BinaryFunctor<T1,T2,R>
    conditional(BinaryFunctor<T1,T2,Boolean> test, BinaryFunctor<T1,T2,R> trueFn,
                BinaryFunctor<T1,T2,R> falseFn)
    {
        return new ConditionalBinary<T1,T2,R>(test, trueFn, falseFn);
    }

    
    /**
     */
    static public <R> Generator<R> constant(R value) {
        return new Constant<R>(value);
    }

    
    /**
     */
    static public <T,R> UnaryFunctor<T,R> constantUnary(R value) {
        return new ConstantUnary<T,R>(value);
    }

    
    /**
     */
    static public <T1,T2,R> BinaryFunctor<T1,T2,R> constantBinary(R value) {
        return new ConstantBinary<T1,T2,R>(value);
    }

    
    /**
     */
    @SuppressWarnings("unchecked")
    static public <T> UnaryFunctor<T,T> identity() {
        return (Identity<T>) _id;
    }

   
    /**
     * Returns a Boolean Generator that produces TRUE when either input generators return TRUE.
     */
    static public Generator<Boolean> or(Generator<Boolean> g1, Generator<Boolean> g2) {
        return new OrGenerator(g1, g2);
    }

    
    /**
     * Returns a Boolean Functor that produces TRUE when either input functors return TRUE for
     * same given argument.
     */
    static public <T> UnaryFunctor<T,Boolean>
    or(UnaryFunctor<T,Boolean> uf1, UnaryFunctor<T,Boolean> uf2)
    {
        return new OrUnary<T>(uf1, uf2);
    }

    
    /**
     * Returns a Boolean Functor that produces TRUE when either input functors return TRUE for
     * same given pair of arguments.
     */
    static public <T1,T2> BinaryFunctor<T1,T2,Boolean>
    or(BinaryFunctor<T1,T2,Boolean> uf1, BinaryFunctor<T1,T2,Boolean> uf2)
    {
        return new OrBinary<T1,T2>(uf1, uf2);
    }


    /**
     */
    @SuppressWarnings("unchecked")
    static public <T1,T2> BinaryFunctor<T1,T2,T1> project1st() {
        return _proj1;
    }

    
    
    /**
     */
    @SuppressWarnings("unchecked")
    static public <T1,T2> BinaryFunctor<T1,T2,T2> project2nd() {
        return _proj2;
    }
}

