// ============================================================================
// $Id: BindNth.java,v 1.2 2009/09/06 22:30:06 davidahall Exp $
// Copyright (c) 2009  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.adaptor;

import net.sf.jga.fn.Functor;
import net.sf.jga.util.ArrayUtils;

/**
 * Functor that binds values into the argument list of a nested functor at
 * an arbtirary location.
 * <p>
 * Copyright &copy; 2009  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class BindNth<R> extends Functor<R> {

    private static final long serialVersionUID = -4337766404337539952L;

    // The functor whose arguments are to be bound
    private Functor<R> fn;
    
    // The values to bind to the functors arguments 
    private Object[] constArgs;
    
    // The starting index of the bound arguments
    private int startingIdx;
    
    
    public BindNth(Functor<R> fn, int idx, Object binding0, Object... bindings) {
        if (fn == null) {
            throw new IllegalArgumentException("Must supply a Function whose arguments are to be bound");
        }
        
        if (idx < 0) {
            throw new IllegalArgumentException("Index must be non-negative");
        }
        
        this.fn = fn;
        this.startingIdx = idx;
        if (bindings == null || bindings.length == 0) {
            constArgs = new Object[]{ binding0 };
        }
        else {
            constArgs = new Object[bindings.length + 1];
            constArgs[0] = binding0;
            System.arraycopy(bindings, 0, constArgs, 1, bindings.length);
        }
    }

    /**
     * Returns the results of the functor, binding the constant values given at construction
     * to the arguments starting with the given index.
     * @return fn.eval(args[0], ..., args[i - 1], bindings[0], ..., bindings[bindings.len - 1], args[i], ...)
     * @throws ArrayIndexOutOfBoundsException if the length of the args array is less than the 
     * starting index given at construction.
     */
    public R eval(Object... args) {
        return fn.eval(ArrayUtils.insert(args, startingIdx, constArgs));
    }
    
    /**
     * Calls the Visitor's <code>visit(BindNth)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof BindNth.Visitor)
            ((BindNth.Visitor)v).visit(this);
        else
            v.visit(this);
    }
    
    // Object overrides

    public String toString() {
        return fn+".bindNth("+startingIdx+", "+ constArgs+")";
    }
    
    // AcyclicVisitor
    
    /**x
     * Interface for classes that may interpret an <b>BindNth</b> functor.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(BindNth<?> bf);
    }
}
