// ============================================================================
// $Id: ChainUnary.java,v 1.10 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2002-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.adaptor;

import net.sf.jga.fn.UnaryFunctor;

/**
 * Unary Functor that passes the results of one Unary Functor as the argument
 * to another Unary Functor.  This allows for the construction of compound
 * functors from the primitives found in the arithmetic, logical, property, and
 * comparison packages.
 * <p>
 * Copyright &copy; 2002-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class ChainUnary<T,F,R> extends UnaryFunctor<T,R> {

    static final long serialVersionUID = -6690274081307420708L;

    // the outer fn
    private UnaryFunctor<F,R> _f;

    // the inner fn
    private UnaryFunctor<T,F> _g;
    
    /**
     * Builds a BinaryCompose functor, given the outer functor <b>f</b> the
     * inner functor <b>g</b>.
     * @throws IllegalArgumentException if any of the functors is missing
     */
    public ChainUnary(UnaryFunctor<F,R> f, UnaryFunctor<T,F> g)
    {
        if (f == null || g == null) {
            throw new IllegalArgumentException("Two functors are required");
        }
        
        _f = f;
        _g = g;
    }

    /**
     * Returns the outer functor
     * @return the outer functor
     */
    public UnaryFunctor<F,R> getOuterFunctor() { return _f; }
    
    /**
     * Returns the inner functor
     * @return the inner functor
     */
    public UnaryFunctor<T,F> getInnerFunctor() { return _g; }

    // UnaryFunctor interface
    
    /**
     * Given argument <b>x</b>, passes x to the inner functor, and passes the
     * result to the outer functor.
     * 
     * @return f(g(x))
     */
    public R fn(T x) {
        return _f.fn(_g.fn(x));
    }

    // Object overrides

    public String toString() {
        return _f+".compose("+_g+")";
    }
    
    /**
     * Calls the Visitor's <code>visit(BinaryCompose)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof ChainUnary.Visitor)
            ((ChainUnary.Visitor)v).visit(this);
        else
            v.visit(this);
    }
    
    // AcyclicVisitor
    
    /**
     * Interface for classes that may interpret a <b>ChainUnary</b> functor.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(ChainUnary<?,?,?> bf);
    }
}
