// ============================================================================
// $Id: ComposeBinary.java,v 1.10 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2002-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.adaptor;

import net.sf.jga.fn.BinaryFunctor;

/**
 * Binary Functor that passes the results of two inner Binary Functors as the
 * arguments to an outer Binary Functor.  This allows for the construction of
 * compound functors from the primitives found in the arithmetic, logical,
 * property, and comparison packages.
 * <p>
 * Copyright &copy; 2002-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class ComposeBinary<T1,T2,F1,F2,R> extends BinaryFunctor<T1,T2,R> {

    static final long serialVersionUID = -6427141311027965863L;

    // The first of two inner functors
    private BinaryFunctor<T1,T2,F1> _f;

    // The second of two inner functors
    private BinaryFunctor<T1,T2,F2> _g;

    // The outer functor
    private BinaryFunctor<F1,F2,R> _h;

    /**
     * Builds a ComposeBinary functor, given two inner functors <b>f</b> and
     * <b>g</b>, and outer functor <b>h</b>.
     * @throws IllegalArgumentException if any of the functors is missing
     */
    public ComposeBinary(BinaryFunctor<T1,T2,F1> f,
                         BinaryFunctor<T1,T2,F2> g,
                         BinaryFunctor<F1,F2,R> h)
    {
        if (f == null || g == null || h == null) {
            throw new IllegalArgumentException("Three functors are required");
        }
        
        _f = f; _g =g; _h = h;
    }

    /**
     * Returns the first of two inner functors
     * @return the first of two inner functors
     */
    public BinaryFunctor<T1,T2,F1> getFirstInnerFunctor() { return _f; }

    /**
     * Returns the second of two inner functors
     * @return the second of two inner functors
     */
    public BinaryFunctor<T1,T2,F2> getSecondInnerFunctor() { return _g; }

    /**
     * Returns the outer functor
     * @return the outer functor
     */
    public BinaryFunctor<F1,F2,R> getOuterFunctor() { return _h; }

    // BinaryFunctor interface
    
    /**
     * Given argument <b>x</b>, passes x to both inner functors, and passes the
     * results of those functors to the outer functor.
     * 
     * @return h(f(x,y), g(x,y))
     */
    public R fn(T1 x, T2 y) {
        return _h.fn(_f.fn(x,y), _g.fn(x,y));
    }
    
    /**
     * Calls the Visitor's <code>visit(ComposeBinary)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof ComposeBinary.Visitor)
            ((ComposeBinary.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return _h+".compose("+_f+","+_g+")";
    }
    
    // Acyclic Visitor
    
    /**
     * Interface for classes that may interpret a <b>ComposeBinary</b> functor.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(ComposeBinary<?,?,?,?,?> host);
    }
}
