// ============================================================================
// $Id: ComposeBinaryNth.java,v 1.1 2009/09/06 22:30:06 davidahall Exp $
// Copyright (c) 2009  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.adaptor;

import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.Functor;
import net.sf.jga.util.ArrayUtils;

/**
 * Functor that uses a Generator to produce the Nth argument to a given Functor.
 * The generated result is inserted into the argument list given to this functor
 * with the modified list being passed to the nested functor.
 * <p>
 * Copyright &copy; 2009  David A. Hall
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

public class ComposeBinaryNth<R> extends Functor<R> {
    
    private static final long serialVersionUID = 1371141864350370835L;

    // The Functor whose argument is to be generated
    private Functor<R> fn;
    
    // The nested Functor
    private BinaryFunctor<?,?,?> bf;
    
    // The index of the generated argument
    private int idx;
        
    public ComposeBinaryNth(Functor<R> fn, int idx, BinaryFunctor<?,?,?> bf) {
        if (fn == null) {
            throw new IllegalArgumentException("Must supply a Functor whose argument is to be modified");
        }
        
        if (idx < 0) {
            throw new IllegalArgumentException("Index must be non-negative");
        }
            
        if (bf == null) {
            throw new IllegalArgumentException("Must supply a BinaryFunctor to replace a pair of runtime arguments");
        }
        
        this.fn = fn;
        this.bf = bf;
        this.idx = idx;
    }

    /**
     * Returns the Functor that is invoked
     */

    public Functor<R> getFunctor() { return fn; }

    /**
     * Returns the generator that produces the indicated argument.
     */

    public BinaryFunctor<?,?,?> getBinaryFunctor() { return bf; }
    
    /**
     * Returns the results of the functor, using the nested generator to produce the Nth argument.
     * If there are fewer than N+1 arguments, no modification will be attempted.
     * @return fn.eval(args[0], ..., args[N-1], bf(args[N], args[N+1]), args[N+2], ...)
     */
    public R eval(Object... args) {
        if (idx >= args.length - 1) {
            return fn.eval(args);
        }
        
        // Compute the new argument given the two arguments that it will replace
        Object newArg = bf.eval(args[idx], args[idx+1]);
        
        // Removes one arg position from the array.  newArgs[idx] is now args[idx+1]
        Object[] newArgs = ArrayUtils.remove(args, idx);
        
        // Replaces newArgs[idx] with the result from before
        newArgs[idx] = newArg;
        
        return fn.eval(newArgs);
    }

    /**
     * Calls the Visitor's <code>visit(Generate2nd)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof ComposeBinaryNth.Visitor)
            ((ComposeBinaryNth.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return fn+".composeNth("+idx+":"+bf+")"; 
    }
    
    // Acyclic Visitor
    
    /**
     * Interface for classes that may interpret a <b>Generate2nd</b>
     * functor.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(ComposeBinaryNth<?> host);
    }
}
