// ============================================================================
// $Id: ConditionalBinary.java,v 1.8 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2002-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.adaptor;

import net.sf.jga.fn.BinaryFunctor;

/**
 * BinaryFunctor that tests a condition, executes one of two given functors, and
 * returns the result.  The arguments used to evaluate the condition will also
 * be passed to whichever functor is executed.  This functor implements the
 * traditional <code><b>?:</b></code> operator.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class ConditionalBinary<T1,T2,R> extends BinaryFunctor<T1,T2,R> {

    static final long serialVersionUID = -2682605534388798188L;
    
    // The condition to be evaluated
    private BinaryFunctor<T1,T2,Boolean> _test;

    // The functor to execute when the condition passes
    private BinaryFunctor<T1,T2,R> _then;

    // The functor to execute when the condition fails
    private BinaryFunctor<T1,T2,R> _else;

    /**
     * Builds a ConditionalBinary functor, given the condition to test, and the two
     * functors that may be executed.
     * @throws IllegalArgumentException if any of the functors is missing
     */
    public ConditionalBinary(BinaryFunctor<T1,T2,Boolean> test, BinaryFunctor<T1,T2,R> trueFn,
                       BinaryFunctor<T1,T2,R> falseFn)
    {
        if (test == null || trueFn == null || falseFn == null) {
            throw new IllegalArgumentException("Three functors are required");
        }

        _test = test;
        _then = trueFn;
        _else = falseFn;
    }

    /**
     * Returns the test functors
     * @return the test functors
     */
    public BinaryFunctor<T1,T2,Boolean> getCondition() { return _test; }

    /**
     * Returns the functor that is executed when the condition is true
     * @return the functor that is executed when the condition is true
     */
    public BinaryFunctor<T1,T2,R> getTrueFunctor() { return _then; }

    /**
     * Returns the functor that is executed when the condition is false
     * @return the functor that is executed when the condition is false
     */
    public BinaryFunctor<T1,T2,R> getFalseFunctor() { return _else; }

    // BinaryFunctor interface
    
    /**
     * Given arguments <b>x</b> and <b>x</b> evaluates test(x,y); if true,
     * returns trueFn(x,y), otherwise, returns falseFn(x,y).
     * 
     * @return test(x,y) ? trueFn(x,y) : falseFn(x,y)
     */
    public R fn(T1 x, T2 y) {
        return _test.fn(x,y) ? _then.fn(x,y) : _else.fn(x,y);
    }
    
    /**
     * Calls the Visitor's <code>visit(ConditionalBinary)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof ConditionalBinary.Visitor)
            ((ConditionalBinary.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return "ConditionalBinary["+_test+"?"+_then+":"+_else+"]";
    }
    
    // Acyclic Visitor
    
    /**
     * Interface for classes that may interpret a <b>ConditionalBinary</b> functor.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(ConditionalBinary<?,?,?> host);
    }
}
