// ============================================================================
// $Id: ConditionalUnary.java,v 1.7 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2002-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.adaptor;

import net.sf.jga.fn.UnaryFunctor;

/**
 * UnaryFunctor that tests a condition, executes one of two given functors, and
 * returns the result.  The argument used to evaluate the condition will also
 * be passed to whichever functor is executed.  This functor implements the
 * traditional <code><b>?:</b></code> operator.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class ConditionalUnary<T,R> extends UnaryFunctor<T,R> {

    static final long serialVersionUID = -8509676654602764401L;

    // The condition to be evaluated
    private UnaryFunctor<T,Boolean> _test;

    // The functor to execute when the condition passes
    private UnaryFunctor<T,R> _then;

    // The functor to execute when the condition fails
    private UnaryFunctor<T,R> _else;

    /**
     * Builds a ConditionalUnary functor, given the condition to test, and the two
     * functors that may be executed.
     * @throws IllegalArgumentException if any of the functors is missing
     */
    public ConditionalUnary(UnaryFunctor<T,Boolean> test, UnaryFunctor<T,R> trueFn,
                            UnaryFunctor<T,R> falseFn)
    {
        if (test == null || trueFn == null || falseFn == null) {
            throw new IllegalArgumentException("Three functors are required");
        }

        _test = test;
        _then = trueFn;
        _else = falseFn;
    }

    /**
     * Returns the test functors
     * @return the test functors
     */
    public UnaryFunctor<T,Boolean> getCondition() { return _test; }

    /**
     * Returns the functor that is executed when the condition is true
     * @return the functor that is executed when the condition is true
     */
    public UnaryFunctor<T,R> getTrueFunctor() { return _then; }

    /**
     * Returns the functor that is executed when the condition is false
     * @return the functor that is executed when the condition is false
     */
    public UnaryFunctor<T,R> getFalseFunctor() { return _else; }

    // UnaryFunctor interface
    
    /**
     * Given argument <b>x</b>, evaluates test(x); if true, returns trueFn(x),
     * otherwise, returns falseFn(x).
     * 
     * @return test(x) ? trueFn(x) : falseFn(x)
     */
    public R fn(T x) {
        return _test.fn(x) ? _then.fn(x) : _else.fn(x);
    }
    
    /**
     * Calls the Visitor's <code>visit(ConditionalUnary)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof ConditionalUnary.Visitor)
            ((ConditionalUnary.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return "ConditionalUnary["+_test+"?"+_then+":"+_else+"]";
    }
    
    // Acyclic Visitor
    
    /**
     * Interface for classes that may interpret a <b>ConditionalUnary</b> functor.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(ConditionalUnary<?,?> host);
    }
}
