// ============================================================================
// $Id: Distribute.java,v 1.9 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2002-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.adaptor;

import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.UnaryFunctor;

/**
 * Binary Functor that passes its two arguments to two inner Unary Functors,
 * and uses the results as arguments to an outer Binary Functor.  This allows
 * for the construction of compound functors from the primitives found in the
 * arithmetic, logical, property, and comparison packages.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class Distribute<T1,T2,F1,F2,R> extends BinaryFunctor<T1,T2,R> {

    static final long serialVersionUID = -8288483375404557210L;

    // The functor to which the first arg is passed
    private UnaryFunctor<T1,F1> _f;

    // The functor to which the second arg is passed
    private UnaryFunctor<T2,F2> _g;

    // The functor that evaluates the results of the two inner functors
    private BinaryFunctor<F1,F2,R> _h;

    /**
     * Builds a Distribute functor, given two inner functors <b>f</b> and
     * <b>g</b>, and outer functor <b>h</b>.
     * @throws IllegalArgumentException if any of the functors is missing
     */
    public Distribute(UnaryFunctor<T1,F1> f,
                      UnaryFunctor<T2,F2> g,
                      BinaryFunctor<F1,F2,R> h)
    {
        if (f == null || g == null || h == null) {
            throw new IllegalArgumentException("Three functors are required");
        }
        
        _f = f; _g =g; _h = h;
    }

    /**
     * Returns the first of two inner functors
     * @return the first of two inner functors
     */
    public UnaryFunctor<T1,F1> getFirstInnerFunctor() { return _f; }

    /**
     * Returns the second of two inner functors
     * @return the second of two inner functors
     */
    public UnaryFunctor<T2,F2> getSecondInnerFunctor() { return _g; }

    /**
     * Returns the outer functor
     * @return the outer functor
     */
    public BinaryFunctor<F1,F2,R> getOuterFunctor() { return _h; }

    // BinaryFunctor interface
    
    /**
     * Passes arguments <b>x</b> and <b>y</b> to the corresponding inner
     * functors, and passes the results of those functors to the outer functor.
     * 
     * @return h(f(x), g(y))
     */
    public R fn(T1 x, T2 y) {
        return _h.fn(_f.fn(x), _g.fn(y));
    }
    
    /**
     * Calls the Visitor's <code>visit(Distribute)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof Distribute.Visitor)
            ((Distribute.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return _h+".distribute("+_f+","+_g+")";
    }
    
    // Acyclic Visitor
    
    /**
     * Interface for classes that may interpret a <b>Distribute</b> functor.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(Distribute<?,?,?,?,?> host);
    }
}
