// ============================================================================
// $Id: Identity.java,v 1.11 2009/09/06 22:28:37 davidahall Exp $
// Copyright (c) 2002-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.adaptor;

import net.sf.jga.fn.Generator;
import net.sf.jga.fn.UnaryFunctor;
    
/**
 * Unary Functor that returns the runtime argument.
 * <p>
 * Copyright &copy; 2002-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 */

public class Identity<T> extends UnaryFunctor<T, T> {
    
    static final long serialVersionUID = -1170897535180697025L;

    // the last value passed to fn()
    private transient T _last;
    
    // accessor functor
    private transient Accessor<T> _accessor = new Accessor<T>(this);
    
    /**
     * Returns the last argument that was given, or null if called prior to the
     * first call to fn().  This value does not persist when the functor is
     * serialized.
     * 
     * @return the last argument that was given or null.
     */
    public T arg() {
        return _last;
    }
    
    /**
     * Returns a Generator<T> that will always return the last value passed to
     * this functor.
     */
    public synchronized Generator<T> accessor() { 
        if (_accessor == null)
            _accessor = new Accessor<T>(this);
        
        return _accessor; 
    }
    
    // UnaryFunctor interface
    
    /**
     * Given one argument, returns it.
     *
     * @return x
     */
    public T fn(T x) {
        _last = x;
        return x;
    }

    /**
     * Calls the Visitor's <code>visit(Identity)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof Identity.Visitor)
            ((Identity.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return "Identity";
    }
    
     // AcyclicVisitor
     
    /**
     * Interface for classes that may interpret an <b>Identity</b> functor.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(Identity<?> host);
    }

    /**
     * A Generator that returns the last value passed to an associated Identity.  This allows an
     * Identity to serve as a cache for common subexpressions: if you compose the subexpression
     * within an Identity in it's first appearance in a compound functor, you can use the 
     * Identity's Accessor as a replacement for all other appearances of the same subexpression.
     */
    
    static public class Accessor<T> extends Generator<T> {
        private static final long serialVersionUID = -229232404902638696L;
        
        private Identity<T> id;
        
        private Accessor(Identity<T> id) {
            this.id = id;
        }
        
        public Identity<T> getIdentity() { return id; }
        
        @Override 
        public T fn() { return id.arg(); }
                
        // Object overrides
        @Override 
        public String toString() {
            return id+".Accessor";
        }
        
        /**
         * Calls the Visitor's <code>visit(Constant)</code> method, if it
         * implements the nested Visitor interface.
         */
        public void accept(net.sf.jga.fn.Visitor v) {
            if (v instanceof AccessorVisitor)
                ((AccessorVisitor)v).visit(this);
            else
                v.visit(this);
        }
    }
    
    // AcyclicVisitor for the Accessor functor (interfaces can't be defiend
    // in nested classes).
    
    /**
     * Interface for classes that may interpret an <b>Identity.Accessor</b>.
     */
    public interface AccessorVisitor extends net.sf.jga.fn.Visitor {
        public void visit(Accessor<?> host);
    }
}
        
