// ============================================================================
// $Id: Accumulate.java,v 1.13 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================
package net.sf.jga.fn.algorithm;

import java.util.Iterator;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.UnaryFunctor;

/**
 * Applies a BinaryFunctor to each element in an iteration, and returns the
 * final result.  Each member of the collection is passed to the functor along
 * with the previous result.  If the two arg constructor is used, then the given
 * value is used on the first invocation of the functor: otherwise the first
 * element of the iteration is consumed and passed as the starting value.
 * <p>
 * If the iteration was empty, then the result of this function is the starting
 * value or null if no starting value was given.
 * <p>
 * If no starting value was given, and the iteration has exactly one element,
 * then the element is returned without the BinaryFunctor being used.
 * <p>
 * To Serialize an Accumulate, the generic parameter T must be serializable.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 * @deprecated This functionality is moved into {@link net.sf.jga.algorithms.Summarize} 
 */

public class Accumulate<T> extends UnaryFunctor<Iterator<? extends T>, T> {
    
    static final long serialVersionUID = 4611344190624502921L;
    
    // Functor applied to each element in an iteration in turn
    private BinaryFunctor<T,T,T> _fn;

    // The start value
    private T _value;

    // Flag indicating that the start value was given
    private boolean _givenValue = false;

    /**
     * Builds an Accumulate functor that will use the given functor to process
     * elements in an iteration.  The first element in the iteration will be
     * used as the start value.
     */
    public Accumulate(BinaryFunctor<T,T,T> fn) {
        if (fn == null)
            throw new IllegalArgumentException();
        
        _fn = fn;
    }

    /**
     * Builds an Accumulate functor that will use the given start value and
     * functor to process elements in an iteration.  The first element in the
     * iteration will be used as the start value.
     */
    public Accumulate(T startValue, BinaryFunctor<T,T,T> fn){
        this(fn);
        _value = startValue;
        _givenValue = true;
    }

    /**
     * Returns the functor used to process elements in the iteration.
     */
    public BinaryFunctor<T,T,T> getFunction() {
        return _fn;
    }

    /**
     * Returns the start value, or null if no start value was given.
     */

    public T getStartValue() { return _value; }

    /**
     * Returns true if a start value was passed at construction.
     */

    public boolean hasStartValue() { return _givenValue; }

    /**
     * Apply the functor to the elements of the iteration and return the final
     * result.  Results do not accumulate from one invocation to the next: each
     * time this method is called, the accumulation starts over with the given
     * start value.
     */
    public T fn(Iterator<? extends T> iterator) {
        T value = _givenValue ? _value :
            iterator.hasNext() ? iterator.next() : null;
        
        while (iterator.hasNext()) {
            value = _fn.fn(value, iterator.next());
        }

        return value;
    }
    
    /**
     * Calls the Visitor's <code>visit(Accumulate)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof Accumulate.Visitor)
            ((Accumulate.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return "Accumulate";
    }
    
    // AcyclicVisitor
    
    /**
     * Interface for classes that may interpret an <b>Accumulate</b> functor.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(Accumulate<?> host);
    }
}
