// ============================================================================
// $Id: ElementOf.java,v 1.14 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================
package net.sf.jga.fn.algorithm;

import java.util.Collection;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.util.FindIterator;

/**
 * BinaryPredicate that returns true if a given value is an element of a
 * given collection.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

// The collection extends T because we're reading from it, and this allows the
// objects read to be assumed to be T's.  You might expect a collection of ?superT,
// (ie, a collection that we know can hold T's), but the erasure in that case is
// Object (not T) so comparison using functors won't work.

public class ElementOf<T> extends BinaryFunctor<T,Collection<? extends T>,Boolean> {
    
    static final long serialVersionUID = 4639100512962835854L;
    
    // The functor used to compare the element to the items in the collection
    private BinaryFunctor<T,T,Boolean> _eq;

    /**
     * Builds an ElementOf predicate that will use the collection's built in
     * contains() method.  This form is potentially more efficient than the
     * other constructed form, if the collection passed at evaluation
     * implements contains() using an algorithm more efficient than a linear
     * search.
     */
    public ElementOf() {}
    
    /**
     * Builds an ElementOf predicate that will use the given functor to
     * determine collection membership.  The collection will be searched
     * sequentially for the first element for which the functor returns true.
     */
    public ElementOf(BinaryFunctor<T,T,Boolean> eq) {
        _eq = eq;
    }

    /**
     * Returns the (possibly null) functor used to compare a value to the
     * contents of a collection.
     */
    public BinaryFunctor<T,T,Boolean> getComparisonFn() {
        return _eq;
    }

    /**
     * Return true if the given value is an element of the collection
     */
    public Boolean fn(T value, Collection<? extends T> collection) {
        if (_eq == null)
            return collection.contains(value);
        else {
            FindIterator<T> finder = new FindIterator<T>(collection.iterator());
            UnaryFunctor<T,Boolean> uf = _eq.bind2nd(value);
            return finder.findNext(uf);
        }
    }
    
    /**
     * Calls the Visitor's <code>visit(ElementOf)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof ElementOf.Visitor)
            ((ElementOf.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return "ElementOf";
    }
    
    // AcyclicVisitor
    
    /**
     * Interface for classes that may interpret an <b>ElementOf</b> functor.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(ElementOf<?> host);
    }
}
