// ============================================================================
// $Id: ArithmeticFactory.java,v 1.10 2009/06/27 23:39:42 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.arithmetic;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.HashMap;

/**
 * Builds and distributes implementations of the Arithmetic and
 * IntegerArithmetic interfaces that are available for supported Number
 * classes.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 * 
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 */

public class ArithmeticFactory {
    static private HashMap<Class<? extends Number>,Arithmetic<? extends Number>>
        _arithmeticMap = new HashMap<Class<? extends Number>,Arithmetic<? extends Number>>();
    
    static private HashMap<Class<? extends Number>,IntegerArithmetic<? extends Number>>
        _integralMap = new HashMap<Class<? extends Number>,IntegerArithmetic<? extends Number>>();

    /**
     * Registers the Arithmetic implementation for the given class
     */
    static public <T extends Number> void register (Class<T> c, Arithmetic<T> math) {
        _arithmeticMap.put(c, math);
    }
    
    /**
     * Registers the IntegerArithmetic implementation for the given class
     */
    static public <T extends Number> void register (Class<T> c, IntegerArithmetic<T> math) {
        _arithmeticMap.put(c, math);
        _integralMap.put(c, math);
    }

    /**
     * Returns the Arithmetic implementation registered for the given class
     * @return the Arithmetic implementation registered for the given class
     */
    @SuppressWarnings("unchecked")
    // the map is private, and the register methods ensure that the specific
    // runtime type of each implementation is associated with the correct
    // class.
    static public <T extends Number> Arithmetic<T> getArithmetic (Class<T> c) {
        return (Arithmetic<T>) _arithmeticMap.get(c);
    }
    
    /**
     * Returns the IntegerArithmetic implementation registered for the given
     * class
     * @return the IntegerArithmetic implementation registered for the given
     * class
     */
    @SuppressWarnings("unchecked")
    // the map is private, and the register methods ensure that the specific
    // runtime type of each implementation is associated with the correct
    // class.
    static public <T extends Number> IntegerArithmetic<T> getIntegralArithmetic (Class<T> c) {
        return (IntegerArithmetic<T>) _integralMap.get(c);
    }
    
    static {
        ByteMath bmath = new ByteMath();
        register(Byte.class,       bmath);
        register(Byte.TYPE,        bmath);
        
        ShortMath smath = new ShortMath();
        register(Short.class,      smath);
        register(Short.TYPE,       smath);
        
        IntegerMath imath = new IntegerMath();
        register(Integer.class,    imath);
        register(Integer.TYPE,     imath);

        LongMath lmath = new LongMath();
        register(Long.class,       lmath);
        register(Long.TYPE,        lmath);
        
        FloatMath fmath = new FloatMath();
        register(Float.class,      fmath);
        register(Float.TYPE,       fmath);
        
        DoubleMath dmath = new DoubleMath();
        register(Double.class,     dmath);
        register(Double.TYPE,      dmath);
        
        register(BigDecimal.class, new DecimalMath());
        register(BigInteger.class, new BigIntMath());

    }
}
