// ============================================================================
// $Id: ShiftLeft.java,v 1.8 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.arithmetic;

import java.text.MessageFormat;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.EvaluationException;

/**
 * Returns the first argument shifted left by the number of places given in the
 * second argument.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 * 
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 */

public class ShiftLeft<T extends Number> extends BinaryFunctor<T,Integer,T> {
    
    static final long serialVersionUID = -3242117744551182384L;
    
    // Utility class that performs the bitwise shift as appropriate for the
    // class given at construction
    private IntegerArithmetic<T> _math;
     
    // The type of operands this instance supports
    private Class<T> _type;
     
    /**
     * Builds ShiftLeft functor for the given class.  The class argument must
     * be the same as the generic class argument (when generics are in use)
     * or else a ClassCastException will be thrown when the functor is used.
     *
     * @throws IllegalArgumentException if the given class has no Arithmetic
     *      implementation registered with the ArithmeticFactory
     */
    public ShiftLeft(Class<T> c) {
        _type = c;
        getMath();
    }
    
    /**
     * Returns the type of operands this instance supports
     */
    public Class<T> getType() {
        return _type;
    }
    
    // Binary Functor interface
    
    /**
     * Given integral arguments <b>x</b> and <b>y</b>, returns x &lt;&lt; y
     * @return x &lt;&lt; y
     */
    public T fn(T x, Integer y) {
        try {
            return getMath().shiftLeft(x, y);
        }
        catch (ClassCastException ex) {
            String msg = "ClassCastException: Cannot compute shiftLeft of {0}[{1}] and {2}[{3}]";
            String err = MessageFormat.format(msg, new Object[]{ x.getClass(), x, y.getClass(), y});
            throw new EvaluationException(err, ex);
        }
    }
    
    /**
     * Calls the Visitor's <code>visit(ShiftLeft)</code> method, if it implements
     * the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof ShiftLeft.Visitor)
            ((ShiftLeft.Visitor)v).visit(this);
        else
            v.visit(this);
    }
    
    /**
     */
    private IntegerArithmetic<T> getMath() {
        if (_math == null) {
            _math = ArithmeticFactory.getIntegralArithmetic(_type);
            if (_math == null) {
                String msg = "No implementation of Arithmetic registered for {0}";
                throw new IllegalArgumentException(MessageFormat.format(msg, new Object[]{_type}));
            }
        }

        return _math;
    }
    
    // Object overrides

    public String toString() {
        return "ShiftLeft";
    }
    
    // AcyclicVisitor
    
    /**
     * Interface for classes that may interpret a <b>ShiftLeft</b> functor.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(ShiftLeft<? extends Number> host);
    }
}
