// ============================================================================
// $Id: EqualTo.java,v 1.20 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2002-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.comparison;

import java.util.Comparator;

/**
 * Binary Predicate that returns TRUE for object arguments <b>x</b> and
 * <b>y</b> when x == y using the built-in equals() method or an optional
 * Comparator given at construction. By default, this functor will not throw
 * NullPointerException: it will return true if both runtime arguments are null
 * but false if only one is null.  If an optional comparator is used, then its
 * implementation will determine if a NullPointerException is thrown when passed
 * a null argument.
 * <p>
 * To serialize an Equal functor, the comparator passed at construction(if any)
 * must be Serializable.
 * <p>
 * Copyright &copy; 2002-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class EqualTo<T> extends Equality<T> {

    static final long serialVersionUID = -8880072682296106379L;

    // An optional comparator used to compare values
    Comparator<? super T> _comp;

    /**
     * Builds an EqualTo that uses the built-in equals method.
     */
    public EqualTo() {}
    
    /**
     * Builds an EqualTo that uses the given comparator.
     */
    public EqualTo(Comparator<? super T> comp) {
        _comp = comp;
    }
    
    /**
     * Returns the comparator in use by this functor, if any.
     * @return the comparator in use by this functor, if any.
     */
    public Comparator<? super T> getComparator() { return _comp; }
    
    // BinaryFunctor interface
    
    /**
     * Given arguments <b>x</b> and <b>y</b>, returns x.equals(y).  Will not
     * throw NullPointerException if either x or y are null: will return true
     * if both x and y are null and false if either x or y are null.
     * 
     * @return x.equals(y)
     */
    public Boolean fn(T x, T y) {
        if (_comp == null)
            return (x == null) ? y == null : x.equals(y);
        else
            return _comp.compare(x, y) == 0;
    }

    /**
     * Calls the Visitor's <code>visit(EqualTo)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof EqualTo.Visitor)
            ((EqualTo.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return "EqualTo";
    }
    
    // Acyclic Visitor
    
    /**
     * Interface for classes that may interpret an <b>EqualTo</b> predicate.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(EqualTo<?> host);
    }
}
