// ============================================================================
// $Id: All.java,v 1.20 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2002-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.logical;

import java.util.Collection;
import java.util.Iterator;
import java.util.ArrayList;
import net.sf.jga.fn.UnaryFunctor;

/**
 * Unary Predicate that returns true when each of 0 or more branch predicates
 * returns true.  When the collection of branch predicates is empty, an All
 * predicate will return true (somewhat arbitrarily).  This predicate will
 * short circuit: once one of the branches returns false, none of the
 * subsequent branches will be evaluated.
 * <p>
 * The order of evaluation is dependant on the type of collection used: when
 * using the default constructor, the collection used is a list, and branch
 * predicates will be evaluated in the order given.
 * <p>
 * Copyright &copy; 2002-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class All<T> extends UnaryFunctor<T,Boolean> {
    
    static final long serialVersionUID = 545530729045853706L;
    
    private Collection<UnaryFunctor<T,Boolean>> _branches;

    /**
     * Builds the All predicate with an empty default collection of branch
     * predicates.  The default collection is a list, and branches will be
     * evaluated in the order they are added.
     */
    public All() {
         _branches = new ArrayList<UnaryFunctor<T,Boolean>>();
    }

    /**
     * Builds the All predicate with the given collection of branch predicates.
     * More predicates may be added to the collection after construction.  The
     * order of evaluation of the branch predicates is determined by the
     * collection in use.
     */
    public All(Collection<UnaryFunctor<T,Boolean>> branches) {
        _branches = (branches != null) ? branches : new ArrayList<UnaryFunctor<T,Boolean>>();
    }
     
    /**
     * Returns an Iterator over the branch predicates.
     * @return an Iterator over the branch predicates
     */
    public Iterator<UnaryFunctor<T,Boolean>> branches() {
        return _branches.iterator();
    }

    // UnaryPredicate interface
    
    /**
     * Given arguments <b>x</b>, returns true if all branch predicates return
     * true when given x.  Also returns true when there are no branch
     * predicates.
     * 
     * @return true if all branch predicates return true, false otherwise
     */
    public Boolean fn(T x) {
        for (UnaryFunctor<T,Boolean> pred : _branches) {
            if (! pred.fn(x))
                return false;
        }
        
        return true;
    }
    
    /**
     * Calls the Visitor's <code>visit(All)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof All.Visitor)
            ((All.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        StringBuffer buf = new StringBuffer("All(");
        String sep = "";
        for (UnaryFunctor<T,Boolean> pred : _branches) {
            buf.append(sep).append(pred);
            sep=",";
        }
        buf.append(")");
        return buf.toString();
    }
    
    // AcyclicVisitor
    
    /**
     * Interface for classes that may interpret an <b>All</b> predicate.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(All<?> host);
    }
}
