// ============================================================================
// $Id: CompareProperty.java,v 1.17 2009/09/06 22:36:08 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.property;

import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.fn.comparison.EqualTo;

/**
 * Unary Predicate that compares the value of the named property to the
 * given value.  The type of comparison is a binary predicate: the two
 * arguments passed to the predicate are the value of the argument's property
 * in the first position and the constant value passed at construction in the
 * second position.
 * <p>
 * The test returns
 * <code>bp(getProperty(name).fn(arg), value)</code>.  This is also
 * equivalent to the following:<br>
 * <pre>
 * UnaryPredicate CompareProperty =
 *     new UnaryCompose(new Binder2nd(bp, value),
 *                      new GetProperty(propName))
 * </pre>
 * with one less call to <code>fn</code> at evaluation and somewhat clearer
 * construction syntax.
 * <p>
 * To Serialize a CompareProperty, the generic parameter V must be serializable.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class CompareProperty<T,V> extends UnaryFunctor<T,Boolean> {
    
    static final long serialVersionUID = 8734296101969960336L;

    /**
     * Functor used to compare the value of the property
     */
    private BinaryFunctor<V,V,Boolean> _comp;
    
    /**
     * Functor used to retrieve the value of the property
     */
    private GetProperty<T,V> _gpf;
    
    /**
     * name of the property
     */
    private String _prop;
    
    /**
     * constant value compared to property
     */
    private V _value;

    /**
     * Builds the CompareProperty predicate that will compare the named
     * property of an instance of type argType to the given value using
     * an EqualTo predicate.
     */
    public CompareProperty(Class<T> argType, String propName, V val) {
        this(argType, propName, new EqualTo<V>(), val);
    }
    
    /**
     * Builds the CompareProperty predicate that will compare the named
     * property of an instance of type argType to the given value.  The
     * comparison can be any type of BinaryFunctor where both arguments
     * are of the same type.
     */
    public CompareProperty(Class<T> argType, String propName, BinaryFunctor<V,V,Boolean> pred, V val){
        _prop = propName;
        _comp = pred;
        _value = val;
        _gpf = new GetProperty<T, V>(argType, propName);
    }

    // TODO:
    // public CompareProperty(Class<T> argType, String propName, UnaryFunctor<V,Boolean> pred)
    
    /**
     * Returns the name of the property to be compared
     * @return the name of the property to be compared
     */
    public String getPropName() { return _prop; }
    
    /**
     * Returns the constant value to which properties are compared
     * @return the constant value to which properties are compared
     */
    public V getValue() { return _value; }
    
    /**
     * Returns the predicate used to compare property values
     * @return the predicate used to compare property values.
     */
    public BinaryFunctor<V,V,Boolean> getPredicate() {
        return _comp;
    }
    
    // UnaryPredicate interface
    
    /**
     * Tests the designated property of the argument against the value given at
     * construction.
     * <p>
     * @return the boolean value of the comparison
     */
    public Boolean fn(T arg) {
        return _comp.fn(_gpf.fn(arg), _value);
    }

    /**
     * Calls the Visitor's <code>visit(CompareProperty)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof CompareProperty.Visitor)
            ((CompareProperty.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return "CompareProperty["+_prop+" "+_comp+" "+_value+"]";
    }
    
    // Acyclic Visitor
    
    /**
     * Interface for classes that may interpret a <b>CompareProperty</b>
     * predicate.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(CompareProperty<?,?> host);
    }
}
    


