// ============================================================================
// $Id: Construct.java,v 1.12 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.property;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.text.MessageFormat;
import net.sf.jga.fn.EvaluationException;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.util.ArrayUtils;

/**
 * Unary Functor that constructs an object of the given class.  The functor
 * argument is an array containing the arguments to the class' constructor.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class Construct<R> extends UnaryFunctor<Object[],R> {

    static final long serialVersionUID = -4682718030847292345L;

    // the type of the object built by the constructor
    private Class<R> _objclass;

    // the argument types for the constructor
    private Class<?>[] _argclasses;

    // the constructor being invoked
    private transient Constructor<R> _ctor;

    
    /**
     * Builds a functor that will build an object of class
     * <code>ctorclass</code>, given arguments of classes in
     * <code>argclasses</code>.   
     * @throws IllegalArgumentException if either argument is omitted, or if the
     * constructor cannot be found.
     */
    public Construct(Class<R> ctorclass, Class<?>... argclasses) {
        this(argclasses, ctorclass);
    }
    
    /**
     * Builds a functor that will build an object of class
     * <code>ctorclass</code>, given arguments of classes in
     * <code>argclasses</code>.   
     * @throws IllegalArgumentException if either argument is omitted, or if the
     * constructor cannot be found.
     */
    public Construct(Class<?>[] argclasses, Class<R> ctorclass) {
        if (argclasses == null) {
            String msg = "Argument Classes must be specified";
            throw new IllegalArgumentException(msg);
        }

        if (ctorclass == null) {
            String msg = "Class to be constructed must be specified";
            throw new IllegalArgumentException(msg);
        }

        _objclass = ctorclass;
        _argclasses = argclasses;
        
        try {
            _ctor = ctorclass.getConstructor(argclasses);
        }
        catch (NoSuchMethodException x) {
            String msg = "No constructor for class {0} takes argument(s) of type(s) {1}";
            Object[] msgargs = new Object[]{_objclass.getName(), ArrayUtils.toString(_argclasses)};
            IllegalArgumentException x1 =
                new IllegalArgumentException(MessageFormat.format(msg,msgargs));
            x1.initCause(x);
            throw x1;
        }
    }

    /**
     * Builds a functor that will build an object using the given constructor.
     * @throws IllegalArgumentException if the constructor is omitted.
     */
    public Construct(Constructor<R> ctor) {
        if (ctor == null) {
            String msg = "constructor must be specified";
            throw new IllegalArgumentException(msg);
        }

        _ctor = ctor;
        _objclass = ctor.getDeclaringClass();
        _argclasses = ctor.getParameterTypes();
    }
    
    /**
     * Lazy loads the constructor (used if the functor is called after
     * deserialization)
     */
    public synchronized Constructor<R> getConstructor() throws NoSuchMethodException {
        if (_ctor == null) 
            _ctor = _objclass.getConstructor(_argclasses);

        return _ctor;
    }

    // UnaryFunctor interface
    
    /**
     * Builds an object via the designated constructor, passing the given
     * array of argument values.
     * <p>
     * @return the object built by the constructor
     */
    public R fn(Object[] args) {
        try {
            return (R) getConstructor().newInstance(args);
        }
        catch (NoSuchMethodException x) {
            String msg = "No constructor for class {0} takes argument(s) of type(s) {1}";
            Object[] msgargs = new Object[]{_objclass.getName(), ArrayUtils.toString(_argclasses)};
            throw new EvaluationException(MessageFormat.format(msg,msgargs), x);
        }
        catch (InstantiationException x) {
            String msg = "class {0} is abstract: cannot be constructed";
            Object[] msgargs = new Object[]{_objclass.getName()};
            throw new EvaluationException(MessageFormat.format(msg,msgargs), x);
        }
        catch (IllegalAccessException x) {
            String msg = "class {0} ctor({1}) is not accessible";
            Object[] msgargs = new Object[]{_objclass.getName(), ArrayUtils.toString(_argclasses)};
            throw new EvaluationException(MessageFormat.format(msg,msgargs), x);
        }
        catch (InvocationTargetException x) { 
            String msg = "class {0} ctor({1}) failed: "+x.getMessage();
            Object[] msgargs = new Object[]{_objclass.getName(), ArrayUtils.toString(_argclasses)};
            throw new EvaluationException(MessageFormat.format(msg,msgargs), x);
        }
        catch (ClassCastException x) {
            String msg = "ClassCastException: " +_objclass +"("+ArrayUtils.toString(args)+")";
            throw new EvaluationException(msg, x);
        }
        catch (IllegalArgumentException x) {
            String msg = "class {0} ctor({1}) cannot be called with {2}";
            Object[] msgargs = new Object[]{_objclass.getName(),  ArrayUtils.toString(_argclasses),
                                            ArrayUtils.toString(args)};
            throw new EvaluationException(MessageFormat.format(msg,msgargs), x);
        }
    }

    /**
     * Calls the Visitor's <code>visit(Construct)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof Construct.Visitor)
            ((Construct.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return "Construct("+_objclass.getName()+".getConstructor("+ArrayUtils.toString(_argclasses)+"))";
    }
    
    // Acyclic Visitor
    
    /**
     * Interface for classes that may interpret a <b>Construct</b>
     * predicate.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(Construct<?> host);
    }
}
