// ============================================================================
// $Id: ConstructDefault.java,v 1.11 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2004-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.property;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.text.MessageFormat;
import net.sf.jga.fn.EvaluationException;
import net.sf.jga.fn.Generator;

/**
 * Generator that constructs an object of the given class using its default
 * constructor.
 * <p>
 * Copyright &copy; 2004-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class ConstructDefault<R> extends Generator<R> {

   static final long serialVersionUID = 9150185598879755334L;

    private Class<R> _objclass;
    
    private transient Constructor<R> _ctor;

    /**
     * Builds a generator that will return instances of the class
     * <code>ctorclass</code>, built using the default constructor.   
     * @throws IllegalArgumentException if the constructor cannot be found
     * @throws IllegalArgumentException if the class argument is omitted or does not
     * define an accessible default constructor.
     */
    
    public ConstructDefault(Class<R> ctorclass) {
        if (ctorclass == null) {
            String msg = "Class to be constructed must be specified";
            throw new IllegalArgumentException(msg);
        }
        
        Class<?>[] argclasses = new Class[0];

        _objclass = ctorclass;
        
        try {
            _ctor = ctorclass.getConstructor(argclasses);
        }
        catch (NoSuchMethodException x) {
            String msg = "No default constructor for class {0}";
            Object[] args = new Object[]{_objclass.getName()};
            IllegalArgumentException x1 =
                new IllegalArgumentException(MessageFormat.format(msg,args));
            x1.initCause(x);
            throw x1;
        }
    }

    /**
     * Lazily loads the constructor
     */
    public synchronized Constructor<R> getConstructor() throws NoSuchMethodException{
        if (_ctor == null)
            _ctor = _objclass.getConstructor(new Class[0]);

        return _ctor;
    }

    /**
     * Builds an object via the default constructor
     * <p>
     * @return the object built by the constructor
     */
    public R fn() {
        try {
            R val = (R) getConstructor().newInstance(new Object[0]);
            return val;
        }
        catch (NoSuchMethodException x) { 
            String msg = "No default constructor for class {0}";
            Object[] args = new Object[]{_objclass.getName()};
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
        catch (InstantiationException x) { 
            String msg = "class {0} is abstract: cannot be constructed";
            Object[] args = new Object[]{_objclass.getName()};
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
        catch (IllegalAccessException x) { 
            String msg = "default ctor in class {0} is not accessible";
            Object[] args = new Object[]{_objclass.getName()};
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
        catch (InvocationTargetException x) { 
            String msg = "default ctor in class {0} failed: "+x.getMessage();
            Object[] args = new Object[]{_objclass.getName()};
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
    }
    
    /**
     * Calls the Visitor's <code>visit(ConstructDefault)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof ConstructDefault.Visitor)
            ((ConstructDefault.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return "new "+_objclass.getName()+"()";
    }
    
    // Acyclic Visitor
    
    /**
     * Interface for classes that may interpret a <b>ConstructDefault</b>
     * functor.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(ConstructDefault<?> host);
    }
}
