// ============================================================================
// $Id: ConstructUnary.java,v 1.17 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.property;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.text.MessageFormat;
import net.sf.jga.fn.EvaluationException;
import net.sf.jga.fn.UnaryFunctor;

/**
 * Unary Functor that constructs an object of the given class via a
 * one-argument constructor.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class ConstructUnary<T,R> extends UnaryFunctor<T,R> {
    
    static final long serialVersionUID = -6046849006655934333L;

    // the type of the object built by the constructor
    private Class<R> _objclass;

    // the argument type for the constructor
    private Class<T> _argclass;

    // the constructor being invoked
    private transient Constructor<R> _ctor;

    /**
     * Builds a functor that will build an object of class
     * <code>ctorclass</code>, given an argument of class <code>argclass</code>.
     * The classes passed to the constructor must be assignable to the generic
     * arguments (when generics are in use):
     * <code>
     *    ConstructUnary<String,Integer> ctor =
     *         new ConstructUnary<String,Integer>(String.class,Integer.class);
     * </code>
     * @throws IllegalArgumentException if either argument is omitted, or if the
     * constructor cannot be found.
     */
    
    public ConstructUnary(Class<T> argclass, Class<R> ctorclass) {
        if (argclass == null) {
            String msg = "Argument Class must be specified";
            throw new IllegalArgumentException(msg);
        }

        if (ctorclass == null) {
            String msg = "Class to be constructed must be specified";
            throw new IllegalArgumentException(msg);
        }
        
        try {
            _argclass = argclass;
            _objclass = ctorclass;
            _ctor = getConstructor();
        }
        catch (NoSuchMethodException x) {
            String msg = "No constructor found for class {0} that takes an argument of class {1}";
            Object args[] = new Object[] { ctorclass, argclass };
            
            IllegalArgumentException x1 =
                new IllegalArgumentException(MessageFormat.format(msg, args));
            x1.initCause(x);
            throw x1;
        }
    }

    
    /**
     * Lazily loads the constructor
     */
    public synchronized Constructor<R> getConstructor() throws NoSuchMethodException {
        if (_ctor == null) 
            _ctor = _objclass.getConstructor(new Class[]{_argclass});

        return _ctor;
    }

    
    // UnaryFunctor interface
    
    /**
     * Builds an object via a one-arg constructor, passing the given value.
     * <p>
     * @return the object built by the constructor
     */
    public R fn(T arg) {
        
        try {
            R val = (R) getConstructor().newInstance(new Object[]{arg});
            return val;
        }
        catch (NoSuchMethodException x) {
            String msg = "No constructor for class {0} that takes an argument of type {1}";
            Object[] args = new Object[]{_objclass.getName(), _argclass.getName()};
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
        catch (InstantiationException x) {
            String msg = "class {0} is abstract: cannot be constructed";
            Object[] args = new Object[]{_objclass.getName()};
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
        catch (IllegalArgumentException x) {
            String msg = "class {0} ctor({1}) cannot be called with {2}";
            Object[] args = new Object[]{_objclass.getName(), _argclass.getName(), arg};
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
        catch (IllegalAccessException x) {
            String msg = "class {0} ctor({1}) is not accessible";
            Object[] args = new Object[]{_objclass.getName(), _argclass.getName()};
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
        catch (InvocationTargetException x) { 
            String msg = "class {0} ctor({1}) failed: "+x.getMessage();
            Object[] args = new Object[]{_objclass.getName(), _argclass.getName()};
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
    }

    
    /**
     * Calls the Visitor's <code>visit(ConstructUnary)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof ConstructUnary.Visitor)
            ((ConstructUnary.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return "new "+_objclass.getName()+"(["+_argclass.getName()+"])";
    }
    
    // Acyclic Visitor
    
    /**
     * Interface for classes that may interpret a <b>ConstructUnary</b>
     * predicate.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(ConstructUnary<?,?> host);
    }
}
