// ============================================================================
// $Id: GetField.java,v 1.12 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2002-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.property;

import java.lang.reflect.Field;
import java.text.MessageFormat;
import net.sf.jga.fn.EvaluationException;
import net.sf.jga.fn.UnaryFunctor;

/**
 * Unary Functor that returns the value of the named field for the argument.
 * The field name and type are set at construction.
 * <p>
 * Copyright &copy; 2002-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class GetField<T,R> extends UnaryFunctor<T,R> {
    
    static final long serialVersionUID = 5733484457689881481L;

    // The name of the field to be retrieved.
    private String _fieldName;

    // The class from which the field is retrieved.
    private Class<T> _argtype;
    
    // The Field whose value is is returned
    private transient Field _field;

    /**
     * Builds a GetField for the given field
     * @throws IllegalArgumentException if the field is null, or is not a field
     * of the given class
     */
    @SuppressWarnings("unchecked")
    // There's nothing we can do about this other than warn the users
    // to make sure that they don't use an inappropriate argument type
    public GetField(Field field) {
        if (field == null) {
            throw new IllegalArgumentException("Must supply field");
        }

        _field = field;
        _fieldName = field.getName();
        _argtype = (Class<T>) field.getDeclaringClass();
    }
    
    /**
     * Builds a GetField for the given field
     * @throws IllegalArgumentException if the field is null, or is not a field
     * of the given class
     */
    public GetField(Class<T> argclass, Field field) {
        if (field == null) {
            throw new IllegalArgumentException("Must supply field");
        }

        if (!field.getDeclaringClass().isAssignableFrom(argclass)) {
            throw new IllegalArgumentException(buildNoSuchFieldMessage(field.getName(), argclass));
        }
        
        _field = field;
        _fieldName = field.getName();
        _argtype = argclass;
    }

    /**
     * Builds a GetField for the given field, using the given class to
     * find the desired field.
     * @throws IllegalArgumentException if the field name is null or empty,
     *    
     */
    public GetField(Class<T> argclass, String fieldName) {
        if (fieldName == null || fieldName.length() == 0) {
            throw new IllegalArgumentException("Must supply field name");
        }

        _fieldName = fieldName;
        _argtype = argclass;
        
        try {
            _field = argclass.getField(_fieldName);
        }
        catch (NoSuchFieldException x) {
            IllegalArgumentException iax =
                new IllegalArgumentException(buildNoSuchFieldMessage(fieldName, argclass));
            
            iax.initCause(x);
            throw iax;
        }
    }

    /**
     * Returns the class on which the field is retrieved.
     */
    // UNTESTED
    public Class<T> getObjectType() {
        return _argtype;
    }

    /**
     * Returns the name of the field that this functor retrieves.
     */
    public String getFieldName() {
        return _fieldName;
    }

    /**
     * Returns the type of field that this functor retrieves.  
     */
    @SuppressWarnings("unchecked")
    // There's nothing we can do about this other than warn the users
    // to make sure that they don't use an inappropriate argument type
    public Class<R> getFieldType() {
        return (Class<R>) _field.getType();
    }

    
    /**
     * Lazily loads the Field object.
     */
    
    // Instances of Reflection classes are not serializable, so this must be loaded lazily
    // in case this object was read from a stream (the first time the functor is invoked,
    // the field object will be instantiated)
    
    public synchronized Field getField() throws NoSuchFieldException {
        if (_field == null) 
            _field = _argtype.getField(_fieldName);

        return _field;
    }

    // Unary interface

    /**
     * Returns the value of the designated field of the argument
     * @return the value of the designated field of the argument
     * @throws EvaluationException if the argument does not have the designated
     *         public field, or if it is not of the correct type. 
     */
    @SuppressWarnings("unchecked")
    // There's nothing we can do about this other than warn the users
    // to make sure that they don't use an inappropriate return type
    public R fn(T arg) {
        try {
            R val = (R) getField().get(arg);
            return val; 
        }
        catch (NoSuchFieldException x) {
            throw new EvaluationException(buildNoSuchFieldMessage(_fieldName, _argtype));
        }
        catch (ClassCastException x) { 
            String msg = "{0}.{1} is of type {2}";
            Object[] args = new Object[]{ arg.getClass().getName(), _field.getName(),
                                          _field.getType().getName() };
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
        catch (IllegalAccessException x) {
            String msg = "{0}.{1} is not accessible";
            Object[] args = new Object[]{ _argtype.getName(), _fieldName};
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
    }


    /**
     */
    private String buildNoSuchFieldMessage(String fieldname, Class<?> clasz) {
        String msg = "class {0} does not have field {1}";
        Object[] args = new Object[]{ clasz.getName(), fieldname };
        return MessageFormat.format(msg,args);
    }

    /**
     * Calls the Visitor's <code>visit(GetField)</code> field, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof GetField.Visitor)
            ((GetField.Visitor)v).visit(this);
        else
            v.visit(this);
    }
    
    // Object overrides

    public String toString() {
        return "GetField["+_argtype.getName()+"."+_fieldName+"]";
    }
    
   // AcyclicVisitor
    
    /**
     * Interface for classes that may interpret a <b>GetField</b>
     * function.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(GetField<?,?> host);
    }
}
