// ============================================================================
// $Id: InvokeMethod.java,v 1.20 2009/09/06 22:35:23 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.property;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.MessageFormat;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.EvaluationException;
import net.sf.jga.util.ArrayUtils;

/**
 * Functor that invokes a method described at construction and returns the
 * result or null if the result is void.  The arguments must be passed to the
 * functor in an array, and the values must be assignable to the corresponding
 * classes given at construction.
 * <p>
 * Note that declaring the return type incorrectly can result in
 * ClassCastExceptions being thrown when the functor is invoked: the compiler
 * cannot check the return type of a reflectively loaded method, unless we 
 * add more arguments to the constructor.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class InvokeMethod<T,R> extends BinaryFunctor<T,Object[],R> {

    static final long serialVersionUID = -4632096384509133504L;

    // The class in which the method is accessed (not necessarily the class in which it is defined)
    private Class<T> _objclass;

    // The name of the method to invoke
    private String _methName;

    // The types of the arguments the method requires
    private Class<?>[] _argtypes;

    // The method to be invoked (lazily loaded)
    private transient Method _meth;

    /**
     * Builds an InvokeMethod functor for the given method. 
     * <p>
     * NOTE: this constructor has lower type-safety guarantee than the other constructors,
     * in that the compiler cannot check that the method's declaring class is T.  When a  
     * functor built by this constructor is evaluated, it will throw IllegalArgumentException
     * if given an incompatible argument.
     * @throws IllegalArgumentException if the method is null or if it takes arguments
     */
    @SuppressWarnings("unchecked") // unchecked cast of the class object for the method
    public InvokeMethod(Method method) {
        this((Class<T>) method.getDeclaringClass(), method);
    }
    
    /**
     * Builds a InvokeMethod for a given method
     * @throws IllegalArgumentException if the method or class is null; or if the method is not supplied 
     * by the given class; 
     */
    public InvokeMethod(Class<T> objClass, Method method) {
        if (method == null) {
            String msg = "Must supply method";
            throw new IllegalArgumentException(msg);
        }

        if (objClass == null) {
            String msg = "Must supply object class";
            throw new IllegalArgumentException(msg);
        }

        if (!method.getDeclaringClass().isAssignableFrom(objClass)) {
            String msg = "Method {0} not defined for class {1}";
            Object[] args = { method.getName(), objClass.getName() };
            throw new IllegalArgumentException(MessageFormat.format(msg, args));
        }
        
        _objclass = objClass;
        _meth     = method;
        _methName = method.getName();
        _argtypes = method.getParameterTypes();
    }

    /**
     * Builds an InvokeMethod for a given method, using the given class
     * to find the desired method.  This is a convenience constructor
     * for a common case: the case where the method takes a single argument.
     * It is still necessary for the argument passed to the fn method to be an
     * array, in this case it must be an array of length 1 containing a value of
     * of type argtype.  A ClassCastException may be thrown if this functor is
     * called incorrectly.
     * @throws IllegalArgumentException if class, method name, or argument type array is null;
     * or if the class does not supply the described method. 
     */
    public InvokeMethod(Class<T> objClass, String methName, Class<?> argtype) {
        this(objClass, methName, new Class[]{argtype});
    }

    /**
     * Builds a InvokeMethod for the given method, using the given class array
     * to find the desired method.
     * @throws IllegalArgumentException if class, method name, or argument type array is null;
     * or if the class does not supply the described method. 
     */
    public InvokeMethod(Class<T> objClass, String methName, Class<?>... argtypes) {
        if (methName == null || methName.length() == 0) {
            String msg = "Must supply method name";
            throw new IllegalArgumentException(msg);
        }

        if (objClass == null) {
            String msg = "Must supply object class";
            throw new IllegalArgumentException(msg);
        }

        _methName = methName;
        _objclass = objClass;
        _argtypes = argtypes;
        
        try {
            _meth = _objclass.getMethod(_methName, _argtypes);
        }
        catch (NoSuchMethodException x) {
            String msg = "No method {0} for class {1} that takes an argument(s) of type {2}";
            Object[] args =
                new Object[]{_methName, _objclass.getName(), ArrayUtils.toString(_argtypes) };
            IllegalArgumentException iax =
                new IllegalArgumentException(MessageFormat.format(msg,args));
            iax.initCause(x);
            throw iax;
        }
    }

    /**
     * Returns the class which supplied the method is invoked.
     */
    // UNTESTED
    public Class<T> getObjectType() {
        return _objclass;
    }

    /**
     * Returns the name of the method that this functor invokes.
     */
    public String getMethodName() {
        return _methName;
    }


    /**
     * Returns the type of the method 
     */
    @SuppressWarnings("unchecked")
    // There's nothing we can do about this other than warn the users
    // to make sure that they don't use an inappropriate return type
    public Class<R> getReturnType() {
        return (Class<R>) _meth.getReturnType();
    }

    
    /**
     * Lazy loads the method
     */
    public synchronized Method getMethod() throws NoSuchMethodException {
        if (_meth == null) 
            _meth = _objclass.getMethod(_methName, _argtypes);

        return _meth;
    }

    
    // BinaryFunctor interface
    
    @SuppressWarnings("unchecked")
    // There's nothing we can do about this other than warn the users
    // to make sure that they don't use an inappropriate return type
    public R fn(T obj, Object[] args) {
        try {
            return (R) getMethod(/*_argtypes*/).invoke(obj, args);
        }
        catch (NoSuchMethodException x) {
            String msg = "No method {0} for class {1} that takes an argument(s) of type {2}";
            Object[] msgargs =
                new Object[]{_methName,  _objclass.getName(), ArrayUtils.toString(_argtypes) };
            throw new EvaluationException(MessageFormat.format(msg,msgargs), x);
        }
        catch (ClassCastException x) {
            String msg = "ClassCastException: " +_objclass +"." +_methName
                +"("+ArrayUtils.toString(args)+")";
            throw new EvaluationException(msg, x);
        }
        catch (IllegalAccessException x) {
            String msg = _objclass +"." +_methName +" is not accessible";
            throw new EvaluationException(msg, x);
        }
        catch (InvocationTargetException x) {
            String xmsg = x.getMessage();
            String msg = "InvocationException: " +_objclass +"." +_methName
                +"("+ArrayUtils.toString(args)+")"+(xmsg != null ? (":"+xmsg) : "");
            throw new EvaluationException(msg, x);
        }
    }

    @Override
    @SuppressWarnings("unchecked") // asserting that the first argument is correctly typed is unsafe
    public R eval(Object... args) {
        if (args.length == 0)
            return fn((T) null, new Object[0]);
        
        if (args.length == 1) 
            return fn((T) args[0], new Object[0]);
        
        Object[] newArgs = new Object[args.length - 1];
        System.arraycopy(args, 1, newArgs, 0, newArgs.length);
        return fn((T)args[0], newArgs);
    }
    
    /**
     * Calls the Visitor's <code>visit(InvokeMethod)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof InvokeMethod.Visitor)
            ((InvokeMethod.Visitor)v).visit(this);
        else
            v.visit(this);
    }
    
    // Object overrides

    public String toString() {
        return "InvokeMethod("+_meth+")";
    }
    
   // AcyclicVisitor
    
    /**
     * Interface for classes that may interpret a <b>InvokeMethod</b>
     * function.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(InvokeMethod<?,?> host);
    }
}
