// ============================================================================
// $Id: SetField.java,v 1.10 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2002-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.property;

import java.lang.reflect.Field;
import java.text.MessageFormat;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.EvaluationException;

/**
 * Binary Functor that sets the named field of the first argument to the
 * value.  The field name and type are set at construction.  The return
 * value will be null.
 * <p>
 * Note that declaring the return type incorrectly can result in
 * ClassCastExceptions being thrown when the functor is invoked: the compiler
 * cannot check the return type of a reflectively loaded field.
 * <p>
 * Copyright &copy; 2002-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

// NOTE: compiling this class yields one unchecked cast warning.  It is really
// up to the user to declare this class properly (the return type must be
// correctly specified)

public class SetField<T,R> extends BinaryFunctor<T,R,R> {

    static final long serialVersionUID = -5051261348109487737L;
    
    // The name of the field
    private String _fieldName;

    // The field to set
    private transient Field _field;

    /**
     * Builds a SetField that will update the value of the given field
     * @throws IllegalArgumentException if either argument is omitted, or if
     * there is no such setter field in type argType.
     */
    @SuppressWarnings("unchecked")
    // There's nothing we can do about this other than warn the users
    // to make sure that they don't use an inappropriate return type
    public SetField(Field field) {
        this((Class<T>) field.getDeclaringClass(), field, (Class<R>) field.getType());
    }
    

    /**
     * Builds a SetField that will update the value of the given field
     * @throws IllegalArgumentException if either argument is omitted, or if
     * there is no such setter field in type argType.
     */
    @SuppressWarnings("unchecked")
    // There's nothing we can do about this other than warn the users
    // to make sure that they don't use an inappropriate return type
    public SetField(Class<T> argType, Field field) {
        this(argType, field, (Class<R>)field.getType());
    }
    

    /**
     * Builds a SetField that will update the value of the given field
     * @throws IllegalArgumentException if either argument is omitted, or if
     * there is no such setter field in type argType.
     */
    public SetField(Class<T> argType, Field field, Class<R> fieldType) {
        if (field == null) {
            throw new IllegalArgumentException("Must supply field");
        }
        
        if (!field.getDeclaringClass().isAssignableFrom(argType)) {
            throw new IllegalArgumentException(buildNoSuchFieldMessage(field.getName(), argType));
        }

        _fieldName = field.getName();
        _field = checkFieldType(field, fieldType);
    }
    

    /**
     * Builds a SetField that will update the value of the named field
     * of an instance of type argType.  The field will be of type propType.
     * @throws IllegalArgumentException if either argument is omitted, or if
     * there is no such setter field in type argType.
     */
    public SetField(Class<T> argType, String fieldName, Class<R> fieldType) {
        if (fieldName == null || fieldName.length() == 0) {
            throw new IllegalArgumentException("Must supply field name");
        }
        if (fieldType == null) {
            throw new IllegalArgumentException("Must supply field type");
        }
        
        _fieldName = fieldName;
        
        try {
            _field = checkFieldType(argType.getField(_fieldName), fieldType);
        }
        catch (NoSuchFieldException x) {
            IllegalArgumentException iax =
                new IllegalArgumentException(buildNoSuchFieldMessage(fieldName, argType));
            iax.initCause(x);
            throw iax;
        }
    }
    

    private Field checkFieldType(Field field, Class<R> fieldType) throws IllegalArgumentException {
        if (!field.getType().isAssignableFrom(fieldType)) {
            String msg = "Field {0} is of the wrong type: is {1}, expected {2}";
            Object[] args = new Object[] { field.getName(), field.getType().getName(),
                                               fieldType.getName() };
            throw new IllegalArgumentException(MessageFormat.format(msg, args));                
        }

        return field;
    }

    
    /**
     * Returns the name of the field that this functor sets.
     */
    public String getFieldName() {
        return _fieldName;
    }

    // Binary interface
    
    /**
     * Sets the designated field of the argument to the given value and returns
     * null.
     * <p>
     * @return null
     * @throws EvaluationException if the argument does not have the designated
     *         public field, or if it is accept the given value. 
     */
    public R fn(T arg, R val) {
        try {
            getField(arg).set(arg, val);
            return null;
        }
        catch (ClassCastException x) { 
            String msg = "{0}.{1} is of type {2}";
            Field f = getField(arg);
            Object[] args = new Object[]{ arg.getClass().getName(), f.getName(),
                                          f.getType().getName() };
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
        catch (IllegalAccessException x) {
            String msg = "{0}.{1} is not accessible";
            Object[] args = new Object[]{ arg.getClass().getName(), getField(arg).getName()};
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
    }

    private Field getField(T arg) {
        if (_field == null) {
            try {
                _field = arg.getClass().getField(_fieldName);
            }
            catch (NoSuchFieldException x) {
                throw new EvaluationException(x);}
        }

        return _field;
    }

    private String buildNoSuchFieldMessage(String fieldname, Class<?> clasz) {
        String msg = "class {0} does not have field {1}";
        Object[] args = new Object[]{ clasz.getName(), fieldname };
        return MessageFormat.format(msg,args);
    }

    /**
     * Calls the Visitor's <code>visit(etField)</code> field, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof SetField.Visitor)
            ((SetField.Visitor)v).visit(this);
        else
            v.visit(this);
    }
    
    // Object overrides

    public String toString() {
        return "SetField("+_fieldName+")";
    }
    
    // AcyclicVisitor
    
    /**
     * Interface for classes that may interpret a <b>SetField</b>
     * function.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(SetField<?,?> host);
    }
}
