// ============================================================================
// $Id: SetProperty.java,v 1.15 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2002-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.property;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.MessageFormat;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.EvaluationException;

/**
 * Binary Functor that sets the named property of the first argument to the
 * value.  The property name and type are set at construction.  The return
 * value will be that which the argument's property setter method returns
 * (generally either null or the old value).
 * <p>
 * Note that declaring the return type incorrectly can result in
 * ClassCastExceptions being thrown when the functor is invoked: the compiler
 * cannot check the return type of a reflectively loaded method.
 * <p>
 * Copyright &copy; 2002-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

// NOTE: compiling this class yields one unchecked cast warning.  It is really
// up to the user to declare this class properly (the return type must be
// correctly specified)

public class SetProperty<T,R> extends BinaryFunctor<T,R,R> {

    static final long serialVersionUID = 5305970242256716550L;
    
    // The property class, used to find the correct Method using reflection
    private Class<R> _propClass;

    // The name of the property (without the leading 'set').  
    private String _propName;

    // The name of the setter method (same as _propName, but with 'set' prefix)
    private String _methName;

    // The method to invoke
    private transient Method _meth;

    /**
     * Builds a SetProperty that will return the value of the named property
     * of an instance of type argType.  The property will be of type propType.
     * @throws IllegalArgumentException if either argument is omitted, or if
     * there is no such setter method in type argType.
     */
    public SetProperty(Class<T> argType, String propName, Class<R> propType) {
        if (propName == null || propName.length() == 0) {
            throw new IllegalArgumentException("Must supply property name");
        }
        if (propType == null) {
            throw new IllegalArgumentException("Must supply property type");
        }
        
        if (propName.startsWith("set")) {
            _methName = propName;
            _propName = propName.substring(3);
        }
        else {
            _propName = propName;
            _methName = "set" + propName;
        }
        
        _propClass = propType;
        
        try {
            Class<?>[] car = new Class[]{propType}; 
            _meth = argType.getMethod(_methName, car);
        }
        catch (NoSuchMethodException x) {
            String msg = "class {0} does not have property \"{1}\" of type {2}";
            Object[] args = new Object[]{ argType.getName(), propName, propType.getName() };
            IllegalArgumentException iax =
                new IllegalArgumentException(MessageFormat.format(msg,args));
            iax.initCause(x);
            throw iax;
        }
    }
    

    /**
     * Returns the name of the property that this functor sets.
     */
    public String getPropertyName() {
        return _propName;
    }

    // Binary interface
    
    /**
     * Sets the designated property of the argument to the given value.  When
     * the property's setter method returns a value, then this functor will
     * return it (otherwise it will return null).
     * <p>
     * @return the value returned by the designated property's setter method:
     *         generally it is null, but in some cases it might be the old value
     * @throws EvaluationException if the argument does not have the designated
     *         public property, or if it is accept the given value. 
     */
    @SuppressWarnings("unchecked")
    // There's nothing we can do about this other than warn the users
    // to make sure that they don't use an inappropriate return type
    public R fn(T arg, R val) {
        try {
            R ret = (R) getMethod(arg).invoke(arg, new Object[] {val});
            return ret;
        }
        catch (ClassCastException x) { 
            String msg = "{0}.{1} returns type {2}";
            Method m = getMethod(arg);
            Object[] args = new Object[]{ arg.getClass().getName(), m.getName(),
                                          m.getReturnType().getName() };
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
        catch (IllegalAccessException x) {
            String msg = "{0}.{1} is not accessible";
            Object[] args = new Object[]{ arg.getClass().getName(), getMethod(arg).getName()};
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
        catch (InvocationTargetException x) {
            String msg = "{0}.{1}({2}) failed : "+x.getMessage();
            Object[] args = new Object[]{ arg.getClass().getName(), getMethod(arg).getName(), val};
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
    }

    private Method getMethod(T arg) {
        if (_meth == null) {
            try {
                Class<?>[] car = new Class[]{_propClass};
                _meth = arg.getClass().getMethod(_methName, car);
            }
            catch (NoSuchMethodException x) {
                throw new EvaluationException(x);}
        }

        return _meth;
    }

    /**
     * Calls the Visitor's <code>visit(SetProperty)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof SetProperty.Visitor)
            ((SetProperty.Visitor)v).visit(this);
        else
            v.visit(this);
    }
    
    // Object overrides

    public String toString() {
        return "SetProperty("+_methName+")";
    }
    
    // AcyclicVisitor
    
    /**
     * Interface for classes that may interpret a <b>SetProperty</b>
     * function.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(SetProperty<?,?> host);
    }
}
