// ============================================================================
// $Id: Match.java,v 1.9 2007/02/07 12:20:48 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================
package net.sf.jga.fn.string;

import java.util.regex.Pattern;
import net.sf.jga.fn.UnaryFunctor;

/**
 * Unary Functor that tests a string argument against a given regular
 * expression.  The expression is not presumed to be anchored: if any part of
 * the input string matches the regular expression, then the result will be
 * true.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

public class Match extends UnaryFunctor<String,Boolean> {
    
    static final long serialVersionUID = 4842042451096017684L;

    // The regular expression to be matched (in String form for convenience
    // at construction and for reporting)
    private String _regex;

    // The regular expression to be matched (in Pattern form for execution)
    private Pattern _pattern;

    /**
     * Builds a Match with an empty string pattern
     */
    public Match () { this(""); }
    
    /**
     * Builds a Match with a given regular expression
     */
    public Match (String regex){
        _regex = (regex != null) ? regex : "";
        _pattern = Pattern.compile(_regex);
    }

    /**
     * Builds a Match with a given Pattern
     */
    public Match (Pattern pattern) {
        _pattern = pattern;
        _regex = pattern.pattern();
    }

    /**
     * Returns the format object used to present values in formatted form.
     * @return the format used to present values
     */
    
    public String getRegex() {
        return _regex;
    }
    
    // UnaryPredicate interface
    
    /**
     * Tests a string against the regular expression given at construction
     * <p>
     * @param arg the value to tested
     * @return true if the string matches the regular expression given at
     * construction
     */

    public Boolean fn(String arg) {
        if (arg == null)
            throw new NullPointerException();

        return _pattern.matcher(arg).find();
    }
    
    /**
     * Calls the Visitor's <code>visit(Match)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof Match.Visitor)
            ((Match.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return "Match("+_regex+")";
    }
    
    // Acyclic Visitor
    
    /**
     * Interface for classes that may interpret a <b>Match</b>
     * predicate.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(Match host);
    }
}
