// ============================================================================
// $Id: ParseFormat.java,v 1.14 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.fn.string;

import static net.sf.jga.fn.property.PropertyFunctors.*;

import java.text.DateFormat;
import java.text.Format;
import java.text.MessageFormat;
import java.text.NumberFormat;
import java.text.ParseException;
import net.sf.jga.fn.EvaluationException;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.fn.arithmetic.ValueOf;
import net.sf.jga.fn.property.Cast;

/**
 * Unary Functor that parses a given Format.
 * <p>
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 * @author <a href="mailto:davidahall@users.sourceforge.net">David A. Hall</a>
 **/

public class ParseFormat<R> extends UnaryFunctor<String,R> {

    static final long serialVersionUID = 6590747028326789147L;

    // The java.text.Format object that controls parsing.
    private Format _format;

    // A functor that converts the object from the type returned by the format's
    // parse() method to the parameter type
    private UnaryFunctor<Object, R> _converter;

    /**
     * Builds the ParseFormat given a text Format and a functor that can convert
     * Objects returned by the Format's parse(Object) method and the desired
     * type of this functor.
     */
    protected ParseFormat(Format format, UnaryFunctor<Object,R> conv) {
        if (format == null) {
            String msg = "Format must be specified";
            throw new IllegalArgumentException(msg);
        }
        
        _format = format;
        _converter = conv;
    }

    /**
     * @return the format used to parse values
     */

    public Format getFormat() {
        return _format;
    }
    
    // UnaryFunctor interface
    
    /**
     * Parses the value from the given string, using the java.text.Format object
     * passed at construction.
     * <p>
     * @param arg formatted string to be parsed
     * @throws java.text.ParseException when the string cannot be parsed to the
     * correct type
     * @return the value that the string represented
     */

    @SuppressWarnings("unchecked")
    // The cast to R causes an unchecked cast warning: we can't do anything about this 
    // other than warn the user not to use improper formats.  If the user gives us a 
    // converter, we're OK, otherwise all we can do is cast and hope for the best.
    public R fn(String arg) {
        
        try {
            Object obj =  _format.parseObject(arg);
            R val = (_converter == null) ? (R) obj : _converter.fn(obj);
            return val;
        }
        catch (ParseException x) {
            String msg = "Unable to parse \"{0}\"";
            Object[] args = new Object[]{ arg };
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
        catch (ClassCastException x) {
            String msg = "Unable to convert \"{0}\" to correct type";
            Object[] args = new Object[]{ arg };
            throw new EvaluationException(MessageFormat.format(msg,args), x);
        }
    }
    
    /**
     * Calls the Visitor's <code>visit(ParseFormat)</code> method, if it
     * implements the nested Visitor interface.
     */
    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof ParseFormat.Visitor)
            ((ParseFormat.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    // Object overrides

    public String toString() {
        return "ParseFormat["+_format+"]";
    }
    
    // Acyclic Visitor
    
    /**
     * Interface for classes that may interpret a <b>ParseFormat</b>
     * predicate.
     */
    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(ParseFormat<?> host);
    }

    /**
     * ParseFormat functor for use with Dates.  This class exists
     * as an implementation detail that works around a limit in the javac
     * inferencer -- in all substantive ways, this is simply a ParseFormat functor.
     */

    static public class Date<T extends java.util.Date> extends ParseFormat<T> {
        static final long serialVersionUID = 6959869794482290044L;
        @SuppressWarnings("unchecked")
        // If we want to return java.util.Date, then we can simply cast it.
        // Otherwise, we return [ new D(((Date)arg).getTime()) ]
        //    since we know that All Date subclasses have a constructor that
        //    takes a long (and this will break if a Date subclass is used that
        //    does not have such a constructor.
        // The other possibility is to build a date using a default constructor
        // and set its time property [ new D().setTime(((Date)arg).getTime()) ]
        // but that functor is even scarier
        public Date(Class<T> type, DateFormat format) {
            super(format, (UnaryFunctor<Object,T>)
                (type.equals(java.util.Date.class)
                     ? cast(java.util.Date.class)
                     : constructUnary(type, Long.TYPE)
                         .compose(getProperty(java.util.Date.class, "Time", Long.TYPE))
                         .compose(cast(java.util.Date.class))));
        }
    }

    /**
     * ParseFormat functor for use with Numbers.  This class exists
     * as an implementation detail that works around a limit in the javac
     * inferencer -- in all substantive ways, this is simply a ParseFormat functor.
     */

    static public class Number<T extends java.lang.Number> extends ParseFormat<T> {
        static final long serialVersionUID = -5365953713327551298L;
        public Number(Class<T> type, NumberFormat format) {
            super(format, //(UnaryFunctor<Object,R>)
                  new ValueOf<java.lang.Number,T>(type)
                     .compose(new Cast<Object,java.lang.Number>(java.lang.Number.class)));
        }
    }
}
