// ============================================================================
// $Id: GenericParser.java,v 1.18 2009/01/19 03:37:20 davidahall Exp $
// Copyright (c) 2004-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.parser;

import java.text.MessageFormat;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.Functor;
import net.sf.jga.fn.Generator;
import net.sf.jga.fn.UnaryFunctor;

/**
 * FunctorParser wrapper that adds runtime type-safe entry points to the non-generified
 * base parser.  This class maps between the inherently type<b>un</b>safe
 * baseclass and the generic functors that are returned: if the returned functor
 * does not return the correct type, then a ClassCastException will be thrown by
 * the entry point.
 * <p>
 * Copyright &copy; 2004-2005  David A. Hall
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

public class GenericParser implements FunctorParser {

    // global instance
    static private GenericParser _instance;

    /**
     * Returns a globally accessible default instance of a GenericParser.
     */
    static public synchronized GenericParser getInstance() {
        if (_instance == null) {
            _instance = new GenericParser(JFXGParser.getInstance());
        }

        return _instance;
    }


    private FunctorParser _delegate;

    public GenericParser() { _delegate = new JFXGParser(); }
    public GenericParser(FunctorParser delegate) { _delegate = delegate; }
    
    //======================
    // Parser Entry Points
    //======================

    /**
     * Parses the string to create a Functor.
     */
    public <R> Functor<R> parseFunctor(String str, Class<R> returnType, Class<?>... argTypes) 
        throws ClassCastException, ParseException
    {
        // try the three special cases first
        if (argTypes.length == 0)
            return parseGenerator(str, returnType);

        if (argTypes.length == 1)
            return parseUnary(str, argTypes[0], returnType);

        if (argTypes.length == 2)
            return parseBinary(str, argTypes[0], argTypes[1], returnType);

        throw new ParseException("Cannot parse arbitrary functors (yet)"); 
    }

    /**
     * Parses the string to create a Generator.
     */
    @SuppressWarnings("unchecked")
    // Compiler is objecting to an unsafe cast that we're checking at runtime
    public <R> Generator<R> parseGenerator(String str, Class<R> returnType)
        throws ClassCastException, ParseException
    {
        Generator<?> gen = parseGenerator(str);
        if (returnType.isAssignableFrom(getReturnType()))
            return (Generator<R>) gen;
        
        String err = "Expected Generator<{0}>, but expression returns {1}";
        String msg= MessageFormat.format(err, new Object[] {returnType.getName(),
                                                            getReturnType().getName()});
        throw new ClassCastException(msg);
    }

    /**
     * Parses the string to create a UnaryFunctor that takes an argument of the
     * given type.
     */
    @SuppressWarnings("unchecked")
    // Compiler is objecting to an unsafe cast that we're checking at runtime
    public <T,R> UnaryFunctor<T,R> parseUnary(String str, Class<T> argType, Class<R> returnType)
        throws ClassCastException, ParseException
    {
        UnaryFunctor<?,?> uf = parseUnary(str, argType);
        if (returnType.isAssignableFrom(getReturnType()))
            return (UnaryFunctor<T,R>) uf;

        String err = "Expected UnaryFunctor<{0},{1}>, but expression returns {2}";
        String msg= MessageFormat.format(err, new Object[] {argType.getName(),
                                                            returnType.getName(),
                                                            getReturnType().getName()});
        throw new ClassCastException(msg);
    }
    
    /**
     * Parses the string to create a BinaryFunctor that takes arguments of the
     * given types.
     */
    @SuppressWarnings("unchecked")
    // Compiler is objecting to an unsafe cast that we're checking at runtime
    public <T1,T2,R> BinaryFunctor<T1,T2,R>
    parseBinary(String str, Class<T1> arg1Type, Class<T2> arg2Type, Class<R> returnType)
        throws ClassCastException, ParseException
    {
        BinaryFunctor<T1,T2,?> bf = parseBinary(str, arg1Type, arg2Type);
        if (returnType.isAssignableFrom(getReturnType()))
            return (BinaryFunctor<T1,T2,R>) bf;

        String err = "Expected BinaryFunctor<{0},{1},{2}>, but expression returns {3}";
        String msg= MessageFormat.format(err, new Object[] {arg1Type.getName(),
                                                            arg2Type.getName(),
                                                            returnType.getName(),
                                                            getReturnType().getName()});
        throw new ClassCastException(msg);
    }


    //=======================
    // IParser implementation
    //=======================

    /**
     * Parses the string to create a Generator.
     */
    public Generator<?> parseGenerator(String str) throws ParseException {
        return _delegate.parseGenerator(str);
    }
    
    /**
     * Parses the string to create a UnaryFunctor that takes an argument of the
     * given type.
     */
    public <T> UnaryFunctor<T,?> parseUnary(String str, Class<T> argType) throws ParseException {
        return _delegate.parseUnary(str, argType);
    }

    /**
     * Parses the string to create a BinaryFunctor that takes arguments of the
     * given types.
     */
    public <T1,T2> BinaryFunctor<T1,T2,?> parseBinary(String str, Class<T1> arg1Type, Class<T2> arg2Type)
        throws ParseException
    {
        return _delegate.parseBinary(str, arg1Type, arg2Type);
    }
    
    /**
     * Returns the type of object returned by the last functor parsed.
     * @throws IllegalStateException if the parser has not been used or if 
     * parsing the last functor resulted in an exception being thrown.
     */
    public Class<?> getReturnType() {
        return _delegate.getReturnType();
    }
    
    //======================
    // Static entry points
    //======================

   /**
     * Parses the string to create a Generator.
     */
    static public <R> Generator<R> parse(String str, Class<R> returnType)
        throws UncheckedParseException, ClassCastException
    {
        try {
            return (Generator<R>) getInstance().parseGenerator(str, returnType);
        }
        catch(ParseException x) {
            throw new UncheckedParseException(x);
        }
    }

 
    /**
     * Parses the string to create a UnaryFunctor that takes an argument of the
     * given type.
     */
    static public <T,R> UnaryFunctor<T,R> parse(String str, Class<T> argType, Class<R> returnType)
            throws UncheckedParseException, ClassCastException
    {
        try {
            return getInstance().parseUnary(str, argType, returnType);
        }
        catch(ParseException x) {
            throw new UncheckedParseException(x);
        }
    }

 
    /**
     * Parses the string to create a BinaryFunctor that takes arguments of the
     * given types.
     */
    static public <T1,T2,R> BinaryFunctor<T1,T2,R>
    parse(String str, Class<T1> arg1Type, Class<T2> arg2Type, Class<R> returnType)
        throws UncheckedParseException, ClassCastException
    {
        try {
            return getInstance().parseBinary(str, arg1Type, arg2Type, returnType);
        }
        catch(ParseException x) {
            throw new UncheckedParseException(x);
        }
    }
}
