// ============================================================================
// $Id: ParserContext.java,v 1.8 2009/07/21 11:37:33 davidahall Exp $
// Copyright (c) 2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.parser;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.sf.jga.algorithms.Summarize;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.fn.arithmetic.ValueOf;
import net.sf.jga.fn.property.ConstructUnary;
import net.sf.jga.fn.property.InvokeNoArgMethod;
import net.sf.jga.fn.property.PropertyFunctors;

/**
 * Configuration settings for a JFXGParser.
 * <p>
 * Copyright &copy; 2005  David A. Hall
 */

class ParserContext {
    // Maps the argument names to their declared classes (declared to the entry point that
    // defines them: the names are bound to the types given to the parseGenerator, parseUnary,
    // and parseBinary methods)
//    private Map<String,VariableRef<?>> _arguments = new LinkedHashMap<String,VariableRef<?>>();
    private List<VariableRef<?>> _arguments = new ArrayList<VariableRef<?>>();
    
    // Maps local variable names to their declared types
    private Map<String,VariableRef<?>> _variables = new HashMap<String,VariableRef<?>>();
    
    // Imported classes, keyed by the classname within the package (or an alias)
    private Map<String,Class<?>> _imports = new HashMap<String,Class<?>>();

    // Imported packages
    private List<String> _packages = new ArrayList<String>();
    
    // Imported static methods, keyed by the method name (or an alias).  Multiple methods
    // may be keyed to the same name
    private Map<String,List<Method>> _methods = new HashMap<String,List<Method>>();  
    
    // Imported static members.
    private Map<String,Field> _fields = new HashMap<String,Field>();
    
    // The object to which "this" refers (the parser won't reference itself directly)
    private Object _thisObj;

    // True if decimal strings without a 'D' or 'F' are to be interpreted as BigDecimal
    // constants
    private boolean _undecoratedDecimal;

    ParserContext() {
        initializePromotionRules();
    }
    
    /**
     * Returns a copy of this context.
     */
    ParserContext copy() {
        ParserContext peer = new ParserContext();
        peer._variables.putAll(_variables);
        peer._imports.putAll(_imports);
        peer._methods.putAll(_methods);
        peer._fields.putAll(_fields);

        peer._arguments.addAll(_arguments);
        peer._packages.addAll(_packages);

        peer._thisObj = _thisObj;
        peer._undecoratedDecimal = _undecoratedDecimal;

        for (Class<?> fromtype : _promotionRules.keySet()) {
            HashMap<Class<?>,UnaryFunctor<?,?>> ruleset = _promotionRules.get(fromtype);
            for (Class<?> totype : ruleset.keySet()) {
                UnaryFunctor<?,?> fn = ruleset.get(totype);
                peer.addPromotionRule(fromtype, totype, fn);
            }
        }
        
        for (Class<?> fromtype : _demotionRules.keySet()) {
            HashMap<Class<?>,UnaryFunctor<?,?>> ruleset = _demotionRules.get(fromtype);
            for (Class<?> totype : ruleset.keySet()) {
                UnaryFunctor<?,?> fn = ruleset.get(totype);
                peer.addDemotionRule(fromtype, totype, fn);
            }
        }
        
        return peer;
    }

    /**
     * Declares a named argument to the functor being built, and its type.
     */
    public <T> VariableRef<T> declareArgument(int idx, String name, Class<T> type) {
        VariableRef<T> ref = new VariableRef<T>(name, type);
        increaseArgCount(idx + 1);
        _arguments.set(idx, ref);
        return ref;
    }
    
    /**
     * Declares a local variable
     */
    public <T> VariableRef<T> declareVariable(String name, Class<T> type){
        VariableRef<T> ref = new VariableRef<T>(name, type);
        _variables.put(name, ref);
        return ref;
    }

    
    /**
     * Returns an array of VariableRefs describing the argument space
     */
    public VariableRef<?>[] describeArgumentSpace() {
        return _arguments.toArray(new VariableRef<?>[_arguments.size()]);
        
    }
    
    /**
     * Returns an array of unary functors that provides the handling for the arguments
     * to the function being parsed.
     * @return
     */
    public UnaryFunctor<?,?>[] getArgumentSpace() {
        VariableRef<?>[] arguments = describeArgumentSpace(); 
        UnaryFunctor<?,?>[] handlers= new UnaryFunctor<?,?>[arguments.length];
        for(int i = 0; i < arguments.length; ++i) {
            handlers[i] = arguments[i].mutator();
        }
        
        return handlers;
    }

    /**
     * Returns the type of the named argument, or null if the name is unrecognized
     */
    public Class<?> getTypeForName(String name) {
        VariableRef<?> ref = resolveName(name);
        return (ref != null) ? ref.getType() : null;
    }
    
    
    /**
     * Returns the VariableRef associated with the given name, if a local variable
     * exists with the given name; otherwise, returns null;
     */
    @SuppressWarnings({ "serial" })
    public VariableRef<?> resolveName(final String name) {
        VariableRef<?> ref = _variables.get(name);
        if (ref == null) {
            ref = Summarize.lookup(_arguments,
                    PropertyFunctors.compareProperty(VariableRef.class, "Name", name));
        }
        
        return ref;
    }
    
    
    /**
     * Returns the type of the argument at the given integral position, or null if 
     * the index is out of bounds.
     */
    public Class<?> getTypeForPosition(int idx) {
        VariableRef<?> ref = resolvePosition(idx);
        return (ref != null) ? ref.getType() : null;
    }
    
    /**
     * Returns the VariableRef at the given integral position, creating it if this is the first reference
     */
    @SuppressWarnings("unchecked") // due to the untyped reference to VariableRef ctor
    public VariableRef<?> resolvePosition(int idx) {
        VariableRef<?> ref = null;
        try {
            ref = resolvePosition(idx, null);
        }
        catch (ParseException x) {
            // shouldn't happen -- the exception is thrown when the positional arg carries
            // invalid type information
        }
        
        return ref;
    }
    
    /**
     * Returns the VariableRef at the given integral position, creating it if this is the first reference.
     */
    @SuppressWarnings("unchecked") // due to the untyped reference to VariableRef ctor
    public VariableRef<?> resolvePosition(int idx, Class<?> type) throws ParseException {
        if (idx < 0)
            return null;
        
        if (idx >= _arguments.size()) {
            return declareArgument(idx, String.valueOf(idx), type != null ? type : Object.class);
        }

        VariableRef<?> ref = _arguments.get(idx);
        
        if (ref == VariableRef.UNDECLARED) {
            ref = new VariableRef(String.valueOf(idx), type != null ? type : Object.class);
            _arguments.set(idx, ref);
        }
        else if (type != null) {
            String msg = "Only the first reference to a positional argument may carry type information";
            throw new ParseException(msg);
        }

        return ref;
    }
    
    private void increaseArgCount(int size) {
        for (int i = _arguments.size(); i < size; ++i) {
            _arguments.add(VariableRef.UNDECLARED);
        }
    }
    
    private Class<?> inferredType;
    
    public void setInferredType(Class<?> type) { inferredType = type; }
    public Class<?> getInferredType() { return inferredType; }
    
    /**
     * Imports the given class into the parser.  The class will not need to be
     * fully qualified for use in expressions.  This will replace any existing
     * imported class that happens to have the same name.
     */
    public void importClass(Class<?> clasz) {
        importClass(ParserUtils.getSimpleName(clasz), clasz);
    }

    
    /**
     * Imports the given class into the parser under the given alias.  The alias
     * may be used in place of the fully qualified name of the class hereafter.
     * This will replace any existing imported class associated with the alias.
     */
    public void importClass(String alias, Class<?> clasz) {
        if (alias != null && clasz != null)
            _imports.put(alias, clasz);
    }


    /**
     * Removes information about the class imported under the given name.
     */
    public void deportClass(String alias) {
        if (alias != null)
            _imports.remove(alias);
    }

    /**
     * Returns the class associated with the given alias
     */
    public Class<?> getImportedClass(String alias) {
        Class<?> klass = _imports.get(alias);
        if (klass != null)
            return klass;

        List<Class<?>> potentials = new ArrayList<Class<?>>();
        for (String pkgName : _packages) {
            try {
                potentials.add(Class.forName(pkgName+"."+alias));
            }
            catch (Exception x) {
                // There's no way to know that a class exists, other than to try and
                // load it.  Exceptions for flow control are the only option, here.
            }
        }

        if (potentials.size() == 1) {
            return potentials.get(0);
        }

        return null;
    }

    
    /**
     * Imports all classes in the given package into the parser.
     */
    public void importPackage(String pkgName) {
        if ( ! _packages.contains(pkgName)) {
            _packages.add(pkgName);
        }
    }

    
    /**
     * Removes information about the package imported under the given name.
     */
    public void deportPackage(String pkgName) {
        _packages.remove(pkgName);
    }
    
    
    /**
     * Imports all of the static public methods and members in the given class.
     */
    public void importStatics(Class<?> clasz) {
        // First, do the fields
        Field[] fields = clasz.getFields();
        for (int i = 0; i < fields.length; ++i) {
            if (Modifier.isStatic(fields[i].getModifiers()))
                importField(fields[i]);
        }

        // next, do the methods
        Method[] methods = clasz.getMethods();
        for (int i= 0; i < methods.length; ++i) {
            if (Modifier.isStatic(methods[i].getModifiers()))
                importMethod(methods[i]);
        }
    }

    
    /**
     * Imports the given static member.  The member will not need to be qualified
     * with its classname for use in expressions.
     * @throws java.lang.NoSuchFieldException if the named field does not exist
     * @throws java.lang.IllegalArgumentException if the named field is not static
     */
    public void importField(Class<?> clasz, String name) throws NoSuchFieldException {
        importField(clasz.getField(name));
    }

    /**
     * Imports the given static member.  The member will not need to be qualified
     * with its classname for use in expressions.
     * @throws java.lang.IllegalArgumentException if the field is not static
     */
    public void importField(Field field) throws IllegalArgumentException {
        if (! Modifier.isStatic(field.getModifiers())) {
            String msg = "Cannot import non-static field {0}";
            throw new IllegalArgumentException(MessageFormat.format(msg, new Object[]{ field }));
        }
                                               
        _fields.put(field.getName(), field);
    }

    /**
     * Returns the imported field with the given name, or null if no such field has
     * been imported.
     */
    public Field getImportedField(String name) {
        return _fields.get(name);
    }
    
    /**
     * Imports the given static method(s).  
     * @throws java.lang.NoSuchMethodException if the named method does not exist, or if
     * it/they is/are not static
     */
    public void importMethod(Class<?> clasz, String name) throws NoSuchMethodException {
        Method[] methods = clasz.getMethods();
        int numImported = 0;
        for (int i= 0; i < methods.length; ++i) {
            if (name.equals(methods[i].getName()) && Modifier.isStatic(methods[i].getModifiers())) {
                importMethod(name, methods[i]);
                ++numImported;
            }
        }

        if (numImported == 0) {
            String msg = "No non-static method {0} found in class {1}";
            Object[] args = new Object[] {name, clasz.getName()};
            throw new NoSuchMethodException(MessageFormat.format(msg, args));
        }
    }

    /**
     * Imports a static method into the parser.  The method will not need to be
     * qualified with its classname for use in expressions.  Multiple methods
     * may be imported with the same name: the first best fit will be used when
     * the method name is used.
     * @throws java.lang.NoSuchMethodException if the named method does not exist
     * @throws java.lang.IllegalArgumentException if the named method is not static
     */
    public void importMethod(Method meth) {
        importMethod(meth.getName(), meth);
    }
    
    /**
     * Imports a static method into the parser under the given name.  The method
     * will not need to be qualified with its classname for use in expressions.
     * Multiple methods may be imported with the same name: the first best fit
     * will be used when the method name is used.
     * @throws java.lang.IllegalArgumentException if the method is not static
     */
    public void importMethod(String name, Method meth) {
        if (! Modifier.isStatic(meth.getModifiers())) {
            String msg = "Cannot import non-static method {0}";
            throw new IllegalArgumentException(MessageFormat.format(msg, new Object[]{ meth }));
        }
                                               
        List<Method> methodList = _methods.get(name);
        if (methodList == null) {
            methodList = new ArrayList<Method>();
            _methods.put(name, methodList);
        }
        methodList.add(meth);
    }


    /**
     * Returns a list of methods with the given name
     */
    public Method[] getImportedMethods(String name) {
        List<Method> methodList = _methods.get(name);
        if (methodList == null)
            return new Method[0];
        else
            return methodList.toArray(new Method[0]);
    }
    

    /**
     * Binds the object to which 'this' refers    
     */
    public void bindThis(Object thisBinding) {
        _thisObj = thisBinding;
    }

    
    /**
     * Returns the current object to which 'this' refers
     */
    protected Object getBoundObject() {
        return _thisObj;
    }
    
    
    /**
     * Enables/Disables the interpretation of undecorated decimal literals
     * as BigDecimals.  When true, an undecorated number containing a
     * decimal, for example <tt>1.50</tt> is interpreted as a BigDecimal
     * literal.  When false, the same token is interpreted as a Double, as
     * is the case in standard java
     */
    public void setUndecoratedDecimal(boolean flag) { _undecoratedDecimal = flag; }

    
    /**
     * When true, an undecorated number containing a
     * decimal, for example <tt>1.50</tt> is interpreted as a BigDecimal
     * literal.  When false, the same token is interpreted as a Double, as
     * is the case in standard java
     */
    public boolean isUndecoratedDecimal() { return _undecoratedDecimal; }


    /**
     * Registers a type-coercion rule that will promote values of the from type to
     * values of the to type for use in expressions.
     */
    public void addPromotionRule(Class<?> fromType, Class<?> toType, UnaryFunctor<?,?> converter) {
        addCoersion(fromType, toType, converter, _promotionRules);
    }

    /**
     * Registers a type-coercion rule that will promote values of the from type to
     * values of the to type for use in expressions.
     */
    public void addDemotionRule(Class<?> fromType, Class<?> toType, UnaryFunctor<?,?> converter) {
        addCoersion(fromType, toType, converter, _demotionRules);
    }

    /**
     */
    private void addCoersion(Class<?> fromType, Class<?> toType, UnaryFunctor<?,?> converter, 
                             HashMap<Class<?>, HashMap<Class<?>, UnaryFunctor<?,?>>> rules) {
        HashMap<Class<?>, UnaryFunctor<?,?>> ruleset = rules.get(fromType);
        if (ruleset == null) {
            ruleset = new HashMap<Class<?>, UnaryFunctor<?,?>>();
            rules.put(fromType, ruleset);
        }

        ruleset.put(toType, converter);
        
    }

    /**
     * Removes the type-coercion rule (if any) that would have promoted values of the
     * from type to values of the to type for use in expressions.
     */
    public void removePromotionRule(Class<?> fromType, Class<?> toType) {
        HashMap<Class<?>, UnaryFunctor<?,?>> ruleset = _promotionRules.get(fromType);
        if (ruleset != null)
            ruleset.remove(toType);
    }

    /**
     * Returns a type-coercion rule that will promote values of the from type to
     * values of the to type for use in expressions, if any such rule exists.
     * Returns null if no such rule exists.
     */
    public UnaryFunctor<?,?> getPromotionRule(Class<?> fromType, Class<?> toType) {
        return getCoersion(fromType, toType, _promotionRules);
    }
    
    public UnaryFunctor<?,?> getDemotionRule(Class<?> fromType, Class<?> toType) {
        return getCoersion(fromType, toType, _demotionRules);
    }
    
    private UnaryFunctor<?,?> 
    getCoersion(Class<?> fromType, Class<?> toType, 
            HashMap<Class<?>, HashMap<Class<?>, UnaryFunctor<?,?>>> rules) {
        HashMap<Class<?>, UnaryFunctor<?,?>> ruleset = rules.get(fromType);
        if (ruleset == null)
            return null;
        
        return ruleset.get(toType);
    }
    
    private HashMap<Class<?>, HashMap<Class<?>, UnaryFunctor<?,?>>> _promotionRules = 
        new HashMap<Class<?>, HashMap<Class<?>, UnaryFunctor<?,?>>>();

    private HashMap<Class<?>, HashMap<Class<?>, UnaryFunctor<?,?>>> _demotionRules = 
        new HashMap<Class<?>, HashMap<Class<?>, UnaryFunctor<?,?>>>();

    private UnaryFunctor<? extends Number,Byte> toByte          = new ValueOf<Number,Byte>(Byte.class);
    private UnaryFunctor<? extends Number,Short> toShort        = new ValueOf<Number,Short>(Short.class);
    private UnaryFunctor<? extends Number,Integer> toInteger    = new ValueOf<Number,Integer>(Integer.class);
    private UnaryFunctor<? extends Number,Long> toLong          = new ValueOf<Number,Long>(Long.class);
    private UnaryFunctor<? extends Number,Float> toFloat        = new ValueOf<Number,Float>(Float.class);
    private UnaryFunctor<? extends Number,Double> toDouble      = new ValueOf<Number,Double>(Double.class);
    private UnaryFunctor<? extends Number,BigDecimal> toDecimal = new ValueOf<Number,BigDecimal>(BigDecimal.class);
    private UnaryFunctor<? extends Number,BigInteger> toBigInt  = new ValueOf<Number,BigInteger>(BigInteger.class);
    
    @SuppressWarnings("unchecked")
    private ConstructUnary boxByte   = new ConstructUnary<Byte,Byte>(Byte.TYPE, Byte.class);
    
    @SuppressWarnings("unchecked")
    private ConstructUnary boxShort  = new ConstructUnary<Short,Short>(Short.TYPE, Short.class);
    
    @SuppressWarnings("unchecked")
    private ConstructUnary boxInt    = new ConstructUnary<Integer,Integer>(Integer.TYPE, Integer.class);
    
    @SuppressWarnings("unchecked")
    private ConstructUnary boxLong   = new ConstructUnary<Long,Long>(Long.TYPE, Long.class);
    
    @SuppressWarnings("unchecked")
    private ConstructUnary boxFloat  = new ConstructUnary<Float,Float>(Float.TYPE, Float.class);
    
    @SuppressWarnings("unchecked")
    private ConstructUnary boxDouble = new ConstructUnary<Double,Double>(Double.TYPE, Double.class);

    @SuppressWarnings("unchecked")
    private void initializePromotionRules() {
        addPromotionRule(Byte.class,      Short.class,     toShort);
        addPromotionRule(Byte.class,      Integer.class,   toInteger);
        addPromotionRule(Byte.class,      Long.class,      toLong);
        addPromotionRule(Byte.class,      BigInteger.class,toBigInt);
        addPromotionRule(Byte.class,      Double.class,    toDouble);
        addPromotionRule(Byte.class,      Float.class,     toFloat);
        addPromotionRule(Byte.class,      BigDecimal.class,toDecimal);
        
        addPromotionRule(Byte.TYPE,       Short.class,     toShort.compose(boxByte));
        addPromotionRule(Byte.TYPE,       Integer.class,   toInteger.compose(boxByte));
        addPromotionRule(Byte.TYPE,       Long.class,      toLong.compose(boxByte));
        addPromotionRule(Byte.TYPE,       BigInteger.class,toBigInt.compose(boxByte));
        addPromotionRule(Byte.TYPE,       Double.class,    toDouble.compose(boxByte));
        addPromotionRule(Byte.TYPE,       Float.class,     toFloat.compose(boxByte));
        addPromotionRule(Byte.TYPE,       BigDecimal.class,toDecimal.compose(boxByte));
       
        addPromotionRule(Short.class,     Integer.class,   toInteger);
        addPromotionRule(Short.class,     Long.class,      toLong);
        addPromotionRule(Short.class,     BigInteger.class,toBigInt);
        addPromotionRule(Short.class,     Double.class,    toDouble);
        addPromotionRule(Short.class,     Float.class,     toFloat);
        addPromotionRule(Short.class,     BigDecimal.class,toDecimal);
        
        addPromotionRule(Short.TYPE,      Integer.class,   toInteger.compose(boxShort));
        addPromotionRule(Short.TYPE,      Long.class,      toLong.compose(boxShort));
        addPromotionRule(Short.TYPE,      BigInteger.class,toBigInt.compose(boxShort));
        addPromotionRule(Short.TYPE,      Float.class,     toFloat.compose(boxShort));
        addPromotionRule(Short.TYPE,      Double.class,    toDouble.compose(boxShort));
        addPromotionRule(Short.TYPE,      BigDecimal.class,toDecimal.compose(boxShort));

        addPromotionRule(Integer.class,   Long.class,      toLong);
        addPromotionRule(Integer.class,   BigInteger.class,toBigInt);
        addPromotionRule(Integer.class,   Float.class,     toFloat);
        addPromotionRule(Integer.class,   Double.class,    toDouble);
        addPromotionRule(Integer.class,   BigDecimal.class,toDecimal);
        
        addPromotionRule(Integer.TYPE,    Long.class,      toLong.compose(boxInt));
        addPromotionRule(Integer.TYPE,    BigInteger.class,toBigInt.compose(boxInt));
        addPromotionRule(Integer.TYPE,    Float.class,     toFloat.compose(boxInt));
        addPromotionRule(Integer.TYPE,    Double.class,    toDouble.compose(boxInt));
        addPromotionRule(Integer.TYPE,    BigDecimal.class,toDecimal.compose(boxInt));

        addPromotionRule(Long.class,      BigInteger.class,toBigInt);
        addPromotionRule(Long.class,      Float.class,     toFloat);
        addPromotionRule(Long.class,      Double.class,    toDouble);
        addPromotionRule(Long.class,      BigDecimal.class,toDecimal);
        
        addPromotionRule(Long.TYPE,       BigInteger.class,toBigInt.compose(boxLong));
        addPromotionRule(Long.TYPE,       Float.class,     toFloat.compose(boxLong));
        addPromotionRule(Long.TYPE,       Double.class,    toDouble.compose(boxLong));
        addPromotionRule(Long.TYPE,       BigDecimal.class,toDecimal.compose(boxLong));

        addPromotionRule(BigInteger.class,Float.class,     toFloat);
        addPromotionRule(BigInteger.class,Double.class,    toDouble);
        addPromotionRule(BigInteger.class,BigDecimal.class,toDecimal);
        
        addPromotionRule(Float.class,     Double.class,    toDouble);
        addPromotionRule(Float.class,     BigDecimal.class,toDecimal);
        
        addPromotionRule(Float.TYPE,      Double.class,    toDouble.compose(boxFloat));
        addPromotionRule(Float.TYPE,      BigDecimal.class,toDecimal.compose(boxFloat));

        addPromotionRule(Double.class,    BigDecimal.class,toDecimal);
        
        addPromotionRule(Double.TYPE,     BigDecimal.class,toDecimal.compose(boxDouble));
        //
        
//        Cast<?,Integer> castInt = new Cast<Number,Integer>(Integer.TYPE);
        UnaryFunctor<Byte,Byte> byteValue  = new InvokeNoArgMethod(Byte.class,"byteValue");
        addDemotionRule(Short.class,      Byte.class,      toByte);
        addDemotionRule(Integer.class,    Byte.class,      toByte);
        addDemotionRule(Long.class,       Byte.class,      toByte);
        addDemotionRule(Float.class,      Byte.class,      toByte);
        addDemotionRule(Double.class,     Byte.class,      toByte);
        
        addDemotionRule(Short.class,      Byte.TYPE,       byteValue.compose(toByte));
        addDemotionRule(Integer.class,    Byte.TYPE,       byteValue.compose(toByte));
        addDemotionRule(Long.class,       Byte.TYPE,       byteValue.compose(toByte));
        addDemotionRule(Float.class,      Byte.TYPE,       byteValue.compose(toByte));
        addDemotionRule(Double.class,     Byte.TYPE,       byteValue.compose(toByte));
        
        UnaryFunctor<Short,Short> shortValue  = new InvokeNoArgMethod(Short.class,"shortValue");
        addDemotionRule(Integer.class,    Short.class,      toShort);
        addDemotionRule(Long.class,       Short.class,      toShort);
        addDemotionRule(Float.class,      Short.class,      toShort);
        addDemotionRule(Double.class,     Short.class,      toShort);
        
        addDemotionRule(Integer.class,    Short.TYPE,       shortValue.compose(toShort));
        addDemotionRule(Long.class,       Short.TYPE,       shortValue.compose(toShort));
        addDemotionRule(Float.class,      Short.TYPE,       shortValue.compose(toShort));
        addDemotionRule(Double.class,     Short.TYPE,       shortValue.compose(toShort));

        UnaryFunctor<Integer,Integer> intValue = new InvokeNoArgMethod(Integer.class, "intValue");
        addDemotionRule(Long.class,       Integer.class,    toInteger);
        addDemotionRule(Float.class,      Integer.class,    toInteger);
        addDemotionRule(Double.class,     Integer.class,    toInteger);
        
        addDemotionRule(Long.class,       Integer.TYPE,     intValue.compose(toInteger));
        addDemotionRule(Float.class,      Integer.TYPE,     intValue.compose(toInteger));
        addDemotionRule(Double.class,     Integer.TYPE,     intValue.compose(toInteger));

        UnaryFunctor<Long,Long> longValue = new InvokeNoArgMethod(Long.class, "longValue");
        addDemotionRule(Float.class,      Long.class,    toLong);
        addDemotionRule(Double.class,     Long.class,    toLong);
        
        addDemotionRule(Float.class,      Long.TYPE,     longValue.compose(toLong));
        addDemotionRule(Double.class,     Long.TYPE,     longValue.compose(toLong));

        UnaryFunctor<Float,Float> floatValue = new InvokeNoArgMethod(Float.class, "floatValue");
        addDemotionRule(Double.class,     Float.class,    toFloat);
        addDemotionRule(Double.class,     Float.TYPE,     floatValue.compose(toFloat));
    }
}
