// ============================================================================
// $Id: FunctorProxy.java,v 1.9 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.swing;

import java.io.Serializable;
import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.Collections;
import java.util.EventListener;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;

import net.sf.jga.algorithms.Filter;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.parser.GenericParser;
import net.sf.jga.parser.ParseException;

/**
 * Factory that builds Proxy objects that map Functors to event notification methods.
 * Using this proxy is a two step process: first build the proxy instance calling
 * <tt>makeListenerFor</tt> or <tt>newProxyInstance</tt>; second, register the
 * desired functors with the various registration methods.
 * <p>
 * Copyright &copy; 2005  David A. Hall
 */

public class FunctorProxy {

    /**
     * Manufactures a Proxy that can listen to any/all events that the given
     * object may fire.  The generated Proxy's invocation handler will be an
     * instance FunctorInvocationHandler.
     */
    static public Proxy makeListenerFor(Object obj)  {
        return newProxyInstance(obj.getClass().getClassLoader(), identifyListeners(obj.getClass()));
    }

    
    /**
     * Returns an instance of a proxy class for the specified interfaces that dispatches
     * method invocations to functors registed via a FunctorInvocationHandler.
     */
    static public Proxy newProxyInstance(ClassLoader loader, Class<?>[] interfaces) {
        return (Proxy) Proxy.newProxyInstance(loader, interfaces,
                                              new FunctorInvocationHandler(interfaces));
    }

    // ------------
    // registration
    // ------------
    
    /**
     * Registers the functor described by the given expression, to be invoked by the proxy
     * when a method with the given method is called.  If the proxy supports more than one
     * method with this name, then which one will be invoked is undefined.
     *
     * @throws NoSuchMethodException if no method with the given name is found
     * @throws ParseException if the expression cannot be parsed 
     * @throws ClassCastException if the given proxy's handler is not a FunctorInvocationHandler
     */
    static public void register(Proxy proxy, String methodName, String expression)
        throws ParseException, NoSuchMethodException
    {
        getFunctorInvocationHandler(proxy).register(methodName, expression);
    }

    
    /**
     * Registers the given functor to be invoked by the proxy when a method with the given
     * method is called.  If the proxy supports more than one method with this name, then
     * which one will be invoked is undefined.
     *
     * @throws NoSuchMethodException if no method with the given name is found
     * @throws ClassCastException if the given proxy's handler is not a FunctorInvocationHandler
     */
    static public void register(Proxy proxy, String methodName, UnaryFunctor<?,?> functor)
        throws NoSuchMethodException
    {
        getFunctorInvocationHandler(proxy).register(methodName, functor);
    }
    

    /**
     * Registers the given functor to be invoked by the proxy when the given method is called.
     * @throws ClassCastException if the given proxy's handler is not a FunctorInvocationHandler
     */
    static public void register(Proxy proxy, Method method, UnaryFunctor<?,?> functor) {
        getFunctorInvocationHandler(proxy).register(method, functor);
    }


    /**
     * Registers a functor to be evaluated when the given method is invoked.
     * @throws IllegalArgumentException if the proxy does not implement the listenerClass
     * @throws NoSuchMethodException if the proxy does not have a method with the given
     *    name that takes an argument of class eventClass
     * @throws ClassCastException if the given proxy's handler is not a FunctorInvocationHandler
     */
    static public void register(Proxy proxy, Class<?> listenerClass, Class<?> eventClass,
                                String methodName, UnaryFunctor<?,?> functor) 
        throws NoSuchMethodException, IllegalArgumentException
    {
        if (listenerClass.isInstance(proxy)) {
            Method m = listenerClass.getMethod(methodName, new Class<?>[]{eventClass});
            getFunctorInvocationHandler(proxy).register(m, functor);
            return;
        }

        throw new IllegalArgumentException("This does not implement " +listenerClass.getName());
    }

    
    // ----------------------
    // implementation details
    // ----------------------

    
    // Maps 'Listenee' classes to the arrays of Listener interfaces they support
    static private Map<Class<?>,WeakReference<?>> proxyClasses =
        Collections.synchronizedMap(new WeakHashMap<Class<?>,WeakReference<?>>());

    
    // A predicate that is true when a method is an add${FOO}Listener
    static private final UnaryFunctor<Method,Boolean> isAddListenerMethod =
        GenericParser.parse("x.getName().matches(\"^add.*Listener$\")", Method.class, Boolean.class);

    /**
     * Returns a list of the ${FOO}Listener interfaces which an object of Class clasz
     * may fire events (ie, the Listeners which such an object can add). 
     */
    static private Class<?>[] identifyListeners(Class<?> clasz) {
        synchronized (proxyClasses) {
            // See if we've already generated such a list: if so, return it.
            Object value = proxyClasses.get(clasz);
            if (value instanceof Reference) {
                Class<?>[] interfaceArray = (Class<?>[]) ((Reference<?>) value).get();
                if (interfaceArray != null) // Return the array if it has not been GC'd.
                    return interfaceArray;
            }

            // Otherwise, we'll need to build the list.  For now, we can find all methods
            // whose names are of the form "add${FOO}Listener", and that take a single
            // argument of a type derived from EventListener.
            Set<Class<?>> interfaceSet = new HashSet<Class<?>>();
            
//          for(Method m : Filter.filter(clasz.getMethods(), isAddListenerMethod)) {
            Iterator<Method> iter = Filter.filter(clasz.getMethods(),isAddListenerMethod).iterator();
            while(iter.hasNext()) {
                Method m = iter.next();
                Class<?>[] argtypes = m.getParameterTypes();
                if (argtypes.length == 1) {
                    Class<?> c = m.getParameterTypes()[0];
                    if (EventListener.class.isAssignableFrom(c)) {
                        interfaceSet.add(c);
                    }
                }
            }

            // Save ourselves the work next time around.
            Class<?>[] interfaceArray = interfaceSet.toArray(new Class<?>[interfaceSet.size()]);
            proxyClasses.put(clasz, new WeakReference<Class<?>[]>(interfaceArray));
            return interfaceArray;
        }
    }


    /**
     * @returns the FunctorInvocationHandler associated with the given proxy
     * @throws ClassCastException if the given proxy's handler is not a FunctorInvocationHandler
     */
    static private FunctorInvocationHandler getFunctorInvocationHandler(Proxy proxy) {
        return (FunctorInvocationHandler) Proxy.getInvocationHandler(proxy);
    }


    /**
     * The InvocationHandler for FunctorProxy instances.  This handler allows individual
     * methods to be mapped to Functors of the form UnaryFunctor<Event>.  When the method
     * (which must be an event notification method of one of the listener interfaces that
     * the proxy object implements) is invoked, the event is passed to the appropriate
     * functor for evaluation.
     */
    static public class FunctorInvocationHandler implements InvocationHandler, Serializable {
        
        static final long serialVersionUID = -6537167128759416788L;

        // The interfaces for which this handler can route method invocations
        private Class<?>[] _interfaces;

        // Maps methods to the functors that will be invoked
        private Map<Method,UnaryFunctor<?,?>> _functors = new HashMap<Method,UnaryFunctor<?,?>>();

        // Functor that returns true if a given method has the given name
        static private final BinaryFunctor<Method,String,Boolean> isNameEq =
            GenericParser.parse("x.getName() == y", Method.class, String.class, Boolean.class);

        private FunctorInvocationHandler(Class<?>[] interfaces) {
            _interfaces = interfaces;
        }

        
        /**
         * Returns the first method found with the given name
         * @throws NoSuchMethodException if no method with the given name is found
         */
        private Method getMethod(String name) throws NoSuchMethodException {
            UnaryFunctor<Method,Boolean> hasGivenName = isNameEq.bind2nd(name);

            // Loop through the interfaces that this handler supports to find
            // the one with the named method, and return the method if found

            for (int i = 0; i < _interfaces.length; ++i) {
                Class<?> iface = _interfaces[i];
                
                Iterator<Method> iter = Filter.filter(iface.getMethods(), hasGivenName).iterator();
                if(iter.hasNext()) {
                    return iter.next(); 
                }
            }
            
            throw new NoSuchMethodException(name);
        }


        /**
         * Registers the functor described by the given expression, to be invoked a
         * method with the given method is called.  If this handler has more than one
         * method with this name, then which one will be invoked is undefined.
         *
         * @throws NoSuchMethodException if no method with the given name is found
         * @throws ParseException if the expression cannot be parsed 
         */
        private void register(String methodName, String expression)
            throws ParseException, NoSuchMethodException
        {
            Method m = getMethod(methodName);
            UnaryFunctor<?,?> fn =
                GenericParser.parse(expression, m.getParameterTypes()[0], m.getReturnType());
            
            register(m, fn);
        }
    
    
        /**
         * Registers the given functor, to be invoked a method with the given method
         * is called.  If this handler has more than one method with this name, then
         * which one will be invoked is undefined.
         *
         * @throws NoSuchMethodException if no method with the given name is found
         */
        private void register(String methodName, UnaryFunctor<?,?> functor)
            throws NoSuchMethodException
        {
            Method m = getMethod(methodName);
            register(m, functor);
        }
    

        /**
         * Registers the given functor, to be invoked when the given method is called.
         */
        private void register(Method method, UnaryFunctor<?,?> functor) {
            _functors.put(method, functor);
        }

        
        
        /**
         */
        private UnaryFunctor<?,?> getFunctor(Method method){
            return _functors.get(method);
        }

        // --------------------------------
        // InvocationHandler Implementation 
        // --------------------------------

        /**
         * Type-unsafe implementation of invoke, but invoke is an inherently un-typesafe operation
         * anyway.  Be careful to pass arguments of the correct type in the argument array, lest
         * ClassCastExceptions be thrown.
         */
        @SuppressWarnings("unchecked")
        // There's no possiblity of making this method typesafe: we're ultimately calling a method
        // reflectively using the given array of arguments.  
        public Object invoke(Object object, Method method, Object[] objectArray) throws Throwable {
            UnaryFunctor functor = getFunctor(method);
            if (functor != null) {
                functor.fn(objectArray[0]);
            }
            
            return null;
        }
    }
}
