// ============================================================================
// $Id: GenericCellEditor.java,v 1.14 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.swing;

import java.awt.Color;
import javax.swing.DefaultCellEditor;
import javax.swing.JTextField;
import javax.swing.border.LineBorder;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.fn.property.ConstructUnary;
import net.sf.jga.fn.string.DefaultFormat;
import net.sf.jga.fn.EvaluationException;

/**
 * CellEditor that uses a pair of functors to parse/format a value.  Empty
 * strings, or strings containing only whitespace will correspond to null
 * values of type T.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

public class GenericCellEditor<T> extends DefaultCellEditor {

    static final long serialVersionUID = -6265710748312156901L;

    // correctly typed reference to the editor component, saving excess casts
    protected JTextField _component;

    // default value returned when the string is empty.
    private T _defaultValue;

    
    /**
     * Builds a GenericCellEdior using default format for the given class.  The
     * resulting editor can support any class that defines toString() and also
     * has a one argument constructor that takes a String.
     */
    public GenericCellEditor (Class<T> type) {
        this(new DefaultFormat<T>(), new ConstructUnary<String,T>(String.class, type));
    }
    
    /**
     * Builds the GenericCellEditor using the given functor pair.
     */
    public GenericCellEditor (UnaryFunctor<T,String> formatter, UnaryFunctor<String,T> parser) {
        this(formatter, parser, null);
    }

    /**
     * Builds the GenericCellEditor using the given functor pair and default
     * value for empty strings.
     */
    public GenericCellEditor (UnaryFunctor<T,String> formatter, UnaryFunctor<String,T> parser,
                              T defaultValue)
    {
        super(new JTextField());
        if (formatter == null) {
            String msg = "Formatter (functor) is required";
            throw new IllegalArgumentException(msg);
        }
        
        if (parser == null) {
            String msg = "Parser (functor) is required";
            throw new IllegalArgumentException(msg);
        }
        
        _component = (JTextField) editorComponent;
        _defaultValue = defaultValue;

        // overrides the protected delegate 
        delegate = new FormattedDelegate(formatter, parser);
    }

    /**
     * Retrieves the value from the editor component, giving a visual clue when
     * the value is invalid
     */
    
	public boolean stopCellEditing() {
        try {
            // The FormattedDelegate class provides the implementation of
            // getCellEditorValue, and it is known to always return a T.
            // We're doing this solely to catch the exception that might
            // be thrown
            getCellEditorValue();
	    }
	    catch (Exception e) {
            _component.setBorder(new LineBorder(Color.red));
            return false;
	    }
        
	    return super.stopCellEditing();
	}

    /**
     * Returns the horizontal alignment: one of the values defined by JTextField.
     */
    public int getHorizontalAlignment() {
        return _component.getHorizontalAlignment();
    }
    
    /**
     * Sets the horizontal alignment to one of the values defined by JTextField.
     */
    public void setHorizontalAlignment(int alignment) {
        _component.setHorizontalAlignment(alignment);
    }

    public String toString() {
        return "GenericTableCellEditor["+delegate+"]";
    }
    
    //
    // Overrides the base class nested EditorDelegate class, using the given
    // functor pair to fill in and interpret the value of the component.
    //

    protected class FormattedDelegate extends EditorDelegate {

        static final long serialVersionUID = -3670848351305705031L;
        
        private UnaryFunctor<T,String> _formatter;
        private UnaryFunctor<String,T> _parser;
        
        public FormattedDelegate(UnaryFunctor<T,String> formatter,
                                 UnaryFunctor<String,T> parser)
        {
            _formatter = formatter;
            _parser = parser;
        }
        
        /**
         * Sets the text in the editor component by formatting the given value.
         * Will use the default value if one is available when given a null,
         * otherwise will use an empty string.
         */
        @SuppressWarnings("unchecked")
        // The cast to T is unsafe, but we're also explicitly catching and sort-of dealing with
        // the ClassCastException
        public void setValue(Object value) {
            if (value != null) {
                // @SuppressWarnings
                try {
                    _component.setText(_formatter.fn((T)value));
                    return;
                }
                catch (ClassCastException x) {
                    // The value we're passed is not of the correct type, so
                    // fall through to one of the defaults.
                }
            }
            
            if (_defaultValue != null) {
                _component.setText(_formatter.fn(_defaultValue));
                return;
            }
            
            _component.setText("");
        }
        
        /**
         * Retrieves the value in the component.  If the component contains an
         * empty string (or one with only whitespace), or if an exception
         * occurs, then returns the default value (which may be null).
         */
        public Object getCellEditorValue() {
            String txt = _component.getText();
            if (txt == null || txt.trim().length() == 0) {
                return _defaultValue;
            }

            try {
                Object val = _parser.fn(txt);
                return val;
            }
            catch (EvaluationException x) {
                return _defaultValue;
            }
        }

        public String toString() { return "FormattedDelegate["+_formatter+","+_parser+"]"; }
    }
}
