// ============================================================================
// $Id: GenericTableCellRenderer.java,v 1.15 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================
package net.sf.jga.swing;

import javax.swing.table.DefaultTableCellRenderer;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.fn.string.DefaultFormat;
import net.sf.jga.util.Formattable;

/**
 * TableCellRenderer that passes the contents through a functor to render the
 * contents, instead of calling toString() on the contents.  It is up to the user
 * to ensure that the contents of the table are of the correct type, otherwise,
 * a ClassCastException will be thrown during the compo
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

public class GenericTableCellRenderer<T> extends DefaultTableCellRenderer implements Formattable<T> {

    static final long serialVersionUID = -1084070441724877620L;

    private UnaryFunctor<T,String> _formatter;
    private boolean _formatNulls = false;

    /**
     * Builds a TableCellRenderer that uses a default formatter
     */
    public GenericTableCellRenderer () {
        this(new DefaultFormat<T>(), false);
    }

    /**
     * Builds a TableCellRenderer that uses the given functor to format values.
     */
    public GenericTableCellRenderer (UnaryFunctor<T,String> formatter) {
        this(formatter, false);
    }

    /**
     * Builds a TableCellRenderer that uses the given functor to format values,
     * and the given flag to determine if nulls are formatted.
     */
    public GenericTableCellRenderer (UnaryFunctor<T,String> formatter,
                                     boolean formatsNulls)
    {
        if (formatter == null) {
            throw new IllegalArgumentException("Non-null functor required");
        }
        
        _formatter = formatter;
        _formatNulls = formatsNulls;
    }

    /**
     * Overrides the base class setValue: 
     */
    @SuppressWarnings("unchecked")
    // The cast to T is unsafe, but we're also explicitly catching and sort-of dealing with
    // the ClassCastException
    protected void setValue(Object value) {
        if (value == null && !_formatNulls) {
            setText("");
            return;
        }
        
        try {
            setText(_formatter.fn((T)value));
        }
        catch (ClassCastException x) {
            setText("#" +value +"#");
        }
        catch (IllegalArgumentException x) {
            setText("#" +value +"#");
        }
    }

    public String toString() {
        return "GenericTableCellRenderer["+_formatter+"]";
    }
    
    // ---------------------
    // Formattable interface
    // ---------------------
    
    /**
     * Changes the format of an existing renderer
     * @throws IllegalArgumentException
     */
    public void setFormat(UnaryFunctor<T,String> formatter) {
        if (formatter == null) {
            throw new IllegalArgumentException("Non-null functor required");
        }
        
        _formatter = formatter;
    }
}
