// ============================================================================
// $Id: GenericTableColumn.java,v 1.8 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================
package net.sf.jga.swing;

import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.UnaryFunctor;

/**
 * Column class used in conjunction with GenericTableModel.  It is up to the user
 * to ensure that the contents of the list are of the correct type, otherwise,
 * a ClassCastException will be thrown during the component's paint routine.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

public class GenericTableColumn<R,C>  extends TableColumn {
    
    static final long serialVersionUID = 1960293231258498520L;
    
    private Class<C> _coltype;
    private UnaryFunctor<R,C> _getFn;
    private BinaryFunctor<R,C,C> _setFn;

    /**
     * Builds a read-only column that will apply the given functor to objects of
     * the given class to get the value of specific cells.  A default table
     * renderer will be used, as defined by the Table that uses this column.
     */
    public GenericTableColumn(Class<C> coltype, UnaryFunctor<R,C> getFn) {
        this(coltype, getFn, null);
    }

    /**
     * Builds a potentially editable column that will apply the given functors to
     * objects of the given class to get and set the value of specific cells.
     * The setFn can be null: if it is, then this column will not be editable.
     * A default table renderer and editor will be used, as defined by the Table
     * that uses this column.
     *
     */
    public GenericTableColumn(Class<C> coltype, UnaryFunctor<R,C> getFn,
                              BinaryFunctor<R,C,C> setFn)
    {
        _coltype = coltype;
        _getFn   = getFn;
        _setFn   = setFn;
    }
    
    /**
     * Builds a potentially editable column that will apply the given functors
     * to objects of the given class to get and set the value of specific cells.
     * The setFn can be null: if it is, then this column will not be editable.
     * The formatter will be used to build a GenericTableCellRenderer for the
     * column if it is non-null.  If formatter, parser, and setFn are all
     * non-null, then a GenericCellEditor will be built for this column.  If any
     * of these fields are null, the the Table that uses this column will supply
     * the default renderer/editor. 
     */
    public GenericTableColumn(Class<C> coltype, UnaryFunctor<R,C> getFn,
                              BinaryFunctor<R,C,C> setFn, 
                              UnaryFunctor<C,String> formatter,
                              UnaryFunctor<String,C> parser)
    {
        this(coltype, getFn, setFn);
     
        if (formatter != null) {
            setCellRenderer(new GenericTableCellRenderer<C>(formatter));

            if (parser != null && setFn != null) {
                setCellEditor(new GenericCellEditor<C>(formatter, parser));
            }
        }
    }
    

    /**
     * Builds a potentially editable column that will apply the given functors
     * to objects of the given class to get and set the value of specific cells.
     * The setFn can be null: if it is, then this column will not be editable.
     * Either the renderer or editor may be null: the Table will supply the
     * default renderer/editor in this case.  Note that if the setFn is null,
     * there will be no editor configured (as it would never be called anyway).
     */
    public GenericTableColumn(Class<C> coltype, UnaryFunctor<R,C> getFn,
                              BinaryFunctor<R,C,C> setFn,
                              TableCellRenderer renderer, TableCellEditor editor)
    {
        this(coltype, getFn, setFn);
        setCellRenderer(renderer);
        if (setFn != null)
            setCellEditor(editor);
    }
    

    /*
     *
     */
    public C getValueAt (R rowvalue) {
        return _getFn.fn(rowvalue);
    }

    /*
     * @throw ClassCastException if the given value is not of class C
     */
    @SuppressWarnings("unchecked")
    // There's no way to guarantee that the user has configured this correctly, ie, 
    // ensure that the value to be set is of class C.
    public void setValueAt (R rowvalue, Object obj) {
        if (_setFn != null) {
            _setFn.fn(rowvalue, (C) obj); 
        }
    }
    
    /*
     *
     */
    public boolean isEditable() {
        return _setFn != null;
    }

    /*
     *
     */
    public Class<C> getColumnClass() {
        return _coltype;
    }
    
    /*
     *
     */
    public String toString() {
        return super.toString() +"["+_getFn+","+_setFn+"]";
    }
}
