// ============================================================================
// $Id: GenericTableModel.java,v 1.16 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================
package net.sf.jga.swing;

import java.util.Iterator;
import java.util.List;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableColumnModel;
import javax.swing.table.TableColumnModel;
import net.sf.jga.fn.EvaluationException;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.fn.property.GetProperty;
import net.sf.jga.fn.property.SetProperty;

/**
 * TableModel that uses a list of data for storage, and whose columns contain
 * functors that read (and possibly write) properties of the objects in the
 * list.
 * <p>
 * This class uses GenericTableColumns to store information about each specific
 * column.  To keep the correct column classes in use by the table, it is
 * important to build the table using the constructor that takes both the data
 * model and the column model, as shown in the code sample below.  Otherwise,
 * the table will build a default TableColumnModel of its own, and additions
 * of columns to the data model won't be reflected in the column model.
 * <p>
 * Another caveat on the use of this class is that if columns are to be added to
 * the model after it has been given to a table, then the table's
 * autoCreateColumnsFromModel flag must be false (otherwise, the table will
 * discard the existing columns and attempt to build new ones: JTable won't know
 * how to create instances of the generic classes.
 * <P>
 * Working with this class can be fairly simple.  To build a simple table that
 * allows for the display and editing of a typical business object (Item, in
 * this example),
 * <p>
 * <pre>
 * List<Item> data = // initialized from somewhere
 * GenericTableModel<Item> model =
 *     new GenericTableModel<Item>(Item.class, data);
 *
 *  // adds a read-only column for the object's id
 *  model.addColumn(Integer.class, "ID");
 *  
 *  // adds a read-only column for the object's name
 *  model.addColumn(String.class, "Name");
 *  
 *  // adds an editable column for the object's description
 *  model.addColumn(String.class, "Desc", true);
 *
 *  // adds an editable column for the object's count
 *  model.addColumn(Integer.class, "Count", true);
 *
 *  // adds an editable column for the object's price
 *  model.addColumn(BigDecimal.class, "Price", true);
 * 
 * JTable table = new JTable(model, model.getColumnModel());
 * </pre>
 * <p>
 * In this example, the Item class is presumed to have the appropriate
 * getter/setter methods defined.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

// TODO: there should be an associated GenericTableColumnModel to go along
//       with this.  It might be possible for addColumn to be safe after the
//       table has been assigned (if the table consults the TableColumnModel
//       when it rebuilds (but it probably doesn't))
// NOTE: Tried this -- there are problems.  Specifically, the column model
//       really wants to share (or take over)some of the internal state.  It
//       leads to having column, model, and column model, all of which need to
//       be used together or not at all.  The only real way to keep everything
//       in sync is to create a GenericTable that overrides the default class
//       used for model,column,and column model.  I didn't want to go that far.

// TODO: this should wrap a ListModel

public class GenericTableModel<T> extends AbstractTableModel {
    
    static final long serialVersionUID = 8996104516840561845L;

    private Class<T> _rowtype;
    private List<T> _values;

    private DefaultTableColumnModel _columnModel = new DefaultTableColumnModel();
    
    /**
     * Builds a GenericTableModel for the given list of data.  The columns
     * must be added separately before the table may be rendered.
     */
    public GenericTableModel (Class<T> rowtype, List<T> values) {
        _rowtype = rowtype;
        _values = values;
    }
    
    /**
     * Builds a GenericTableModel for the given list of data, using the
     * given list of columns.
     */
    public GenericTableModel (Class<T> rowtype, List<T> values,
                              List<GenericTableColumn<T,?>> columns)
    {
        _rowtype = rowtype;
        _values = values;
        Iterator<GenericTableColumn<T,?>> iter = columns.iterator();
        while(iter.hasNext())
            _columnModel.addColumn(iter.next());
    }

    /**
     * Returns the column model that should be used in conjunction with this
     * data model.
     */

    public TableColumnModel getColumnModel() {
        return _columnModel;
    }
    
    /**
     * Returns the value that corresponds to the given row of the table
     */
    public T getRowValue(int row) {
        return _values.get(row);
    }

    // Flag that determines if the columns' names are to be used in the table
    // header.  Generally, that will be the case, so this defaults to true.
    private boolean _nameUsedInHeader = true;
    
    /**
     * Enables/Disables use of property names in the column headers of all
     * columns subsequently added to the table.  Existing header values will
     * not be changed when this value changes.  Defaults to true (enabled).
     */
    public void setNameUsedInHeader(boolean b) { _nameUsedInHeader = b; }

    
    /**
     * Returns true if property names are to be used in the headers of the
     * corresponding columns.
     */
    public boolean isNameUsedInHeader() { return _nameUsedInHeader; }
    
    /**
     * Adds a read-only column that uses a GetProperty functor for the named
     * property.  Do not call this method after the model has been given to a
     * table unless the table's autoCreateColumnsFromModel flag is unset.
     */
    
    public <C> GenericTableColumn<T,C> addColumn(Class<C> coltype, String name){
        GetProperty<T,C> getter = new GetProperty<T,C>(_rowtype, name);
        return addColumnImpl(name, new GenericTableColumn<T,C>(coltype, getter));
    }
    
    /**
     * Adds a possibly editable column that uses a GetProperty functor for
     * the named property, and a SetProperty functor if the column is
     * editable.  Do not call this method after the model has been given to
     * a table unless the table's autoCreateColumnsFromModel flag is unset.
     */
    
    public <C> GenericTableColumn<T,C> addColumn(Class<C> coltype, String name,
                                                 boolean editable)
    {
        GetProperty<T,C> getter = new GetProperty<T,C>(_rowtype, name);
        SetProperty<T,C> setter = editable ? new SetProperty<T,C>(_rowtype, name, coltype) : null;
        return addColumnImpl(name, new GenericTableColumn<T,C>(coltype, getter, setter));
    }
        
    /**
     * Adds a read-only column that uses a GetProperty functor for the named
     * property and renders the property value in the given formatter.  Do not
     * call this method after the model has been given to a table unless the
     * table's autoCreateColumnsFromModel flag is unset.
     */
    public <C> GenericTableColumn<T,C> addColumn(Class<C> coltype, String name,
                                                 UnaryFunctor<C,String> formatter)
    {
        GetProperty<T,C> getter = new GetProperty<T,C>(_rowtype, name);
        return addColumnImpl(name, new GenericTableColumn<T,C>(coltype,getter,null,formatter,null));
    }
    

    /**
     * Adds an editable column that uses a GetProperty functor for
     * the named property, and a SetProperty functor if the column is
     * editable, and is rendered (and edited) using the given format.  Do not
     * call this method after the model has been given to a table unless the
     * table's autoCreateColumnsFromModel flag is unset.
     */
    public <C> GenericTableColumn<T,C> addColumn(Class<C> coltype, String name,
                                                 UnaryFunctor<C,String> formatter,
                                                 UnaryFunctor<String,C> parser)
    {
        GetProperty<T,C> getter = new GetProperty<T,C>(_rowtype, name);
        SetProperty<T,C> setter = (parser != null) ?
            new SetProperty<T,C>(_rowtype, name, coltype) : null;

        return addColumnImpl(name,
                             new GenericTableColumn<T,C>(coltype,getter,setter,formatter,parser));
    }


    /**
     * Completes the configuration of a new column and adds it to the table
     */
    private <C> GenericTableColumn<T,C> addColumnImpl(String name, GenericTableColumn<T,C> column) {
        if (_nameUsedInHeader)
            column.setHeaderValue(name);
        
        addColumn(column);
        return column;
    }

    /**
     * Adds the given column to the table.  Do not call this method
     * after the model has been given to a table unless the table's
     * autoCreateColumnsFromModel flag is unset.
     */
    
    public void addColumn(GenericTableColumn<T,?> col) {
        col.setModelIndex(getColumnCount());
        _columnModel.addColumn(col);
    }

    /**
     * Returns the GenericTableColumn associated with the given column index.
     */
    
    @SuppressWarnings("unchecked")
    // The column model is private, and is known to contain only GenericTableColumns.
    public GenericTableColumn<T,?> getGenericColumn(int col) {
        return (GenericTableColumn<T,?>) _columnModel.getColumn(col);
    }
    
    // -------------------------------------------------------------
    // TableModel interface
    // -------------------------------------------------------------
    
    public int getRowCount() {
        return _values.size();
    }

    public int getColumnCount() {
        return _columnModel.getColumnCount();
    }

    public Class<?> getColumnClass(int col) {
        return getGenericColumn(col).getColumnClass();
    }
    
    public String getColumnName(int col) {
        GenericTableColumn<T,?> column = getGenericColumn(col);
        Object name = column.getHeaderValue();
        return (name != null) ? name.toString() : super.getColumnName(col);
    }
    
    public boolean isCellEditable(int row, int col) {
        return getGenericColumn(col).isEditable();
    }

    public Object getValueAt(int row, int col) {
        T rowvalue = getRowValue(row);
        GenericTableColumn<T,?> column = getGenericColumn(col);
        try {
            return column.getValueAt(rowvalue);
        }
        catch (EvaluationException x) {
            x.printStackTrace();
            return "### ERROR ###";
        }
    }

    public void setValueAt(Object value, int row, int col) {
        T rowvalue = getRowValue(row);
        GenericTableColumn<T,?> column = getGenericColumn(col);
        assert column.getColumnClass().isInstance(value) || value == null;
        
        column.setValueAt(rowvalue, value);
        fireTableCellUpdated(row,col);        
    }
}
