// ============================================================================
// $Id: Application.java,v 1.10 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2004-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.swing.spreadsheet;

import static net.sf.jga.fn.property.PropertyFunctors.setProperty;
import static net.sf.jga.fn.property.PropertyFunctors.invokeMethod;
import static net.sf.jga.fn.property.PropertyFunctors.invokeNoArgMethod;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URL;
import javax.swing.Action;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.UIManager;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.fn.adaptor.ApplyBinary;
import net.sf.jga.fn.adaptor.ConstantBinary;
import net.sf.jga.fn.adaptor.Project1st;
import net.sf.jga.fn.adaptor.Project2nd;
import net.sf.jga.fn.property.ArrayBinary;
import net.sf.jga.fn.property.ArrayUnary;
//import net.sf.jga.fn.property.InvokeMethod;
import net.sf.jga.fn.property.SetProperty;
import net.sf.jga.parser.JFXGParser;
import net.sf.jga.swing.GenericAction;

/**
 * An application wrapper for Spreadsheet, providing a main method to
 * allow for standalone use.  
 * <p>
 * Copyright &copy; 2004-2005  David A. Hall
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

public class Application {

    private JFrame _frame;

    private Spreadsheet _sheet; 
       
    private JFXGParser _parser; 
        
    private Controller _controller;

    private JFileChooser _chooser;
        
    public Application() {
        _sheet = makeDefaultSheet();
        _parser = new JFXGParser();
        _parser.bindThis(this);

        _controller = new Controller(_sheet);

        _sheet.setUpdateHandler(new SetProperty<Controller,Boolean>(Controller.class,"SheetDirty",
                                                                    Boolean.TYPE)
                                    .bind(_controller, Boolean.TRUE));
        
        // TODO: RunnabableFunctor?
        createUI(_sheet, _controller);
    }

   
    public Spreadsheet makeDefaultSheet() {
        Spreadsheet sheet = new Spreadsheet(16,16);
        sheet.setPreferredScrollableViewportSize(new Dimension(400,250));
        sheet.setEditableByDefault(true);

        sheet.setRowSelectionInterval(0,0);
        sheet.setColumnSelectionInterval(0,0);

        return sheet;
    }

    
    public final void createUI(final Spreadsheet sheet, Controller controller) {
        final JPopupMenu popupMenu = new JPopupMenu("Popup Menu");
        popupMenu.add(new JMenuItem(controller.getCellRenameCmd()));
        popupMenu.add(new JMenuItem(controller.getCellFormatCmd()));
        popupMenu.add(new JMenuItem(controller.getCellTypeCmd()));

        // Add a rightclick mouse listener with a 'standard' popup menu
        sheet.addMouseListener(new MouseAdapter() {
                public void mousePressed(MouseEvent e) {
                    // Right clicking should update selection
                    if (e.getButton() == MouseEvent.BUTTON3) {
                        Spreadsheet sht = (Spreadsheet) e.getComponent();
                        Point p = e.getPoint();
                        int row = sht.rowAtPoint(p);
                        int col = sht.columnAtPoint(p);
                        sht.setRowSelectionInterval(row,row);
                        sht.setColumnSelectionInterval(col,col);
                        
                        if (e.isPopupTrigger()) {
                            popupMenu.show(e.getComponent(), p.x, p.y);
                        }                
                    }
                }
                public void mouseReleased(MouseEvent e) {
                    if (e.isPopupTrigger()) {
                        popupMenu.show(e.getComponent(), e.getX(), e.getY());
                    }
                }
            });
        
        // Setup a status bar, and wire it to the spreadsheet's status functor
        JLabel statusLabel = new JLabel("cell(0,0)");
        sheet.setStatusHandler(setProperty(JLabel.class, "Text", String.class).bind1st(statusLabel));

        // Setup the outer frame
        _frame = new JFrame("Application");
        _frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
        _frame.addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent e) {
                    closeWorksheet();
//                     System.out.println("Calling System.exit()");
                    System.exit(0);
                }
            });
        
        // wire up the various functors that implement user interaction
        controller.setLoadFunctor(buildLoadFunctor());
        controller.setSaveFunctor(buildSaveFunctor());
        controller.setErrorFunctor(buildErrorFunctor());
        controller.setPromptFunctor(buildPromptFunctor());
        controller.setConfirmFunctor(buildConfirmFunctor());
              
        // Setup the frame's content pane
        Container rootPane = _frame.getContentPane();
        rootPane.setLayout(new BorderLayout(5,5));

        JScrollPane pane = new JScrollPane(sheet);
        rootPane.add(pane, BorderLayout.CENTER);
        rootPane.add(statusLabel, BorderLayout.SOUTH);

        // Setup the frame's menu
        JMenu fileMenu = new JMenu("File");
        fileMenu.add(controller.getFileNewCmd());
        fileMenu.add(controller.getFileOpenCmd());
        fileMenu.add(controller.getFileSaveCmd());
        fileMenu.add(controller.getFileSaveAsCmd());
        fileMenu.add(getFileExitCmd());

        JMenu defaultMenu = new JMenu("Default");
        defaultMenu.add(controller.getDefaultEditableCmd());
        defaultMenu.add(controller.getDefaultTypeCmd());
        defaultMenu.add(controller.getDefaultValueCmd());
//         defaultMenu.add(controller.getDefaultFormatCmd());
//         defaultMenu.add(controller.getDefaultRendererCmd());
//         defaultMenu.add(controller.getDefaultEditorCmd());
        
        JMenu sheetMenu = new JMenu("Worksheet");
        sheetMenu.add(controller.getSheetColumnsCmd());
        sheetMenu.add(controller.getSheetRowsCmd());
        sheetMenu.add(defaultMenu);        
        sheetMenu.add(controller.getImportClassCmd());
        
        JMenu cellMenu = new JMenu("Cell");
        cellMenu.add(controller.getCellRenameCmd());
//         cellMenu.add(controller.getCellTypeCmd());
//         cellMenu.add(controller.getCellFormatCmd());
//         cellMenu.add(controller.getCellValidatorCmd());
//         cellMenu.add(controller.getCellRendererCmd());
//         cellMenu.add(controller.getCellEditorCmd());
        
        JMenuBar menu = new JMenuBar();
        menu.add(fileMenu);
        menu.add(sheetMenu);
        menu.add(cellMenu);

        // show the frame
        _frame.setJMenuBar(menu);
        _frame.pack();
        _frame.setVisible(true);
    }
   
    // ----------------------
    // Spreadsheet UI Methods
    // ----------------------

    /**
     * Returns an Action that closes the spreadsheet.
     */
    public Action getFileExitCmd() {
        UnaryFunctor<ActionEvent,?> fn = new Project2nd<ActionEvent,Object>()
            .generate2nd(invokeNoArgMethod(Application.class,"closeWorksheet").bind(this));

        return new GenericAction(fn, "Exit");
    }

    // ----------------------
    // Spreadsheet IO Methods
    // ----------------------

    public int loadFile(Spreadsheet sheet) {
        if(getChooser().showOpenDialog(_frame) != JFileChooser.APPROVE_OPTION)
            return Controller.CANCEL_OPTION;
        
        File file = getChosenFile();
        if (file == null)
            return Controller.CANCEL_OPTION;

        try {
            FileInputStream fis = new FileInputStream(file);
            sheet.readSpreadsheet(fis);
            _controller.setSheetSource(file.toURI().toURL());
            return Controller.YES_OPTION;
        }
        catch (IOException x) {
            _controller.notify(x.getMessage(), Controller.getExceptionName(x));
            return Controller.CANCEL_OPTION;
        }
    }
    
    
    public int saveFile(Spreadsheet sheet, boolean promptForName) {
        URL hint = _controller.getSheetSource();
        if (promptForName || hint == null)
            if (getChooser().showSaveDialog(_frame) != JFileChooser.APPROVE_OPTION)
                return Controller.CANCEL_OPTION;
        
        File file = getChosenFile();
        if (file != null) {
            try {
                FileOutputStream fos = new FileOutputStream(file);
                _controller.setSheetSource(file.toURI().toURL());
                _sheet.writeSpreadsheet(fos);
                fos.close();
                _controller.setSheetDirty(false);
                return Controller.YES_OPTION;
            }
            catch (IOException x) {
                Throwable t = Controller.getRootCause(x);
                _controller.notify(x.getMessage(), Controller.getExceptionName(t));
            }
        }
        
        return Controller.CANCEL_OPTION;
    }

    // ============================================================
    // ============================================================
    

    public void closeWorksheet() {
        int ans = Controller.YES_OPTION;
        if (_controller.isSheetDirty()) {
            ans = _controller.promptAndSave();
        }

        if (ans != Controller.CANCEL_OPTION) {
            _frame.dispose();
        }
    }
        

    private JFileChooser getChooser() {
        if (_chooser != null)
            return _chooser;

        File pwd = new File(".");
        _chooser = new JFileChooser(pwd);

//         // Note: source for ExampleFileFilter can be found in FileChooserDemo,
//         // under the demo/jfc directory in the Java 2 SDK, Standard Edition.
//         ExampleFileFilter filter = new ExampleFileFilter();
//         filter.addExtension("jpg");
//         filter.setDescription("JPG & GIF Images");
//         chooser.setFileFilter(filter);

        return _chooser;
    }


    private File getChosenFile() {
        File dir = _chooser.getCurrentDirectory();
        if (dir == null)
            return null;
     
        return _chooser.getSelectedFile();
    }


    // =========================
    // protected API
    // =========================

    protected Controller getController() { return _controller; }
    
    // =========================
    // Implementation Details
    // =========================
    
    @SuppressWarnings("unchecked")
    // "Type safety : A generic array of BinaryFunctor<String,String,?> is created for a 
    // varargs parameter"
    public BinaryFunctor<String,String,String> buildPromptFunctor() {
        // This gets a little hairy.  We have have to manually assemble the arguments
        // to the JOptionPane.showInputDialog class since we want to bind to the component
        // that is available here and leave the two arguments to the caller.
        BinaryFunctor<JOptionPane,Object[],String> showInput = 
            invokeMethod(JOptionPane.class, "showInputDialog", 
                            Component.class, Object.class, Object.class);
        
        ApplyBinary<String,String> threeArgs =
            new ApplyBinary<String,String>(
                 new ConstantBinary<String,String,Component>(_frame),
                 new Project1st<String,String>(),
                 new Project2nd<String,String>()
            );
            
        return showInput.bind1st(null).compose(threeArgs);
    }

    
    @SuppressWarnings("unchecked")
    // "Type safety : A generic array of BinaryFunctor<String,String,?> is created for a 
    // varargs parameter"
    public BinaryFunctor<String,String,?> buildErrorFunctor() {
        // This is fairly similar to the PromptFunctor.  This one requires
        // four parms, two of which are constants.
        BinaryFunctor<JOptionPane,Object[],Integer> showError = 
            invokeMethod(JOptionPane.class, "showMessageDialog", 
                            Component.class, Object.class, String.class, Integer.TYPE);

        ApplyBinary<String,String> fourArgs =
            new ApplyBinary<String,String>(
                 new ConstantBinary<String,String,Component>(_frame),
                 new Project1st<String,String>(),
                 new Project2nd<String,String>(),
                 new ConstantBinary<String,String,Integer>(JOptionPane.ERROR_MESSAGE)
            );

        return showError.bind1st(null).compose(fourArgs);
    }

    
    @SuppressWarnings("unchecked")
    // "Type safety : A generic array of BinaryFunctor<String,String,?> is created for a 
    // varargs parameter"
    public BinaryFunctor<String,String,Integer> buildConfirmFunctor() {
        BinaryFunctor<JOptionPane,Object[],Integer> showConfirm = 
            invokeMethod(JOptionPane.class, "showConfirmDialog",
                            Component.class, Object.class, String.class, Integer.TYPE);

        ApplyBinary<String,String> fourArgs =
            new ApplyBinary<String,String>(
                 new ConstantBinary<String,String,Component>(_frame),
                 new Project1st<String,String>(),
                 new Project2nd<String,String>(),
                 new ConstantBinary<String,String,Integer>(JOptionPane.YES_NO_CANCEL_OPTION)
            );

        return showConfirm.bind1st(null).compose(fourArgs);
    }

    
    public BinaryFunctor<Spreadsheet,Boolean,Integer> buildSaveFunctor() {
        BinaryFunctor<Application,Object[],Integer> getSave = 
            invokeMethod(Application.class, "saveFile", Spreadsheet.class, Boolean.TYPE);
        
        return getSave.bind1st(this).compose(new ArrayBinary<Spreadsheet,Boolean>());
    }


    public UnaryFunctor<Spreadsheet,Integer> buildLoadFunctor() {
        BinaryFunctor<Application,Object[],Integer> getLoad = 
            invokeMethod(Application.class, "loadFile", Spreadsheet.class);
        
        return getLoad.bind1st(this).compose(new ArrayUnary<Spreadsheet>());
    }

    // ------------------------
    // Standalone entry point
    // ------------------------

    static public void main(String[] args) {
        printStartupHeader();

        try {
            UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
        }
        catch (Exception x) {
            // TODO: log this instead of simply dumping it
            System.err.println("Error loading L&F:" +x);
        }
        
        // TODO: do command line processing

        new Application();
    }
    

    static private void printStartupHeader() {
        System.out.println("");
        System.out.println("/**");
        System.out.println(" * A Java Hacker's Worksheet");
        System.out.println(" * Copyright (c) 2004-2007  David A. Hall");
        System.out.println(" */");
        System.out.println("");
    }
}
