// ============================================================================
// $Id: Cell.java,v 1.14 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2004-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.swing.spreadsheet;

import java.awt.Component;
import java.awt.Point;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Observable;
import java.util.Observer;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.DefaultCellEditor;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;

import net.sf.jga.fn.AdaptorVisitor;
import net.sf.jga.fn.EvaluationException;
import net.sf.jga.fn.Generator;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.fn.adaptor.Constant;
import net.sf.jga.parser.ParseException;
import net.sf.jga.swing.GenericTableCellRenderer;
import net.sf.jga.util.Formattable;

/**
 * A single cell in spreadsheet.  Encapsulates information about the generator
 * that provides the value of the cell, the cell's formatting information, and
 * the editor and/or renderer used to display/update the cell's contents.
 * <p>
 * Copyright &copy; 2004-2005  David A. Hall
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

// Problem: this class can't be generic, as the type T can be changed if the
// user changes the formula to return a different type.  An alternative is to
// allow this type to remain generic, and when the user enters a new formula,
// a new instance if generated.  In this case, we'd need the new cell to be
// able to adopt the listener list of the old cell, and the listener list in
// the base class (Observable) is private.

// TODO: implement the strictTyping flag of the enclosing spreadsheet. When
// strictly typed, a cell may reject a value/formula if it is not of the type
// already set (the first value/formula assigned to an undefined cell will
// never be rejected if it is not of the default type, however)

public class Cell extends Observable implements Observer {

    private Spreadsheet _parent;
    
//     // the type that the cell holds: typechecking has to be implemented
//     // in this class, as the table can represent a heterogenous collection
//     // of cells
//     private Class _type;

    // The name of the cell
    private String _name;
    
    // caches the value held by the cell
    private Object _value;
    
    // the functor that returns the value of the cell: in the simple case,
    // it'll be a Constant<T>, but it can be an arbitrarily complex functor
    private Generator<?> _generator;

    // the formula that describes the value of the cell: this string is
    // parsed to create the Generator
    private String _formula;
    
    // The address of this cell (for the time being, rows & columns won't be
    // inserted (we'll need to update the address of a cell or provide a map)
    private Point _addr;

    // Flag that controls editablilty: we can make cells read-only
    private boolean _editable;

    // Component used to render cells in the paint method 
    private TableCellRenderer _renderer;

    // Component used to edit cell values
    private TableCellEditor _editor;

    // The last exception caught by the cell
    private Throwable _exception;
    
    // The string shown on the UI when a cell encounters an error
    private String _errMsg;

    // Implies that a dependency value is not of the expected type
    static public final String CLASS_CAST_ERR   = "### CLASS ###";

    // Implies that a formula encounters an error when evaluated 
    static public final String EVALUATION_ERR   = "### EVAL ###";

    // Implies that a cell's formula depends directly or indirectly on itself
    static public final String CIRCULAR_REF_ERR = "### CIRC ###";

    // Implies that a cell's formula encounted an unexpected null value when evaluated
    static public final String NULL_POINTER_ERR = "### NULL ###";

    // Implies that a cell reference cannot be resolved
    static public final String REFERENCE_ERR    = "### REF ###";

    // Implies that a cell has been referenced before it is initialized
    static public final String UNDEFINED_ERR    = "### UNDEF ###";

    // Implies that a cell's formula cannot be parsed: nothing past the point in the
    // string representation of the formula where the parser exception was thrown has
    // been analyzed.
    static public final String PARSER_ERR       = "### PARSE ###";

    // The set of cells on which this cell depends.
    private Set<Cell> _dependencies = new HashSet<Cell>();

    // TODO: move this into spreadsheet model (cell should only know about Generators:
    // it would be up to the model to parse string formulas and set the cell's
    // generator or error state)
    
    /**
     * Builds a possibly editable cell at the given address, using the formula to build
     * a generator to retrieve the value.  
     */
    Cell(Spreadsheet parent, Point addr, String formula, boolean editable) {
        if (formula == null || formula.trim().length() <= 0)
            throw new IllegalArgumentException("Formula must be supplied");

        // TODO: delete this temporary hack
//         _type = Object.class;

        _parent = parent;
        _addr = addr;
        _editable = editable;
        setFormula(formula);

//         if (isValid())
//             _type = getParser().getReturnType();
        
        _renderer = new Renderer<Object>(); //GenericTableCellRenderer();
    }

    /**
     * Builds a read-only cell at the given address, using the generator to retrieve the
     * value. 
     */
    <T> Cell(Spreadsheet parent, Class<T> type, Point addr, Generator<T> value) {
        _parent = parent;
//         _type = type;
        _addr = addr;
        _editable = false;
        _renderer = new Renderer<T>(); // GenericTableCellRenderer();

        setFormulaImpl(value);
    }

    Cell(Spreadsheet parent, int row, int col) {
        _parent = parent;
        _addr = new Point(row, col);
        _editable = true;
        _errMsg = UNDEFINED_ERR;
//         _type = parent.getDefaultType();
        _value = parent.getDefaultCellValue();
    }

    Spreadsheet getParent() { return _parent; }

    private Spreadsheet.Parser getParser() { return _parent._parser; }

    public String getName() { return _name; }

    
    void setName(String name) { _name = name; }
    
    /**
     * Returns the type of values held by the cell.
     */
    public Class<?> getType() { return _value == null ? Object.class : _value.getClass(); }

    /**
     * Returns the formula that describes the cell's contents.
     */
    String getFormula() { return _formula; }

    /**
     * Returns true if the cell is editable, false if it is read-only.
     */
    public boolean isEditable() { return _editable; }


    /**
     * Sets the cell as editable (or not).
     */
    public void setEditable(boolean b) { _editable = b; }
    
    /**
     * Returns the address of the cell.
     */
    public Point getAddress() { return _addr; }

    /**
     * Returns the renderer used to paint the cell
     */
    public TableCellRenderer getRenderer() { return _renderer; }

    /**
     * Sets the component used to paint the cell
     */
    public void setRenderer(TableCellRenderer renderer) {
        _renderer = renderer;
    }
   
    /**
     * Returns the component used to edit the cell
     */
    public TableCellEditor getEditor() {
        if (_editor == null)
            _editor = new Editor();
        
        return _editor;
    }
    
    /**
     * Sets the component used to edit the cell
     */
    public void setEditor(TableCellEditor editor) {
        _editor = editor;
    }

    /**
     * Returns true if the cell's value can be retrieved without error
     */
    public boolean isValid ()   {
        return _errMsg == null;
    }

    // TODO: fix this: it's bad to use _errMsg in this way, as the first thing
    // that setFormula does is clear this information.  (It might be OK, but if
    // so, its more of an accident than anything)
    
    /**
     * Returns true if the cell's value is undefined
     */
    public boolean isUndefined ()   {
        return _errMsg == UNDEFINED_ERR;
    }

    /**
     * Returns the textual description of the last error associated with cell.
     * The string will be used by the UI.  If the cell's value can be retrieved
     * without error, then this field will be null.
     */
    public String getErrorMsg() { return _errMsg; }

    /**
     * Puts the cell into the error display mode: the invalid formula is preserved
     * as a string, and the error message is set
     */
    @SuppressWarnings("deprecation")
    private void setError(Throwable t, String formula, String msg) {
        Logger.global.log(Level.FINE, formula, t);
        _exception = t;
        _parent.setStatus(t.getMessage());

//         _type = String.class;
        _formula = formula;
        _errMsg = msg;

        _value = t;
        notifyObservers();
    }

    /**
     * Puts the cell into the standard mode.
     */
    private void clearError() {
        _exception = null;
        _errMsg = null;
    }

    /**
     * Returns the value of the cell.  If an exception is caught while
     * evaluating the cell, then this will return null and the getErrorMsg()
     * method will return a description of the problem.
     */
    public Object getValue() {
        return _value;
    }

    
    /**
     * Evaluates the cell's formula and caches the result.  This can cause the cell to enter
     * an error state, if the forumula throws one of several exceptions.
     */
    private void setValue() {
        try {
            _value = _generator.fn();
            clearError();
        }
        catch (CircularReferenceException x) { setError(x, _formula, CIRCULAR_REF_ERR); } 
        catch (InvalidReferenceException x)  { recover (x, _formula, REFERENCE_ERR); }
        catch (NullPointerException x)       { setError(x, _formula, NULL_POINTER_ERR); } 
        catch (ClassCastException x)         { recover (x, _formula, CLASS_CAST_ERR); }
        catch (EvaluationException x)        { recover (x, _formula, EVALUATION_ERR); }
        catch (IllegalArgumentException x)   { recover (x, _formula, EVALUATION_ERR); }
        catch (Exception x)                  { setError(x, _formula, EVALUATION_ERR); }
    }


    /**
     * Clears the contents of the cell, setting to the default type and value.
     */

    public void clear() {
        doClear(null, _parent.getDefaultCellValue());
    }

    /**
     * Clears the contents of the cell, and sets its state to an invalid state to
     * force all dependant cells to reparse.
     */

    void unlink() {
        doClear(REFERENCE_ERR, new InvalidReferenceException("Cell("+_addr.x+","+_addr.y+")"));
    }

    private void doClear(String msg, Object value) {
        unregister();
        
        _editable = true;
//         _type = parent.getDefaultType();
        _value = value;
        _generator = null;
        _formula = null;
        _errMsg = msg;
        
        notifyObservers();
    }


    /**
     * Sets the contents of the cell to the given value.  
     */
    public final <V> void setValue(V value) {
        if (value instanceof Generator<?>)
            setFormula((Generator<?>) value);

        // If the value we receive is the current exception, then we're at the end
        // of an editing process that went bad.  The Editor has set the cell's state
        // to reflect the error, and we don't want to undo that here.
        else if (value != _exception)
            setFormula(new Constant<V>(value));
    }
    
    /**
     * Sets the contents of the cell to the given formula.
     */
    public final void setFormula(String formula) throws CircularReferenceException {
        if (formula == null || "".equals(formula)) {
            clear();
            return;
        }

        if (formula.startsWith("="))
            formula = formula.substring(1);
     
        try {
            Generator<?> gen = getParser().parseGenerator(this, formula);
            //             _type = getParser().getReturnType();
            setFormula(gen);
            _formula = formula;
        }
        catch (ParseException x)       { setError(x, formula, PARSER_ERR);  }
        catch (ClassCastException x)   { recover(x, formula, CLASS_CAST_ERR); }
        catch (NullPointerException x) { setError(x, formula, NULL_POINTER_ERR); }
    }

    // TODO: define a version of setFormula that the spreadsheet to use: it must do
    // all of this and also set the _formula member to something meaningful
    
    /**
     * Sets the contents of the cell to the given formula, notifying all observers of
     * the change.
     */
    public final void setFormula(Generator<?> formula) {
        clearError();
        unregister();
        setFormulaImpl(formula);
        notifyObservers();
    }


    /**
     * Checks the given formula to prevent circular dependencies, and sets the contents of the
     * cell to the given formula.  A circular reference causes the cell to enter an error state,
     * and its old value will be retained.
     */
    private void setFormulaImpl(Generator<?> formula) {
        try {
            _generator = register(formula);
            setValue();
        }
        catch (CircularReferenceException x) { setError(x, formula.toString(), CIRCULAR_REF_ERR); }
    }
    

    /**
     * Attempts one recovery from the given exception.  The formula stored in the cell
     * is reparsed, to allow recovery in cases where there is a type mismatch between the
     * old formula functor and new values in input cells.
     */
    synchronized private void recover(Exception x, String formula, String msg) {
        if (_inRecovery) {
            setError(x, formula, msg);
        }
        else {
            _inRecovery = true;
            setFormula(_formula);
            if (!isValid() && _errMsg.equals(PARSER_ERR)) {
                setError(x, formula, msg);
            }
            
            _inRecovery = false;
        }
    }

    // guards against re-entering the recover method
    private boolean _inRecovery = false;

    
    /**
     * Sets the format of the cell, if formatting is supported by the cell's
     * renderer.  If formatting is not supported, invoking this method will
     * have no effect.
     * @throws ClassCastException if the generic type of the formatter is not
     * compatable with the generic type of the cell's renderer
     */
    @SuppressWarnings("unchecked")
    // We can't instance check using the generic parm, so Formattable must be a 
    // rawtype.
    public <T> void setFormat(UnaryFunctor<T,String> formatter) {
        if (_renderer instanceof Formattable)
            ((Formattable) _renderer).setFormat(formatter);
    }


    /**
     * Returns a reference to the cell's formula.  
     */
    @SuppressWarnings("unchecked")
    public Generator<?> getReference() {
        return new CellReference(this);
    }

    /**
     * Returns true if this cell references the given cell, either directly or
     * indirectly.  Also returns true if this cell <i>is</i> the given cell.
     */
    public boolean references(Cell c) {
        if (getAddress().equals(c.getAddress())) {
            return true;
        }

        for (Cell c1 : _dependencies) {
            if (c1 == c) {
                return true;
            }
            if (c1.references(c)) {
                return true;
            }
        }

        return false;
    }


    /**
     * returns an iterator over all cells on which this cell directly depends
     */
    public final Iterator<Cell> dependsOn() {
        return _dependencies.iterator();
    }
    

    public String toString() {
        StringBuffer buf = new StringBuffer(256);
        buf.append("Cell(").append(_addr.x).append(",").append(_addr.y).append(") ");
        if (_name != null) {
            buf.append('"').append(_name).append('"').append(' ');
        }
        buf.append("[");
        if (_exception != null)
            buf.append(_exception.getMessage());
        else if (_formula != null)
            buf.append(_formula);
        else
            buf.append(_value);

        return buf.append("]").toString();
    }


    // - - - - - - - - - - - -
    // Observer Interface
    // - - - - - - - - - - - -

    public void update(Observable observable, Object object) {
        setValue();
        notifyObservers();
    }

    /**
     * Triggers cells (and other observers) that this cell's contents have been
     * changed.
     */
    public void notifyObservers() {
        setChanged();
        notifyObservers(_value);
        clearChanged();
    }

    // - - - - - - - - - - - -
    // Registration details
    // - - - - - - - - - - - -

    // Examines the current value generator, adding this as an observer to
    // any cells on which this depends
    private <T> Generator<T> register(Generator<T> formula) throws CircularReferenceException {
        formula.accept(new RegistrationVisitor());
        return formula;
    }

    
    // Examines the current value generator, removing this as an observer to
    // any cells on which this depends.
    private void unregister() {
        if (_generator != null)
            _generator.accept(new UnregistrationVisitor());
        
        _dependencies.clear();
    }

    private class RegistrationVisitor extends AdaptorVisitor
        implements CellReference.Visitor
    {
        public void visit(CellReference<?> host) {
            Cell hostCell = host.getCell();
            
            // check for circular references
            if (hostCell.references(Cell.this))
                throw new CircularReferenceException();

            host.register(Cell.this);
            _dependencies.add(hostCell);
        }
    }

    private class UnregistrationVisitor extends AdaptorVisitor
        implements CellReference.Visitor
    {
        public void visit(CellReference<?> host) {
            host.unregister(Cell.this);
        }
    }

    // - - - - - - - - - - -
    // Editor class
    // - - - - - - - - - - -

    static class Editor extends DefaultCellEditor {
        
        static final long serialVersionUID = -9010644547311806213L;

        private Spreadsheet _sheet;
        private Cell _cell;
        
        public Editor() {
            super(new JTextField());
        }
     
        public Object getCellEditorValue() {
            // The superclass' editorComponent is the textField we passed at construction
            // the super class returns the text fields contents.  we'll parse it and pass
            // the result
            String formula = (String) super.getCellEditorValue();
            if (formula.startsWith("="))
                formula = formula.substring(1);
        
            try {
                return _sheet._parser.parseGenerator(_cell, formula);
            }
            catch(ParseException x) {
                _cell.setError(x, formula, PARSER_ERR);
                return x;
            }
        }
        
        public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected,
                                                     int row, int col)
        {
            _sheet = (Spreadsheet) table;
            _cell = _sheet.getCellAt(row, col);
            return super.getTableCellEditorComponent(table,_cell._formula,isSelected,row,col);
        }
    }


    // - - - - - - - - - - -
    // Renderer class
    // - - - - - - - - - - -

    static class Renderer<T> extends GenericTableCellRenderer<T> {
        static private final long serialVersionUID = 5595981545922985168L;

        public Component getTableCellRendererComponent(JTable table,Object value,boolean isSelected,
                                                       boolean hasFocus, int row, int column)
        {
            Spreadsheet sheet = (Spreadsheet) table;
            Cell cell = sheet.getCellIfPresent(row,column);
            boolean editable = (cell != null) ? cell.isEditable() : sheet.isEditableByDefault();
            return super.getTableCellRendererComponent(table, value, isSelected && editable,
                                                       hasFocus && editable, row, column);
        }
    }
}

// - - - - - - - - - - -
// Cell Reference class
// - - - - - - - - - - -

class CellReference<R> extends Generator<R> {

    static final long serialVersionUID = -2552950462394369940L;
    
    private Cell _cell;

    public CellReference(Cell cell) { _cell = cell; }

    @SuppressWarnings("unchecked")
    public R fn() {
        Object obj = _cell.getValue();
        if (obj instanceof RuntimeException)
            throw (RuntimeException) obj;

        return (R)obj;
    }

    public Cell getCell() { return _cell; }

    public void register(Observer obs) {
        _cell.addObserver(obs);
    }
    
    public void unregister(Observer obs) {
        _cell.deleteObserver(obs);
    }
    
    public String toString() { return "CellReference("+_cell+")"; }

    public void accept(net.sf.jga.fn.Visitor v) {
        if (v instanceof CellReference.Visitor)
            ((CellReference.Visitor)v).visit(this);
        else
            v.visit(this);
    }

    public interface Visitor extends net.sf.jga.fn.Visitor {
        public void visit(CellReference<?> host);
    }
}

class CircularReferenceException extends RuntimeException {
    static final long serialVersionUID = -8670923266092229864L;    
    public String getMessage() { return "Circular Reference"; }
}

class InvalidReferenceException extends RuntimeException {
    static final long serialVersionUID = -7825481889728025018L;
    public InvalidReferenceException(String msg) { super(msg); }
}
    
