// ============================================================================
// $Id: WebStart.java,v 1.6 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2004-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.swing.spreadsheet;

import java.io.IOException;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.net.URL;
import javax.jnlp.FileContents;
import javax.jnlp.FileOpenService;
import javax.jnlp.FileSaveService;
import javax.jnlp.ServiceManager;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

/**
 * An JNLP-based application wrapper for Spreadsheet, providing a main method to
 * allow for standalone use.  This version supplies input and output stream
 * services based on the WebStart API.
 * <p>
 * Copyright &copy; 2004-2005  David A. Hall
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

public class WebStart extends Application {

    private FileContents _fc;

    // ----------------------
    // Spreadsheet IO Methods
    // ----------------------

    /**
     * Loads the file via the JNLP FileOpen service
     */
    public int loadFile(Spreadsheet sheet) {
        final Controller controller = getController();
        try { 
            FileOpenService fos = (FileOpenService)
                ServiceManager.lookup("javax.jnlp.FileOpenService"); 
            if (fos == null) {
                return Controller.CANCEL_OPTION;
            }
         
            // ask user to select a file through this service 
            _fc = fos.openFileDialog(null, null);
            if (_fc == null)
                return Controller.CANCEL_OPTION;
            
            sheet.readSpreadsheet(_fc.getInputStream());
            controller.setSheetSource(new URL("file:///"+_fc.getName()));
            controller.setSheetDirty(false);
            return Controller.YES_OPTION;
        }
        catch (Exception x) {
            displayError(x);
            return Controller.CANCEL_OPTION;
        }

    }


    /**
     * Writes the file via the JNLP FileSave service
     */
    public int saveFile(Spreadsheet sheet, boolean promptForName) {
        PipedInputStream pipeIn = null;
        PipedOutputStream pipeOut = null;
        try {
            // Get a handle to the necessary service.  
            FileSaveService fss = (FileSaveService)
                ServiceManager.lookup("javax.jnlp.FileSaveService");
            if (fss == null) {
                return Controller.CANCEL_OPTION;
            }

            // use a PipeInput/PipeOutput to store the file.  One or the other should
            // be used on a separate thread.
            pipeIn = new PipedInputStream();
            pipeOut = new PipedOutputStream(pipeIn);

            // Start up a background thread to write the spreadsheet to a pipe
            Thread writeThread = new Thread(new Writer(sheet, pipeOut));
            writeThread.setDaemon(true);
            writeThread.start();

            //See if we currently have any idea where the spreadsheet came from.
            Controller controller = getController();
            String hint = getClue(_fc, controller.getSheetSource());

            // Prompt the user for the file to write to.  By the time the dialog has come
            // up, the write thread may already be completed (if the document is smaller
            // than the pipe buffer).
            if (promptForName && _fc != null) {
                _fc = fss.saveAsFileDialog(hint, new String[]{ "hwks", }, _fc);
            }
            else {
                _fc = fss.saveFileDialog(hint, new String[]{ "hwks", }, pipeIn, hint);
            }
            
            if (_fc == null) {
                return Controller.CANCEL_OPTION;
            }

            // update the spreadsheet state associated with file i/o
            controller.setSheetSource(new URL("file:///"+_fc.getName()));
            controller.setSheetDirty(false);
            pipeIn.close();
            return Controller.YES_OPTION;
        }
        catch (Exception x) {
            displayError(x);
            return Controller.CANCEL_OPTION;
        }
    }

    /**
     */

    private String getClue(FileContents fc, URL source) {
        if (fc != null) 
            try { return fc.getName(); }
            catch (IOException x) {} // just fall through: failure _is_ an option in this case

        if (source != null)
            return source.getPath();

        return "worksheet.hwks";
    }

    
    /**
     * Writes the spreadsheet contents
     */
    private class Writer implements Runnable {
        private PipedOutputStream _os;
        private Spreadsheet _sheet;
        public Writer(Spreadsheet sheet, PipedOutputStream os) {
            _os = os; _sheet = sheet;
        }
        public void run() {
            try {
                _sheet.writeSpreadsheet(_os);
            }
            catch (Exception x) {
                displayError(x);
            }
            
            try {
                _os.close();
            }
            catch (IOException x) {
                displayError(x);
            }
        }
    }
    

    /**
     * Displays an error message to the user, via whatever method the Controller
     * has been configured with.
     */
    private synchronized void displayError(final Exception x) {
        // make sure that we're on the event thread.
        if (!SwingUtilities.isEventDispatchThread()) {
            SwingUtilities.invokeLater(new Runnable() {
                    public void run () {
                        displayError(x);
                    }
                });
        }
        else {
            Controller controller = getController();
            String msg = x.getMessage();
            if (msg == null || msg.length() == 0)
                msg = Controller.getExceptionName(x);
        
            controller.notify(msg, Controller.getExceptionName(x));
        }
    }
    

    // ------------------------
    // Standalone entry point
    // ------------------------

    static public void main(String[] args) {
        printStartupHeader();

        try {
            UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
        }
        catch (Exception x) {
            // TODO: log this instead of simply dumping it
            System.err.println("Error loading L&F:" +x);
        }
     
        // TODO: do command line processing

        new WebStart();
    }
    

    static private void printStartupHeader() {
        System.out.println("");
        System.out.println("/**");
        System.out.println(" * A Java Hacker's Worksheet: JNLP Edition");
        System.out.println(" * Copyright (c) 2005  David A. Hall");
        System.out.println(" */");
        System.out.println("");
    }
}
