// ============================================================================
// $Id: Algorithms.java,v 1.47 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html 
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================
package net.sf.jga.util;

import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import net.sf.jga.algorithms.Compare;
import net.sf.jga.algorithms.Find;
import net.sf.jga.algorithms.ListAlgorithms;
import net.sf.jga.algorithms.Summarize;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.UnaryFunctor;
import net.sf.jga.fn.comparison.Equality;

/**
 * Facade for the Algorithms adapted from STL, defined to work primarily with
 * collections.  These algorithms are adapted from STL, with modifications to be
 * consistent with typical java practice.  For example, typical STL algorithms
 * are defined with pairs of iterators defining a half-open range over some
 * implied collection. It works in C++ because the STL iterators can be compared
 * for equality.  Java iterators are not guaranteed to be comparable to each
 * other by contract, so the same signatures wouldn't work.
 * <p>
 * Typically, where an STL algorithm would take a pair of iterators, we'll take
 * a collection.  Where an STL algorithm would return an iterator, we'll return
 * an iterator.  Note that it will always be java.lang.Iterator when using
 * this class: for some of the more powerful uses, use the Iterators class,
 * which will often return an implementation of Iterator that is tailored for
 * the semantics of the algorithm that was called.
 * <p>
 * The algorithms in this class and the same set of algorithms in the Iterators
 * class will always return the same results when called with identical
 * arguments.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 * @deprecated everything likely to be of value has been moved to static classes in the
 * net.sf.jga.algorithms package
 */

public class Algorithms {
    
    // ----------------------------------------------------------------------
    // ForEach algorithms
    // ----------------------------------------------------------------------

    /**
     * Applies the given UnaryFunctor to every element in the collection, and
     * returns the Functor.  This is useful when the Functor gathers information
     * on each successive call.
     * @return the functor, after it has been called once for every element
     * @deprecated use {@link Summarize#forEach(Iterable,UnaryFunctor)}
     */
    static public <T,R> UnaryFunctor<T,R> forEach(Collection<? extends T> c, UnaryFunctor<T,R> fn) {
        return Summarize.forEach(c, fn);
    }


    // ----------------------------------------------------------------------
    // Collection Equality algorithms
    // ----------------------------------------------------------------------

    /**
     * Returns true if the two collections are equal.
     * @return true if the two collections are equal
     * @deprecated use c1.equals(c2) instead
     */
    static public <T> boolean equal(Collection<? extends T> c1, Collection<? extends T> c2) {
        return c1.equals(c2);
    }

    
    /**
     * Returns true if the two collections are equal, using the given comparator
     * to compare the elements in each collection
     * @return true if the two collections are equal.
     * @deprecated use {@link Compare#equal(Iterable,Iterable,Comparator)} instead
     */
    static public <T> boolean
    equal(Collection<? extends T> c1, Collection<? extends T> c2, Comparator<T> comp) {
        return Compare.equal(c1, c2, comp);
    }

    
    /**
     * Returns true if the two collections are equal, using the given functor
     * to compare the elements in each collection.  The functor is expected to
     * evaluate its two argments and return true if they are "equal", therefore
     * this method returns true if the iterations contain the same number of
     * elements and if the functor returns true for all pairs of elements.
     * @return true if the two collections are equal
     * @deprecated use Compare.equal(Iterable,Iterable,BinaryFunctor) instead
     */
    static public <T> boolean
    equal(Collection<? extends T> c1, Collection<? extends T> c2, BinaryFunctor<T,T,Boolean> eq) {
        return Compare.equal(c1, c2, eq);
    }
                                 
    // ----------------------------------------------------------------------
    // Collection Comparison algorithms
    // ----------------------------------------------------------------------

    /**
     * Returns true if the first collection is lexically less than the second,
     * using the default comparison operation to compare the elements in each
     * collection.
     * @return true if c1 < c2
     * @deprecated use Compare.lessThan(Iterable,Iterable) instead
     */
    static public <T extends Comparable<? super T>> boolean
    lessThan(Collection<? extends T> c1, Collection<? extends T> c2) {
        return Compare.lessThan(c1, c2);
    }

    
    /**
     * Returns true if the first collection is lexically less than the second,
     * using the given comparator to compare the elements in each collection.
     * @return true if c1 < c2
     * @deprecated use Compare.lessThan(Iterable,Iterable,Comparator) instead
     */
    static public <T> boolean
    lessThan(Collection<? extends T> c1, Collection<? extends T> c2, Comparator<T> comp) {
        return Compare.lessThan(c1, c2, comp);
    }

    
    /**
     * Returns true if the first collection is lexically less than the second,
     * using the given operator to compare the elements in each collection.  The
     * first is less than the second if it is not longer than the second and if
     * the first corresponding element that is not equal is less.
     * @return true if c1 < c2
     * @deprecated use Compare.lessThan(Iterable,Iterable,BinaryFunctor) instead
     */
    static public <T> boolean
    lessThan(Collection<? extends T> c1, Collection<? extends T> c2, BinaryFunctor<T,T,Boolean> lt) {
        return Compare.lessThan(c1, c2, lt);
    }
                                 
    // ----------------------------------------------------------------------
    // Minimum/Maximum algorithms
    // ----------------------------------------------------------------------

    /**
     * Finds the position of the minimum value in a collection using the natural
     * ordering of the collection's elements.
     * @return an iterator from the collection whose next() [if it hasNext()]
     * will return the first instance of the minimum value in the collection.
     * If the collection is empty, then the returned iterator's hasNext() will
     * report false.
     */
    static public <T extends Comparable<? super T>> Iterator<T> minimum(Collection<? extends T> c) {
        return Find.find(c, Summarize.min(c));
    }

    
    /**
     * Finds the position of the minimum value in a collection using the given
     * comparator.
     * @return an iterator from the collection whose next() [if it hasNext()]
     * will return the first instance of the minimum value in the collection.
     * If the collection is empty, then the returned iterator's hasNext() will
     * report false.
     */
    static public <T> Iterator<T> minimum(Collection<? extends T> c, Comparator<T> comp) {
        return Find.find(c, Summarize.min(c,comp));
    }

    
    /**
     * Finds the position of the minimum value in a collection using the given
     * functor to compare elements.  The functor is presumed to return the lesser
     * of its two arguments.
     * @return an iterator from the collection whose next() [if it hasNext()]
     * will return the first instance of the minimum value in the collection.
     * If the collection is empty, then the returned iterator's hasNext() will
     * report false.
     */
    static public <T> Iterator<T> minimum(Collection<? extends T> c, BinaryFunctor<T,T,T> bf) {
        return Find.find(c, Summarize.min(c,bf));
    }

    
    /**
     * Finds the position of the maximum value in a collection using the natural
     * ordering of the collection's elements.
     * @return an iterator from the collection whose next() [if it hasNext()]
     * will return the first instance of the maximum value in the collection.
     * If the collection is empty, then the returned iterator's hasNext() will
     * report false.
     */
    static public <T extends Comparable<? super T>> Iterator<T> maximum(Collection<? extends T> c) {
        return Find.find(c, Summarize.max(c));
    }

    
    /**
     * Finds the position of the maximum value in a collection using the given
     * comparator.
     * @return an iterator from the collection whose next() [if it hasNext()]
     * will return the first instance of the maximum value in the collection.
     * If the collection is empty, then the returned iterator's hasNext() will
     * report false.
     */
    static public <T > Iterator<T> maximum(Collection<? extends T> c, Comparator<T> comp) {
        return Find.find(c, Summarize.max(c,comp));
    }


    /**
     * Finds the position of the maximum value in a collection using the given
     * functor to compare elements.  The functor is presumed to return the 
     * greater of its two arguments.
     * @return an iterator from the collection whose next() [if it hasNext()]
     * will return the first instance of the maximum value in the collection.
     * If the collection is empty, then the returned iterator's hasNext() will
     * report false.
     */
    static public <T> Iterator<T> maximum(Collection<? extends T> c, BinaryFunctor<T,T,T> bf) {
        return Find.find(c, Summarize.max(c, bf));
    }

    
    /**
     * Applies the UnaryFunctor to each element in the list, and replaces
     * each element with the result. This method would, in an ideal world,
     * belong in the Collections class, as its signature is more like the
     * algorithm methods in that class than in Algorithms.
     * @return the modified list
     * @deprecated use {@link ListAlgorithms#transform(List,UnaryFunctor)}
     */
    static public <T> List<T> transform(List<T> lin, UnaryFunctor<T,T> uf) {
        return ListAlgorithms.transform(lin, uf);
    }

    
    /**
     * Tests each element in the list, and replaces elements that pass with
     * the given value.  This method would, in an ideal world, belong in the
     * Collections class, as its signature is more like the algorithm methods
     * in that class than in Algorithms.
     * @return the modified list
     * @deprecated use {@link ListAlgorithms#replace(List,UnaryFunctor,Object)}
     */
    static public <T> List<T> replaceAll(List<T> lin, UnaryFunctor<T,Boolean> uf, T value) {
        return ListAlgorithms.replace(lin, uf, value);
    }

    /**
     * removes all instances of the given element from the list
     * @return the modified list
     * @deprecated use {@link ListAlgorithms#remove(List,Object)}
     */
    static public <T, LT extends List<? extends T>> LT removeAll(LT lin, T value) {
        return ListAlgorithms.remove(lin, value);
    }

    
    /**
     * removes all instances of the given element from the list
     * @return the modified list
     * @deprecated use {@link ListAlgorithms#remove(List,Object)}
     */
    static public <T, LT extends List<? extends T>> LT removeAll(LT lin, T value, Equality<T> eq) {
        return ListAlgorithms.remove(lin, value, eq);
    }

    
    /**
     * removes all elements from the list for which the functor returns TRUE
     * @return the modified list
     * @deprecated use {@link ListAlgorithms#remove(List,UnaryFunctor)}
     */
    static public <T, LT extends List<? extends T>> LT removeAll(LT lin, UnaryFunctor<T,Boolean> uf){
        return ListAlgorithms.remove(lin, uf);
    }

    
    /**
     * removes all adjacent duplicate values in the given list, using equals()
     * to compare adjacent elements
     * @return the modified list
     * @deprecated use {@link ListAlgorithms#unique(List)}
     */
    static public <T, LT extends List<? extends T>> LT unique(LT lin) {
        return ListAlgorithms.unique(lin);
    }

    
    /**
     * removes all adjacent duplicate values in the given list, using the given
     * functor to compare adjacent elements
     * @return the modified list
     * @deprecated use {@link ListAlgorithms#unique(List,BinaryFunctor)}
     */
    static public <T, LT extends List<? extends T>> LT unique(LT lin, BinaryFunctor<T,T,Boolean> eq){
        return ListAlgorithms.unique(lin, eq);
    }
}
