// ============================================================================
// $Id: ArrayUtils.java,v 1.7 2009/09/06 22:30:06 davidahall Exp $
// Copyright (c) 2004-2009  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.util;

import java.lang.reflect.Array;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.NoSuchElementException;

/**
 * Miscellaneous utilities for working with arrays.
 * <p>
 * Copyright &copy; 2004-2009  David A. Hall
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

public class ArrayUtils {

    // Older versions of java lack a method that does this.
    static public String toString(Object[] arr) {
        StringBuffer buf = new StringBuffer();
        buf.append("[");
        for (int i = 0; i < arr.length; ++i) {
            if (i > 0)
                buf.append(',');
            
            buf.append(arr[i]);
        }

        buf.append("]");
        return buf.toString();
    }

    /**
     * Returns an array containing the contents of the original array with the given
     * items inserted at the given idx.
     * @param <T> tye type of objects stored in the input array
     * @param arr the input array
     * @param idx the index at which to insert new items
     * @param items the new items to be inserted.
     */
    @SuppressWarnings("unchecked")
    static public <T> T[] insert(T[] arr, int idx, T... items) {
        int prefixUsed = Math.min(arr.length, idx);
        int prefixSize = Math.max(arr.length, idx);
        int suffixSize = Math.max(arr.length - idx, 0);
        
        T[] result = (T[]) allocate(arr.getClass(), prefixSize + items.length);
        System.arraycopy(arr, 0, result, 0, prefixUsed);
        System.arraycopy(items, 0, result, idx, items.length);
        System.arraycopy(arr, prefixUsed, result, idx + items.length, suffixSize); 
        return result;
    }
    
    /**
     * Returns an array containing the contents of the original array with the given
     * position removed; the size of the resulting array will be one less than the 
     * input array
     * @param <T> tye type of objects stored in the input array
     * @param arr the input array
     * @param idx the index at which to insert new items
     */
    @SuppressWarnings("unchecked")
    static public <T> T[] remove(T[] arr, int idx) {
        return remove(arr, idx, 1);
    }
    

    /**
     * Returns an array containing the contents of the original array with the given
     * positions removed; the size of the resulting array will be <i>count</i> less than the 
     * input array
     * @param <T> tye type of objects stored in the input array
     * @param arr the input array
     * @param idx the index at which to insert new items
     * @throws ArrayIndexOutOfBoundsException if idx >= arr.length
     */
    @SuppressWarnings("unchecked")
    static public <T> T[] remove(T[] arr, int idx, int count) {
        if (idx >= arr.length) {
            throw new ArrayIndexOutOfBoundsException();
        }
        
        if (idx < 0) {
            throw new IllegalArgumentException("Index must be non-negative");
        }
        
        if (count < 0) {
            throw new IllegalArgumentException("Count must be non-negative");
        }
        
        T[] result = (T[]) allocate(arr.getClass(), arr.length - count);
        System.arraycopy(arr, 0, result, 0, idx);
        System.arraycopy(arr, idx + count, result, idx, arr.length - idx - count); 
        return result;
    }
    

    /**
     * Returns an uninitialized array of the given type, sized to the given length. 
     * @param <T> the type of objects to be stored in the array
     * @param type the type of the array (ie, Array of T)
     * @param len the length of the new array
     * @return
     */
    @SuppressWarnings("unchecked")
    static public <T> T[] allocate(Class<? extends T[]> type, int len) {
        return (T[]) Array.newInstance(type.getComponentType(), len);
    }
    
    /**
     * Produce an ArrayIterator over the given array.
     */
    static public <E> ArrayIterator<E> iterate(E... arr) {
        return new ArrayIterator<E>(arr);
    }

    
    /**
     * Adapts the array to the Iterable interface
     */
    static public <T> Iterable<T> iterable(T[] ts) {
        return new ArrayIterable<T>(ts);
    }

    /**
     * Iterable wrapper around a given array.
     */
    static public class ArrayIterable<T> implements Iterable<T> {

        // The array being iterated over
        private T[] _delegate;

        public ArrayIterable(T[] ts) { _delegate = ts; }

        // - - - - - - - - - - - -
        // Iterable implementation
        // - - - - - - - - - - - -
        
        public Iterator<T> iterator() {
            return (_delegate == null) ? new EmptyIterator<T>() : buildIterator(_delegate);
        }
        
        protected Iterator<T> buildIterator(T[] delegate) {
            return new ArrayIterator<T>(_delegate);
        }
    }

    /**
     * Iterates over an array of objects.  Not safe for use by multiple threads.
     */
    static public class ArrayIterator<T> implements ListIterator<T> {
        
        // the array being iterated over
        private T[] _array;

        // the current pointer into the array
        private int _idx;

        // Null on initialization, TRUE when next() was the last move called,
        // FALSE when previous() was the last move called.
        private Boolean _goingForward;
        
        /**
         * Builds an ArrayIterator for the given array
         */
        public ArrayIterator(T[] array) { _array = array; }

        /**
         * Builds an ArrayIterator for the given array, initialized to the given index
         */
        public ArrayIterator(T[] array, int idx) {
            _array = array;
            _idx = idx;
        }
        
        // - - - - - - - - - - - - - -
        // ListIterator implementation
        // - - - - - - - - - - - - - -
        
        public boolean hasNext() { return _idx < _array.length; }

        public boolean hasPrevious() { return _idx > 0; }

        public int nextIndex() { return _idx; }

        public int previousIndex() { return _idx - 1; }

        public T next() {
            if (_idx >= _array.length )
                throw new NoSuchElementException();

            _goingForward = Boolean.TRUE;
            return _array[_idx++];
        }

        public T previous() {
            if (_idx <= 0)
                throw new NoSuchElementException();

            _goingForward = Boolean.FALSE;
            return _array[--_idx];
        }

        public void set(T value) {
            if (_goingForward == null)
                throw new IllegalStateException();

            if (_goingForward.booleanValue()) // to compile w/ 1.4 for retro test
                _array[_idx - 1] = value;
            else
                _array[_idx] = value;
        }

        // Can't support either of these, as an array is a fixed size structure
        public void add(T value) {
            _goingForward = null;
            throw new UnsupportedOperationException();
        }
    
        public void remove() {
            _goingForward = null;
            throw new UnsupportedOperationException();
        }
    }
}
