// ============================================================================
// $Id: CachingIterator.java,v 1.10 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.util;

import java.util.NoSuchElementException;
import java.util.Iterator;

/**
 * Iterator that allows the program to retain the last few elements returned.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

public class CachingIterator<T> implements Iterator<T>, Iterable<T> {
    // The base iterator
    private Iterator<? extends T> _base;

    // ring buffer, to hold the elements that have been consumed
    private Object[] _cache;

    // index of the buffer's base element.  Once the ring buffer loops around,
    // the base pointer indicates the logical 
    private int _baseptr = 0;

    // number of elements in the buffer
    private int _cnt = 0;

    // size of the buffer
    private int _size;

    /**
     * Builds a CachingIterator that can retain 1 element.
     */
      
    public CachingIterator (Iterator<? extends T> base) {
        this(base, 1);
    }

    /**
     * Builds a CachingIterator that can retain the given number of elements.
     * 
     * @throws IllegalArgumentException if max <= 0.
     */
      
    public CachingIterator (Iterator<? extends T> base, int max) {
        if (max <= 0)
            throw new IllegalArgumentException();
        
        _base = (base != null) ? base : new EmptyIterator<T>();
        _size = max;

        _cache = new Object[_size];
    }

    /**
     * Returns true if there have been at least N elements consumed from the
     * underlying iterator.
     */

    public boolean hasCached(int n) {
        return n > 0 && n <= _cnt;
    }
    
    /**
     * Returns the Nth previous element consumed from the underlying iterator.
     */

    @SuppressWarnings("unchecked")
    // The buffer array is private, and the only things that we ever put
    // into it are known to be T objects because they are read from an
    // Iterator<? extends T>, so we know that the cast is always valid
    public T cached(int n) {
        if (n <= 0 || n > _cnt || n > _size)
            throw new NoSuchElementException();

        return (T) _cache[(_baseptr +_size - n) % _size];
    }
    
    /**
     * Returns the cache size
     */
    public int getCacheSize() {
        return _size;
    }
    
    /**
     * Returns the number of elements in the cache 
     */
    public int getCacheCount() {
        return _cnt;
    }
    
    // - - - - - - - - - - -
    // Iterable<T> interface
    // - - - - - - - - - - -

    public Iterator<T> iterator() { return this; }
    
    // - - - - - - - - - - -
    // Iterator<T> interface
    // - - - - - - - - - - -
    
    public boolean hasNext() {
        return _base.hasNext();
    }

    @SuppressWarnings("unchecked")
    // The buffer array is private, and the only things that we ever put
    // into it are known to be T objects because they are read from an
    // Iterator<? extends T>, so we know that the cast is always valid
    public T next() {
        if (_cnt < _size)
            ++_cnt;
        
        int n = _baseptr;
        _baseptr = ++_baseptr % _size;
        _cache[n] =_base.next();
        return (T) _cache[n];
    }


    /**
     * Removes the last item returned from the base iterator, if the base
     * iterator supports the remove operation.  The item is not removed from
     * the cache.
     * @throws UnsupportedOperationException if the base iterator does not
     * support the remove() operation
     */
    public void remove() {
        _base.remove();
    }
}
