// ============================================================================
// $Id: Iterators.java,v 1.46 2007/03/07 03:00:47 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.util;

import java.util.Comparator;
import java.util.Iterator;
import net.sf.jga.algorithms.Compare;
import net.sf.jga.algorithms.Summarize;
import net.sf.jga.fn.BinaryFunctor;
import net.sf.jga.fn.UnaryFunctor;

/**
 * Facade for the Algorithms adapted from STL, defined to work primarily with
 * iterators.  These algorithms are adapted from STL, with modifications to be
 * consistent with typical java practice.  For example, typical STL algorithms
 * are defined with pairs of iterators defining a half-open range over some
 * implied collection.  It works in C++ because the STL iterators can be
 * compared for equality.  Java iterators are not guaranteed to be comparable
 * to each other by contract, so the same signatures wouldn't work.
 * <p>
 * Typically, where an STL algorithm would take a pair of iterators, this facade
 * will take a single iterator and where an STL algorithm would return an
 * iterator, we'll return an iterator.  In this facade, the iterator returned
 * will frequently be specialized in some way: either a ListIterator or one of
 * the iterators defined in jga.
 * <P>
 * It is best to assume that all of the methods in this class may (but are not
 * guaranteed to) advance the iterator argument.  Also, once an iterator has
 * been passed to one of the methods of this class, it should not be used
 * again.  On the other hand, unless otherwise noted, it is safe to pass the
 * iterator returned by one of the methods in this class back to the method
 * that returned it, or to any of the other methods.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 * @deprecated all of the remaining functionality has been moved into 
 */

public class Iterators {
    /**
     * Applies the given UnaryFunctor to every element in the iteration, and
     * returns the Functor.  This is useful when the Functor gathers information
     * on each successive call.
     * @return the functor, after it has been called once for every element
     * @deprecated moved to
     * {@link Summarize#forEach(Iterator,UnaryFunctor) Summarize.forEach(it,fn)}
     */
    static public <T,R> UnaryFunctor<T,R> forEach(Iterator<? extends T> it, UnaryFunctor<T,R> fn) {
        return Summarize.forEach(it, fn);
    }        

    
    /**
     * Returns true if the two iterations are equal, using the Comparable
     * interface to compare elements in the iterations.
     * @return true if the two iterations are equal
     * @deprecated use Compare.equal(Iterator,Iterator) instead
     */
    static public <T extends Comparable<? super T>> boolean
    equal(Iterator<? extends T> iterator1, Iterator<? extends T> iterator2)
    {
        return Compare.equal(iterator1, iterator2);
    }

    
    /**
     * Returns true if the two iterations are equal, using the given Comparator
     * to compare elements in the iterations.
     * @return true if the two iterations are equal
     * @deprecated use Compare.equal(Iterator,Iterator,Comparator) instead
     */
    static public <T> boolean
    equal(Iterator<? extends T> iterator1, Iterator<? extends T> iterator2, Comparator<T> comp)
    {
        return Compare.equal(iterator1, iterator2, comp);
    }

    
    /**
     * Returns true if the two iterations are equal, using the given
     * BinaryFunctor to compare elements in the iterations.
     * @return true if the two iterations are equal
     * @deprecated use Compare.equal(Iterator,Iterator,BinaryFunctor) instead
     */
    static public <T> boolean
    equal(Iterator<? extends T> iter1, Iterator<? extends T> iter2, BinaryFunctor<T,T,Boolean> eq)
    {
        return Compare.equal(iter1, iter2, eq);
    }

    
    /**
     * Returns true if the first iterator is lexically less than the second,
     * using the default comparison operation to compare the elements in each
     * iterator.
     * @return true if the first iteration is less than the second
     * @deprecated use Compare.lessThan(Iterator,Iterator) instead
     */
    static public <T extends Comparable<? super T>> boolean
    lessThan(Iterator<? extends T> iter1, Iterator<? extends T> iter2)
    {
        return Compare.lessThan(iter1, iter2);
    }

    
    /**
     * Returns true if the first iterator is lexically less than the second,
     * using the given comparator to compare the elements in each iterator.
     * @return true if the first iteration is less than the second
     * @deprecated use Compare.lessThan(Iterator,Iterator,Comparator)
     */
    static public <T> boolean
    lessThan(Iterator<? extends T> iter1, Iterator<? extends T> iter2, Comparator<T> comp)
    {
        return Compare.lessThan(iter1, iter2, comp);
    }

    
    /**
     * Returns true if the first iterator is lexically less than the second,
     * using the given operator to compare the elements in each iterator.  The
     * first is less than the second if it is not longer than the second and if
     * the first corresponding element that is not equal is less.
     * @return true if the first iteration is less than the second
     * @deprecated use Compare.lessThan(Iterator,Iterator,BinaryFunctor) instead
     */
    static public <T> boolean
    lessThan(Iterator<? extends T> i1, Iterator<? extends T> i2, BinaryFunctor<T,T,Boolean> lt)
    {
        return Compare.lessThan(i1, i2, lt);
    }
}
