// ============================================================================
// $Id: LookAheadIterator.java,v 1.20 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.util;

import java.util.NoSuchElementException;
import java.util.Iterator;

/**
 * Iterator that allows the program to look at and operate on the next few
 * elements without consuming them.
 * <p>
 * Copyright &copy; 2003-2005  David A. Hall
 *
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

public class LookAheadIterator<T> implements Iterator<T>, Iterable<T> {
    // The base iterator
    private Iterator<? extends T> _base;

    // ring buffer, to hold the elements that have been read ahead of the
    // current position.
    private Object[] _buffer;

    // index of the next element to read from the buffer
    private int _readptr = 0;

    // index of the next element to write to the buffer
    private int _writeptr = 0;

    // number of elements in the buffer
    private int _cnt = 0;

    // size of the buffer
    private int _size;

    /**
     * Builds a LookAheadIterator that can look ahead 1 element.
     */
      
    public LookAheadIterator (Iterator<? extends T> base) {
        this(base, 1);
    }

    /**
     * Builds a LookAheadIterator that can look ahead the given number of
     * elements.
     * 
     * @throws IllegalArgumentException if max <= 0.
     */
      
    public LookAheadIterator (Iterator<? extends T> base, int max) {
        if (max <= 0)
            throw new IllegalArgumentException();
        
        _base = (base != null) ? base : new EmptyIterator<T>();
        _size = max;

        _buffer = new Object[_size];
    }

    /**
     * Returns true if there is an element at the Nth position.  Put another
     * way, returns true if there are enough elements remaining in the iterator
     * that next() could be called N times without having a
     * NoSuchElementException thrown.
     *
     * @return true if there is an element at the Nth position
     * @throws IllegalArgumentException if n < 0 or n > max lookahead
     */
    
    public boolean hasNextPlus(int n) {
        if (n < 0 || n > _size)
            throw new IllegalArgumentException();

        if (n == 0)
            return hasNext();

        return readAhead(n);
    }

    /**
     * Returns the element at the Nth position.  Put another way, returns the
     * element that the Nth call to next() would return.  The current position
     * of the iteration is not modified.
     *
     * @return the element at the Nth position
     * @throws IllegalArgumentException if n < 0 or n > max lookahead
     * @throws NoSuchElementException if the Nth position is off the end of
     *         the iteration
     */
    @SuppressWarnings("unchecked")
    // The buffer array is private, and the only things that we ever put
    // into it are known to be T objects because they are read from an
    // Iterator<? extends T>, so we know that the cast is always valid
    public T peek(int n) {
        if (n < 0 || n > _size)
            throw new IllegalArgumentException();

        if (!readAhead(n)) {
            throw new NoSuchElementException();
        }
            
        int offset = (_readptr + n - 1) % _size;
        return (T) _buffer[offset];
    }
    
    /**
     * Returns the maximum offset that may be peeked.
     */
    public int getMaxPeekSize() {
        return _size;
    }

    // - - - - - - - - - - -
    // Iterable<T> interface
    // - - - - - - - - - - -

    public Iterator<T> iterator() { return this; }
    
    // - - - - - - - - - - -
    // Iterator<T> interface
    // - - - - - - - - - - -
    
    public boolean hasNext() {
        if (_cnt > 0)
            return true;
        else 
            return _base.hasNext();
    }

    @SuppressWarnings("unchecked")
    // The buffer array is private, and the only things that we ever put
    // into it are known to be T objects because they are read from an
    // Iterator<? extends T>, so we know that the cast is always valid
    public T next() {
        if (_cnt == 0)
            return _base.next();

        T value = (T) _buffer[_readptr];
        _readptr = advance(_readptr);
        _cnt--;
        return value;
    }

    // private implementation
    
    public void remove() { throw new UnsupportedOperationException(); }

        private boolean readAhead(int n) {
        if (n <= _cnt)
            return true;
        
        while (n > _cnt &&_base.hasNext()) {
            // this line generates an unchecked cast warning, although it should
            // be safe in this case: it would be difficult to find a way to
            // corrupt the private array
            _buffer[_writeptr] = (T) _base.next();
            
            _cnt++;
            _writeptr = advance(_writeptr);
        }

        return n == _cnt;
    }
    
    private int advance(int n) {
        return ++n % _size;
    }
}
