// ============================================================================
// $Id: NullComparator.java,v 1.2 2009/01/19 03:37:21 davidahall Exp $
// Copyright (c) 2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================

package net.sf.jga.util;

import java.io.Serializable;
import java.util.Comparator;

/**
 * Comparator Decorator used to gracefully compare null and non-null values.
 * In most cases, passing a null to a Comparator results in a NullPointerException,
 * as (in the general case) null is not part of a class' strict ordering.
 * However, in many applications, null objects must be included in an ordering.
 * This wrapper can be used to shield an underlying Comparator from dealing
 * with where in the ordering null values should fall.
 * <p>
 * A flag is provided that determines whether nulls are less than non-nulls,
 * (the default case) or greater than non-nulls.  When comparing two values,
 * this comparator determines if either argument is null: if both are null,
 * then 0 is returned (ie, all nulls are considered equal).  If neither argument
 * is null, then both arguments are passed to the wrapped comparator.  If one
 * argument is null, then a positive or negative integer is returned, depending
 * on which argument was null and the setting of the flag.
 */

// UNTESTED

public class NullComparator<T> implements Comparator<T>, Serializable {

    private static final long serialVersionUID = 5959598817444151446L;

    // Flag determining if nulls appear before or after non-nulls in an ordering
    private boolean _nullsAreLess = true;

    // Comparator used to compare pairs of non-null values
    private Comparator<T> _comp;

    /**
     * Builds a NullComparator that uses the given comparator to compare pairs
     * of non-null values.  Null values appear before non-null values in the
     * resulting ordering.
     */
    public NullComparator(Comparator<T> comp) {
        this(comp, true);
    }

    /**
     * Builds a NullComparator that uses the given comparator to compare pairs
     * of non-null values, and uses the flag to determine if non-null values
     * appear before non-null values in the resulting ordering.
     */
    public NullComparator(Comparator<T> comp, boolean flag) {
        if (comp == null)
            throw new IllegalArgumentException("A Comparator is required");

        _comp = comp;
        _nullsAreLess = flag;
    }


    /**
     * Returns the comparator used to compare pairs of non-null values.
     */
    public Comparator<T> getComparator() { return _comp; }


    
    public int compare(T x, T y) {
        if (x == null)
            if (y == null)
                return 0;
            else
                return _nullsAreLess ? -1 : 1;
        else
            if (y == null)
                return _nullsAreLess ? 1 : -1;
            else
                return _comp.compare(x, y);
   }
}
