// ============================================================================
// $Id: SingletonIterator.java,v 1.9 2005/08/02 23:45:22 davidahall Exp $
// Copyright (c) 2003-2005  David A. Hall
// ============================================================================
// The contents of this file are subject to the Common Development and
// Distribution License (CDDL), Version 1.0 (the License); you may not use this 
// file except in compliance with the License.  You should have received a copy
// of the the License along with this file: if not, a copy of the License is 
// available from Sun Microsystems, Inc.
//
// http://www.sun.com/cddl/cddl.html
//
// From time to time, the license steward (initially Sun Microsystems, Inc.) may
// publish revised and/or new versions of the License.  You may not use,  
// distribute, or otherwise make this file available under subsequent versions 
// of the License.
// 
// Alternatively, the contents of this file may be used under the terms of the 
// GNU Lesser General Public License Version 2.1 or later (the "LGPL"), in which
// case the provisions of the LGPL are applicable instead of those above. If you 
// wish to allow use of your version of this file only under the terms of the 
// LGPL, and not to allow others to use your version of this file under the 
// terms of the CDDL, indicate your decision by deleting the provisions above 
// and replace them with the notice and other provisions required by the LGPL. 
// If you do not delete the provisions above, a recipient may use your version 
// of this file under the terms of either the CDDL or the LGPL.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// ============================================================================
package net.sf.jga.util;

import java.util.ListIterator;
import java.util.NoSuchElementException;
import java.util.Iterator;

/**
 * Iterates over a single item.  The iterator is structured as a list iterator,
 * but the list is a fixed size (1).  The value may be changed after it has
 * been retrieved at least one time.
 *
 * @author <a href="mailto:davidahall@users.sf.net">David A. Hall</a>
 */

public class SingletonIterator<T> implements ListIterator<T>, Iterable<T> {
    // The value returned by the iterator
    private T _value;

    // Flag indicating that the 'pointer' is set to the beginning of the
    // imaginary list.
    private boolean _beforeValue = true;

    // Flag indicating if the current value can be changed.
    private boolean _canSetValue = false;

    /**
     * Builds a SingletonIterator that will return the given value.
     */
    public SingletonIterator(T value) {
        _value = value;        
    }

    // - - - - - - - - - - -
    // Iterable<T> interface
    // - - - - - - - - - - -

    public Iterator<T> iterator() { return this; }
    
    // - - - - - - - - - - - - -
    // ListIterator<T> interface
    // - - - - - - - - - - - - -
    
    /**
     * Returns true if the item has not yet been returned.
     * @return true if the item has not yet been returned
     */
    public boolean hasNext() {
        return _beforeValue;
    }

    /**
     * Returns the single item
     * @return the single item
     * @throws NoSuchElementException if the item has already been returned
     */
    public T next() {
        if (!_beforeValue)
            throw new NoSuchElementException();

        _beforeValue = false;
        _canSetValue = true;
        return _value;
    }

    /**
     * Returns true if the item has not yet been returned.
     * @return true if the item has not yet been returned
     */
    public boolean hasPrevious() {
        return !_beforeValue;
    }

    /**
     * Returns the single item
     * @return the single item
     * @throws NoSuchElementException if the item has already been returned
     */
    public T previous() {
        if (_beforeValue)
            throw new NoSuchElementException();

        _beforeValue = true;
        _canSetValue = true;
        return _value;
    }

    public int nextIndex() {
        return _beforeValue ? 0 : 1;
    }

    public int previousIndex() {
        return _beforeValue ? -1 : 0;
    }
        
    public void add(T arg)
    {
        throw new UnsupportedOperationException();
    }

    public void set(T value) {
        if (!_canSetValue)
            throw new IllegalStateException();

        _value = value;
        _canSetValue = false;
    }

    public void remove() {
        throw new UnsupportedOperationException();
    }
}
